/////////////////////////////////////////////////////////////////////////////
// Name:        No names yet.
// Purpose:     Contrib. demo
// Author:      Aleksandras Gluchovas
// Modified by:
// Created:     08/05/1999
// RCS-ID:      $Id: finddlg.cpp,v 1.8 2001/11/25 10:40:52 hansvl Exp $
// Copyright:   (c) Aleksandras Gluchovas
// Licence:     wxWindows licence
/////////////////////////////////////////////////////////////////////////////

#ifdef __GNUG__
    #pragma implementation "finddlg.h"
#endif

// For compilers that support precompilation, includes "wx/wx.h".
#include "wx/wxprec.h"

#ifdef __BORLANDC__
#pragma hdrstop
#endif

#ifndef WX_PRECOMP
#include "wx/wx.h"
#endif

#if !wxCHECK_VERSION(2,3,2)
#error wxWorkshop requires wxWindows v2.3.2 or higher to compile!
#endif

#include "wx/pmf/finddlg.h"
#include "wx/pmf/pm.h"

/***** Implementation for class wxFindReplaceHelper *****/

// Declare a event table. 
// So we can set a deafault behavior...
BEGIN_EVENT_TABLE(wxFindReplaceHelper, wxEvtHandler)
    EVT_FIND_CLOSE(-1, wxFindReplaceHelper::OnCloseFindDialog)

    EVT_FIND(-1, wxFindReplaceHelper::OnFindReplace)
    EVT_FIND_NEXT(-1, wxFindReplaceHelper::OnFindReplace)
    EVT_FIND_REPLACE(-1, wxFindReplaceHelper::OnFindReplace)
    EVT_FIND_REPLACE_ALL(-1, wxFindReplaceHelper::OnFindReplace)
END_EVENT_TABLE()


wxFindReplaceHelper::wxFindReplaceHelper( wxWindow *parent,
                                         wxFindReplaceData *data,
                                         const wxString &title,
                                         long flags)
            : wxFindReplaceDialog(parent, data, title, flags)
{
    wxASSERT(data);     // DBG::Make sure this is not NULL!
    wxASSERT(parent);   // DBG::The parent may not be NULL!
    
    // Keep a pointer, so that it can be deleted when this class is destroyed.
    //m_pFindData = data; // Already done...
    
    // Init our default view pointer...
    m_pView = NULL;
}

wxFindReplaceHelper::~wxFindReplaceHelper()
{
    // Clean up event handlers
    if (m_pView)
    {
        // Remove old View event handler
        PopEventHandler(FALSE /* Don't Delete handler */);
    }
    
    // Reset intance, none loaded anymore.
    m_pSingleton = NULL;
}

void wxFindReplaceHelper::SetActiveView( wxWindow *pView )
{
    if (m_pView != pView)
    {
        if (m_pView)
        {
            // Remove old View event handler
            PopEventHandler(FALSE /* Don't Delete handler */);
        }
        
        m_pView = pView;
        
        // Set new event handlers
        if (m_pView)
            PushEventHandler( m_pView );
    }
}

// Event handlers
void wxFindReplaceHelper::OnCloseFindDialog(wxFindDialogEvent& event)
{
    wxASSERT(event.GetDialog()) // DBG::Something is realy wrong...
        
    // Make sure that we start cleaning up!
    if(wxFindReplaceHelper::IsInstanceCreated())
        event.GetDialog()->Destroy();
    
    // Eat close event (no Skip)... no longer save to use anyway...
}

void wxFindReplaceHelper::OnFindReplace(wxFindDialogEvent& event)
{
    // Can we skip the event and thus not diplay a message?
    if (m_pView)
        event.Skip();
    else
        wxMessageBox("Don't know where to perform the search\nMake a searchable control active.",
                     "Find/Replace info",
                     wxOK | wxICON_INFORMATION, this);
}

// Static member functions
bool wxFindReplaceHelper::IsInstanceCreated()
{
    return m_pSingleton != NULL;
}

wxFindReplaceHelper* wxFindReplaceHelper::GetInstance( wxWindow *pView, const bool bFindReplace )
{
    if (!IsInstanceCreated())
    {
        long flag = 0;
        wxString title(FindTitle);
        
        // if we create a find replace dialog we need to adjust the style to it.
        if (bFindReplace)
        {
            flag = wxFR_REPLACEDIALOG;
            title = FindReplaceTitle;
        }
        
        // If not already done, create the data holder for the find dialog.
        if (!m_pFindData)
        {
            m_pFindData = new wxFindReplaceData();
            m_pFindData->SetFlags(wxFR_DOWN);
        }
        // Creat our self :-)
        m_pSingleton = new wxFindReplaceHelper( wxPMBootSvc().GetMainWindow(),
                                                m_pFindData,
                                                title,
                                                flag);
        
        // Set new view...
        m_pSingleton->SetActiveView( pView );
    }
    else
    {
        // if pView == NULL it means we are in a event or so that does not really need to create a new dialog!
        if ((pView != NULL) && 
            ((bFindReplace && !(m_pSingleton->GetWindowStyle() & wxFR_REPLACEDIALOG)) ||
             (!bFindReplace && (m_pSingleton->GetWindowStyle() & wxFR_REPLACEDIALOG))))
        {
            // Delete our self and re-create our self 
            wxFindReplaceHelper::DeleteInstance( FALSE /* Force to delete */ );
            wxFindReplaceHelper::GetInstance( pView, bFindReplace );
        }
    }
    
    wxASSERT(m_pSingleton);
    
    return m_pSingleton;
}

bool wxFindReplaceHelper::CollectFindData(wxString &find, bool &bCase, bool &bWord)
{
    wxFindReplaceData *data = wxFindReplaceHelper::m_pFindData;

    if (data == NULL)
    {
        find = "";      // Empty string to search
        bCase = FALSE;  // No case sencitive find
        bWord = FALSE;  // No hole word find

        return FALSE;
    }
    else
    {
        find = data->GetFindString();
        bCase = ((data->GetFlags() & wxFR_MATCHCASE) != 0);
        bWord = ((data->GetFlags() & wxFR_WHOLEWORD) != 0);

        return TRUE;
    }
}

bool wxFindReplaceHelper::CollectReplaceData(wxString &find, wxString &replace, bool &bCase, bool &bWord)
{
    wxFindReplaceData *data = wxFindReplaceHelper::m_pFindData;

    if (data == NULL)
    {
        find = "";      // Empty string to search
        replace = "";   // Empty replace string.
        bCase = FALSE;  // No case sencitive find
        bWord = FALSE;  // No hole word find

        return FALSE;
    }
    else
    {
        find = data->GetFindString();
        replace = data->GetReplaceString();
        bCase = ((data->GetFlags() & wxFR_MATCHCASE) != 0);
        bWord = ((data->GetFlags() & wxFR_WHOLEWORD) != 0);

        return TRUE;
    }
}

void wxFindReplaceHelper::DeleteInstance(bool bSaveDelete)
{
    if (IsInstanceCreated())
    {
        // In some cases a save delete is not enough! => see GetInstance...
        if (bSaveDelete)
            m_pSingleton->Destroy(); 
        else
            wxDELETE( m_pSingleton );
    }
}

void wxFindReplaceHelper::DeleteFindData()
{
    wxDELETE(m_pFindData);
}

// Init the singleton pointer.
wxFindReplaceHelper* wxFindReplaceHelper::m_pSingleton = NULL;
// Init the data pointer.
wxFindReplaceData* wxFindReplaceHelper::m_pFindData = NULL;

