/////////////////////////////////////////////////////////////////////////////
// Name:        No names yet.
// Purpose:     Contrib. demo
// Author:      Aleksandras Gluchovas
// Modified by:
// Created:     2000/02/08
// RCS-ID:      $Id: build_svc.cpp,v 1.18 2001/11/24 13:49:32 georgetasker Exp $
// Copyright:   (c) Aleksandras Gluchovas (alex@soften.ktu.lt)
// Licence:   	GPL (version 2.0 or later)
/////////////////////////////////////////////////////////////////////////////

#ifdef __GNUG__
    #pragma implementation "build_svc.h"
#endif

// For compilers that support precompilation, includes "wx/wx.h".
#include "wx/wxprec.h"

#ifdef __BORLANDC__
    #pragma hdrstop
#endif

#ifndef WX_PRECOMP
    #include "wx/wx.h"
#endif

#include "build_svc.h"
#include "myutils.h"

#include "wx/xrc/xmlres.h"
#include "wx/wspp/glist.h"
#include "wx/wspp/pagectrl.h"

#include "wx/txtstrm.h"
#include "wx/wfstream.h"

#include "wx/log.h"
#include "wx/timer.h"
#include "wx/msgdlg.h"

#include "wx/pmf/lseditor.h"
#include "wx/pmf/pm_flalg.h"
#include "wx/pmf/pm.h"
#include "wx/pmf/inf/textediting.h"
#include "wx/fl/controlbar.h"

// constants

static const wxString& BUILD_WND_NAME = "build";

/***** Implementation for class ErrorPinPainter *****/

enum
{
	ERRROR_PIN_TC = BOOKMARK_PIN_TC + 10,
	WARINING_PIN_TC
};

class ErrorPinPainter : public TPinPainterBase
{
public:
	ErrorPinPainter( int tc = ERRROR_PIN_TC );

	virtual void DrawPin( TPinBase* pPin, wxTextEditorView& view, wxDC& dc, 
						  const wxPoint& pos, const wxSize& dim );

	virtual void GetVisuals( wxColour& fgCol, wxColour& bgCol, char& ch );
};

ErrorPinPainter::ErrorPinPainter( int tc ) 

	: TPinPainterBase( tc )
{
}

void ErrorPinPainter::DrawPin( TPinBase* pPin, wxTextEditorView& view, wxDC& dc, 
						       const wxPoint& pos, const wxSize& dim )
{
	wxColour fg, bg; char ch;
	GetVisuals( fg, bg, ch );

	wxBrush bkBrush( bg, wxSOLID );
	dc.SetPen( *wxBLACK_PEN );
	dc.SetBrush( bkBrush );

	dc.DrawRoundedRectangle( pos.x+2, pos.y, dim.x-4, dim.y, 4 );
	dc.SetTextForeground( fg );
	dc.SetBackgroundMode( wxTRANSPARENT );
	
	wxString text = ch;
	wxCoord w = 0, h = 0;
	dc.GetTextExtent( text, &w, &h );

	dc.DrawText( text,
				 pos.x + ( dim.x - w ) / 2, 
				 pos.y + ( dim.y - h ) / 2 );
}

void ErrorPinPainter::GetVisuals( wxColour& fgCol, 
								  wxColour& bgCol, char& ch )
{
	ch = 'E';
	fgCol = *wxWHITE;
	bgCol = wxColour(196,0,0);
}

class WarningPainter : public ErrorPinPainter
{
public:
	WarningPainter()
		: ErrorPinPainter( WARINING_PIN_TC )
	{}

	virtual void GetVisuals( wxColour& fgCol, wxColour& bgCol, char& ch )
	{
		ch = 'W';
		fgCol = *wxWHITE;
		bgCol = wxColour(0,0,196);
	}
};

/*** Implementation for class BuildTemplateInfo *****/

BuildTemplateInfo::~BuildTemplateInfo()
{
	WX_CLEAR_ARRAY( mArgs );
}

/*** Implementation for class BuildConfiguration *****/

BuildConfiguration::BuildConfiguration()

	: mpTempl( NULL ),
	  mIsActive( FALSE )
{}

/*** Implementation for class BuildConfigurationsDialog *****/

BEGIN_EVENT_TABLE( BuildConfigurationsDialog, wxDialog )

	EVT_BUTTON( XMLID( "new" ),		 BuildConfigurationsDialog::OnNew )
	EVT_BUTTON( XMLID( "edit" ),	 BuildConfigurationsDialog::OnEdit )
	EVT_BUTTON( XMLID( "activate" ), BuildConfigurationsDialog::OnActivate )
	EVT_BUTTON( XMLID( "remove" ),	 BuildConfigurationsDialog::OnRemove )

	EVT_INIT_DIALOG( BuildConfigurationsDialog::OnInitDialog )

END_EVENT_TABLE()

BuildConfigurationsDialog::BuildConfigurationsDialog( BuildConfigurationListT* pConfigs,
													  BuildTemplateInfoListT* pTemplates,
													  wxPMBuildService* buildSvc )

	: mpConfigs( pConfigs ),
	  mpTemplates( pTemplates ),
	  mpBuildSvc( buildSvc )
{}	 

BuildConfigurationsDialog::~BuildConfigurationsDialog()
{}

void BuildConfigurationsDialog::OnNew( wxCommandEvent& event )
{
	BuildConfiguration* newCfg = new BuildConfiguration();

	BuildConfigurationEditDialog dlg( newCfg, "New Build Configuration", *mpTemplates );

	if ( dlg.ShowModal() == wxID_OK )
	{
		dlg.ExchangeValues( FALSE );

		if ( mpConfigs->Count() == 0 )

			newCfg->mIsActive = TRUE;

		mpConfigs->Add( newCfg );

		wxPMBootSvc().GetActiveProject()->Modify( TRUE );
		mpBuildSvc->UpdateFiles();

		RefillList();
	}
	else
		wxDELETE(newCfg);
}

void BuildConfigurationsDialog::OnEdit( wxCommandEvent& event )
{
    if ( mpList->GetCount() == 0 )
		return;

	BuildConfiguration* cfg = (*mpConfigs)[ mpList->GetSelection() ];

	BuildConfiguration copy = *cfg;

	BuildConfigurationEditDialog dlg( cfg, "Edit Build Configuration", *mpTemplates );

	if ( dlg.ShowModal() != wxID_OK )
	
		*cfg = copy;
	else
	{
		dlg.ExchangeValues( FALSE );

		wxPMBootSvc().GetActiveProject()->Modify( TRUE );
		mpBuildSvc->UpdateFiles();

		RefillList();
	}
}

void BuildConfigurationsDialog::OnActivate( wxCommandEvent& event )
{
	if ( mpList->GetCount() == 0 )
		return;

    size_t i;
	for ( i = 0; i != mpConfigs->Count(); ++i )
    {
		(*mpConfigs)[i]->mIsActive = FALSE;
    }

	(*mpConfigs)[ mpList->GetSelection() ]->mIsActive = TRUE;

	wxPMBootSvc().GetActiveProject()->Modify( TRUE );

	RefillList();
}

void BuildConfigurationsDialog::OnRemove( wxCommandEvent& event )
{
	if ( mpList->GetCount() == 0 )
		return;

	size_t n = (size_t)mpList->GetSelection();
	bool wasActive = (*mpConfigs)[n]->mIsActive;

	delete (*mpConfigs)[n];
	(*mpConfigs).RemoveAt(n);

	if ( wasActive && mpConfigs->Count() )

		(*mpConfigs)[0u]->mIsActive = TRUE;

	wxPMBootSvc().GetActiveProject()->Modify( TRUE );
	mpBuildSvc->UpdateFiles();

	RefillList();
}

void BuildConfigurationsDialog::OnInitDialog( wxInitDialogEvent& event )
{
	mpList = XMLCTRL( *this, "configs_list", wxListBox );

	wxASSERT( mpList ); // DBG::

	RefillList();
}

/*** protected methods ***/

void BuildConfigurationsDialog::RefillList()
{
    if ( mpList->GetCount() )
    {
        mpList->Clear();
    }

    size_t i;
	for ( i = 0; i != mpConfigs->Count(); ++i )
	{
		BuildConfiguration& cfg = *(*mpConfigs)[i];

		if ( cfg.mIsActive )
		{
			mpList->Append( "+" + cfg.mName );
			mpList->SetSelection( i );
		}
		else
			mpList->Append( cfg.mName );
	}
}

/*** Implementation for class BuildConfigurationEditDialog *****/

#define TEMPL_CHOICE_ID 300

BEGIN_EVENT_TABLE( BuildConfigurationEditDialog, wxDialog )

	EVT_CHOICE( TEMPL_CHOICE_ID, BuildConfigurationEditDialog::OnChoice )

END_EVENT_TABLE()

BuildConfigurationEditDialog::BuildConfigurationEditDialog( BuildConfiguration* pCfg, 
															const wxString& title,
															BuildTemplateInfoListT& templates )
	: wxDialog( wxPMBootSvc().GetMainWindow(), 
				-1, title, wxDefaultPosition, wxDefaultSize,
				wxSYSTEM_MENU | wxRESIZE_BORDER | wxRESIZE_BOX ),

	  mpCfg( pCfg ),
	  mpTemplates( &templates )
{
	if ( !mpCfg->mpTempl )

		mpCfg->mpTempl = templates[0u];

	RecreateUI();

	ExchangeValues( TRUE );

	Layout();
}

void BuildConfigurationEditDialog::OnChoice( wxCommandEvent& event )
{
	mpCfg->mpTempl = (*mpTemplates)[ mpTempl->GetSelection() ];
	mpCfg->mFields.Clear();
	mpCfg->mValues.Clear();

	RecreateUI();

	ExchangeValues( TRUE );

	Layout();
}

/*** protected methods ***/

void BuildConfigurationEditDialog::RecreateUI()
{
	wxList children;
	
	wxNode* node = GetChildren().First();

	while( node )
	{
		children.Append( (wxWindowBase*)node->GetData() );

		node = node->GetNext();
	}

	node = children.First();

	while( node )
	{
		((wxWindow*)node->GetData())->Show(FALSE);

		this->RemoveChild( (wxWindow*)node->GetData() );

		node = node->GetNext();
	}

	this->SetSizer( NULL );

	wxBoxSizer* box = new wxBoxSizer( wxVERTICAL );

	wxGridSizer* grid = new wxGridSizer( 2 + mpCfg->mpTempl->mArgs.Count(), 2, 5, 5 );

	wxStaticText* nameLabel = new wxStaticText( this,-1, "Name", wxDefaultPosition );
	mpName = new wxTextCtrl( this, -1, "", wxDefaultPosition );

	size_t n = mpTemplates->Count();
	wxString* choices = new wxString[ n ];

    size_t k;
	for ( k = 0; k != n; ++k )
    {
		choices[k] = (*mpTemplates)[k]->mName;
    }

	wxStaticText* templLabel = new wxStaticText( this,-1, "Template", wxDefaultPosition );
	mpTempl = new wxChoice( this, TEMPL_CHOICE_ID, wxDefaultPosition, wxDefaultSize, n, choices );

	wxDELETEA(choices);

	grid->Add( nameLabel );
	grid->Add( mpName );
	grid->Add( templLabel );
	grid->Add( mpTempl );

	BuildArgInfoListT& lst = mpCfg->mpTempl->mArgs;

    size_t i;
	for ( i = 0; i != lst.Count(); ++i )
	{
		BuildArgInfo& arg = *lst[i];

		wxString txt = arg.mName;

		if ( arg.mValues.GetCount() )
		{
			txt += " (";

            size_t k;
			for ( k = 0; k != arg.mValues.GetCount(); ++k )
			{
				if ( k != 0 ) 
					txt += ",";

				txt += wxString( arg.mValues[k] );
			}

			txt += ")";
		}
		
		wxString val = arg.mValues.GetCount() ? arg.mValues[0u] : "";

		wxStaticText* name = new wxStaticText( this,-1, txt, wxDefaultPosition );
		wxTextCtrl* value = new wxTextCtrl( this,-1, val, wxDefaultPosition );

		grid->Add( name );
		grid->Add( value );

		arg.mpUserData = (void*)value;
	}

	box->Add( grid, 0, wxALL, 5 );

	wxBoxSizer* btnBox = new wxBoxSizer( wxHORIZONTAL );

	btnBox->Add( new wxButton( this, wxID_OK, "&Ok", wxDefaultPosition ), 0, wxALL, 2 );
	btnBox->Add( new wxButton( this, wxID_CANCEL, "&Cancel", wxDefaultPosition ), 0, wxALL, 2 );

	box->Add( btnBox, 0, wxALL, 5 );

	this->SetAutoLayout( TRUE );
	this->SetSizer( box );

	box->Fit( this );
	box->SetSizeHints( this );

	mpName->SetFocus();

	Center( wxBOTH );
}

void BuildConfigurationEditDialog::ExchangeValues( bool toControls )
{
	if ( toControls )
	{
        size_t k;
		for ( k = 0; k != mpTemplates->Count(); ++k )
        {
			if ( (*mpTemplates)[k] == mpCfg->mpTempl )
			{
				mpTempl->SetSelection(k);
				break;
			}
        }

        size_t i;
		for ( i = 0; i != mpCfg->mFields.GetCount(); ++i )
        {
			((wxTextCtrl*)mpCfg->mpTempl->mArgs[i]->mpUserData)
				->SetValue( mpCfg->mValues[i] );
        }
		mpName->SetValue( mpCfg->mName );
	}
	else
	{
		mpCfg->mFields.Clear();
		mpCfg->mValues.Clear();

		mpCfg->mName = mpName->GetValue();
		mpCfg->mpTempl = (*mpTemplates)[ mpTempl->GetSelection() ];

        size_t i;
		for ( i = 0; i != mpCfg->mpTempl->mArgs.Count(); ++i )
		{
			BuildArgInfo& arg = *mpCfg->mpTempl->mArgs[i];

			mpCfg->mFields.Add( arg.mName );
			mpCfg->mValues.Add( ((wxTextCtrl*)arg.mpUserData)->GetValue() );
		}
	}
}

/*** Implementation for class PMProjectConfigSet *****/

PMProjectConfigSet::~PMProjectConfigSet()
{
	CleanUp();
}

void PMProjectConfigSet::CleanUp()
{
	WX_CLEAR_ARRAY( mConfigs );
}

void PMProjectConfigSet::AddConfig( BuildConfiguration* cfg )
{
	mConfigs.Add( cfg );
}

void PMProjectConfigSet::RemoveConfig( BuildConfiguration* cfg )
{
	mConfigs.Remove( cfg );
	wxDELETE(cfg);
}

BuildConfigurationListT& PMProjectConfigSet::GetConfigs()
{
	return mConfigs;
}

BuildConfiguration* PMProjectConfigSet::GetActiveConfig()
{
    size_t i;
	for ( i = 0; i != mConfigs.Count(); ++i )
    {
		if ( mConfigs[i]->mIsActive )
			return mConfigs[i];
    }
	return NULL;
}
/*** Implementation for class BuildScriptTemplateHandler *****/

BuildScriptTemplateHandler::BuildScriptTemplateHandler( wxPMBuildService* buildSvc )

	: BasicScriptProcessorTemplate( buildSvc ),
	  mpBuildSvc( buildSvc )
{}

wxString BuildScriptTemplateHandler::HandleSpecialSubstitution( char* at, ScriptTemplate& templ  )
{
	char* start = at;

	while( *at && *at != '[' && *at != ')' ) ++at;

	wxString tagName;
	ConcatString( tagName, start, size_t( at - start ) );

	if ( tagName == "file_list" )

		return Handle_file_list( at, templ );
	if ( tagName == "script" )

		return Handle_subscript( at, templ );
	else
		return "[ERROR: Unrecognized special-substitution tag: " + tagName + "]";
}

/*** proteced methods ***/

wxString BuildScriptTemplateHandler::Handle_file_list( char* at, ScriptTemplate& templ )
{
	wxString result;
	++at; // skip '['

	if ( *at == '"' ) ++at; // skip starting '"'

	char* start = at; 
	while( *at && *at != ']' && *at != '"' && *at != ',' ) ++at;

	wxString mask;
	ConcatString( mask, start, size_t(at - start) );

	if ( *at == '"' ) ++at;
	++at; // skip ','

	start = at; 
	while( *at && *at != ']' ) ++at;

	wxString script;
	ConcatString( script, start, size_t(at - start) );

	ScriptTemplate* pTempl = mpBuildSvc->FindSubscriptByName( script );

	if ( pTempl == NULL )

		return "[@file_list:ERROR: Requested entry-subscript " + script + " not found!]";

	pTempl->SetParentTemplate( &templ );

	wxPMProject* prj = wxPMBootSvc().GetActiveProject();
	wxASSERT( prj ); // DBG::

	wxString content;

	PMFileInfoListT& files = prj->GetFiles();

    size_t i;
	for ( i = 0; i != files.Count(); ++i )
	{
		wxString fname = files[i]->GetPath();
		fname = prj->GetRelativePath( fname );

		if ( ::wxMatchWild( mask, fname, FALSE ) )
		{
			pTempl->Reset();
			pTempl->SetVar( "FILE", MakeTmpString( fname ) );

			int j = fname.Length() - 1;

			while ( j >= 0 )
			{
				if ( fname[(unsigned int)j] == '.' )
					break;
				--j;
			}

			if ( j == 0 ) 
				j = fname.Length();

			pTempl->SetVar( "FILE_NO_EXT", 
							MakeTmpString( fname.Mid( 0, j ) ) );

			content += pTempl->SubstitueVariables();
		}
	}

	return content;
}

/*** Implementation for class wxPMBuildService *****/

static wxCommandInfo _gBuildCommands[] = {

	wxCommandInfo( XMLID( "build_build" ),			"&Build", "", "F12" ),
	wxCommandInfo( XMLID( "build_execute" ),		"&Execute", "", "Ctrl+F6" ),
	wxCommandInfo( XMLID( "build_and_execute" ),	"Build and &Excecute", "", "F6" ),

	wxCommandInfo( -1, "SEPARATOR" ),
	wxCommandInfo( XMLID( "build_clean" ),			"&Clean", "", "" ),
	wxCommandInfo( XMLID( "build_configure" ),		"Co&nfigurations...", "", "Alt+F7" ),
	wxCommandInfo( -1, "SEPARATOR" ),
	wxCommandInfo( XMLID( "build_next_msg" ),		"&Next Build Message", "", "F4" ),
	wxCommandInfo( XMLID( "build_prev_msg" ),		"&Previous Build Message", "", "Shift+F4" ),
	wxCommandInfo( -1, "SEPARATOR" ),
	wxCommandInfo( XMLID( "build_switch_output" ),  "&Switch to Output Window", "", "Alt+9" ),

	wxCommandInfo( 0, "" ) // list-terminator
};

static wxCommandInfo _gBuildCommands1[] = {

	wxCommandInfo( XMLID( "build_toggle_output" ), "&Output Window", "", "Alt+2", TRUE, TRUE ),

	wxCommandInfo( 0, "" ) // terminator
};

static wxPMMergableMenuChain _gBuildChains[] = {

	wxPMMergableMenuChain( "/&Build/", _gBuildCommands ),
	wxPMMergableMenuChain( "/&View/",  _gBuildCommands1 ),

	wxPMMergableMenuChain( "", NULL ) // list-terminator
};

IMPLEMENT_DYNAMIC_CLASS( wxPMBuildService, wxPMPane )

BEGIN_EVENT_TABLE( wxPMBuildService, wxPMPane )

	EVT_BUTTON( XMLID( "build_build" ), wxPMBuildService::OnBuild )
	EVT_BUTTON( XMLID( "build_clean" ), wxPMBuildService::OnClean )
	EVT_BUTTON( XMLID( "build_configure" ), wxPMBuildService::OnConfigure )
	EVT_BUTTON( XMLID( "build_and_execute" ), wxPMBuildService::OnBuildAndExecute )
	EVT_BUTTON( XMLID( "build_execute" ), wxPMBuildService::OnExecute )

	EVT_MENU( XMLID( "build_build" ), wxPMBuildService::OnBuild )
	EVT_MENU( XMLID( "build_clean" ), wxPMBuildService::OnClean )
	EVT_MENU( XMLID( "build_configure" ), wxPMBuildService::OnConfigure )
	EVT_MENU( XMLID( "build_and_execute" ), wxPMBuildService::OnBuildAndExecute )
	EVT_MENU( XMLID( "build_execute" ), wxPMBuildService::OnExecute )
	EVT_MENU( XMLID( "build_toggle_output" ), wxPMBuildService::OnToggleOutputWindow )
	EVT_MENU( XMLID( "build_switch_output" ), wxPMBuildService::OnSwitchToOutputWindow )

	EVT_MENU( XMLID( "build_next_msg" ), wxPMBuildService::OnNextMessage )
	EVT_MENU( XMLID( "build_prev_msg" ), wxPMBuildService::OnPreviousMessage )

	EVT_BUTTON( XMLID( "notify_build_started" ), wxPMBuildService::OnBuildStarted )
	EVT_BUTTON( XMLID( "notify_message_added" ), wxPMBuildService::OnBuildMessageListUpdated )
	EVT_BUTTON( XMLID( "notify_build_finished" ), wxPMBuildService::OnBuildFinished )

	EVT_END_PROCESS ( XMLID( "on_process_terminated" ), wxPMBuildService::OnProcessTerminated )

	EVT_TIMER( 5, wxPMBuildService::OnBuildTimer )

END_EVENT_TABLE()

wxPMBuildService::wxPMBuildService()

	: mpBuildProcess( NULL ),
      mpBuildingCfg( NULL ),
      mNeedToExecute( FALSE ),  
      mpInfo( NULL ),
      mpPanel( NULL ),
      mpPageCtrl( NULL ),
      mTimer( this, 5 )
{}

wxPMBuildService::~wxPMBuildService()
{
    CleanUp();
}

bool wxPMBuildService::Start()
{
	if ( !mpTopTempl )
	{
		SetPaneName( "Output Window" );

		ScriptTemplate::SetVarPrefixChar( '%' );

		SetProcessorTemplate( new BuildScriptTemplateHandler( this ) );

		wxString sd = wxPMBootSvc().GetStartupDir();

		wxString fname = sd + wxString(wxFILE_SEP_PATH) + "config_build.txt";

		LoadAndProcessConfig( fname );

		wxPMBootSvc().GetDocManager()->AddListener( (wxPMDocumentListener*)this );

		ScriptTemplate::SetVarPrefixChar();

		wxPMBootSvc().RegisterUniqueServiceInstance( this, "I_OutputWindow" );
	}

	return wxPMPane::Start();
}

bool wxPMBuildService::Stop()
{
	if ( mpPanel )

		mpPanel->SetNextHandler( NULL );

	wxPMBootSvc().GetDocManager()->RemoveListener( (wxPMDocumentListener*)this );

	return wxPMPane::Stop();
}

void* wxPMBuildService::GetInterface( const char* infName, const char* version )
{
	if ( strcmp( infName, "I_OutputWindow" ) == 0 )

		return (void*)(I_OutputWindow*)this;
	else
		return NULL;
}

wxPMMergableMenuChain* wxPMBuildService::GetMergableMenuChain()
{
	return _gBuildChains;
}

wxWindow* wxPMBuildService::GetPaneWindow()
{
	return mpPanel;
}

void wxPMBuildService::CreatePaneWindow( wxWindow* pParent )
{
	wxASSERT(pParent); // DBG::	
	mpPanel = wxTheXmlResource->LoadWindow( pParent, wxT("BuildSvc_Panel1") );
	if (!mpPanel)
		return;
	
	mpPanel->SetNextHandler( this );

	mpPageCtrl = (wxPagingCtrl*)wxXmlFindWindow( "output_pages", mpPanel );

	mpInfo = AddOutputWindow( BUILD_WND_NAME, "Build" );
	mpInfo->mpView->SetKeyInterceptor( this );
}

static wxPMFrameLayoutPreference _gOutputLayPref( FL_ALIGN_BOTTOM, 400,0, 100, 180, FALSE );

wxPMLayoutPreference* wxPMBuildService::GetLayoutPefs()
{
	if ( !wxPMBootSvc().FindUniqueServiceInstance( "I_PMPaneLayoutAlgorithm" ) )
		return NULL;

	return &_gOutputLayPref;
}

#define BUILD_SECTION_NAME "/Build_Configurations"

bool wxPMBuildService::OnClosingDocument( wxPMDocument& doc, wxOutputStream* s, wxPMSerializer* ser )
{
	if ( !ser || !doc.IsKindOf( CLASSINFO(wxPMProject) ) )
		return TRUE;

	if ( ser->Exists( BUILD_SECTION_NAME ) )
	
		ser->DeleteGroup( BUILD_SECTION_NAME );

	ser->SetPath( BUILD_SECTION_NAME );

	PMProjectConfigSet* cfgSet = DoGetConfigForProject( doc );
	BuildConfigurationListT& lst = cfgSet->GetConfigs();

    size_t i;
	for ( i = 0; i != lst.Count(); ++i )
	{
		BuildConfiguration* cfg = lst[i];

		ser->SetPath( BUILD_SECTION_NAME + wxString( "/Entry_" + IntToStr( i ) ) );

		ser->Exchange( "build_template", cfg->mpTempl->mName );
		ser->Exchange( "config_name",    cfg->mName );
		ser->Exchange( "is_active",		 cfg->mIsActive );

        size_t k;
		for ( k = 0; k != cfg->mFields.GetCount(); ++k )
		{
			wxString tmpStr(cfg->mValues[k]);
			ser->Exchange( wxString( cfg->mFields[k] ), tmpStr );
		}
	}

	return TRUE;
}

void wxPMBuildService::OnFileAdded( wxPMFileInfo& file, wxPMProject* toProject )
{
	PMProjectConfigSet* set = DoGetConfigForProject( *toProject );
	DoUpdateOutputFiles( set, TRUE );
	DoUpdateOutputFiles( set, FALSE );
}

void wxPMBuildService::OnFileRemoved( wxPMFileInfo& file, wxPMProject* fromProject  )
{
	PMProjectConfigSet* set = DoGetConfigForProject( *fromProject );
	DoUpdateOutputFiles( set, TRUE );
	DoUpdateOutputFiles( set, FALSE );
}

bool wxPMBuildService::OnOpeningDocument( wxPMDocument& doc, wxInputStream* s, wxPMSerializer* ser )
{
	if ( doc.IsKindOf( CLASSINFO( wxPMWorkplace ) ) )

		((wxPMWorkplace*)&doc)->AddListener( this );

	if ( !ser || !doc.IsKindOf( CLASSINFO(wxPMProject) ) )
		return TRUE;

	if ( ser->Exists( BUILD_SECTION_NAME ) )
	{
		PMProjectConfigSet* cfgSet = DoGetConfigForProject( doc );

		cfgSet->CleanUp();

		ser->SetPath( BUILD_SECTION_NAME );

		long idx = 0; wxString name;
		bool ok = ser->GetFirstGroup( name, idx );

		while( ok )
		{
			ser->SetPath( BUILD_SECTION_NAME + wxString( "/" + name ) );

			BuildConfiguration* cfg = new BuildConfiguration();

			wxString templName;
			
			ser->Exchange( "config_name",    cfg->mName );
			ser->Exchange( "build_template", templName  );
			ser->Exchange( "is_active",      cfg->mIsActive );

			cfg->mpTempl = FindTemplateByName( templName );

			wxASSERT( cfg->mpTempl ); // DBG::

            size_t i;
			for ( i = 0; i != cfg->mpTempl->mArgs.Count(); ++i )
			{
				BuildArgInfo* argInfo = cfg->mpTempl->mArgs[i];

				cfg->mFields.Add( argInfo->mName );

				wxString value;
				ser->Exchange( argInfo->mName, value );
				cfg->mValues.Add( value );
			}

			cfgSet->AddConfig( cfg );

			ser->SetPath( BUILD_SECTION_NAME );
			ok = ser->GetNextGroup( name, idx );
		}
	}

	return TRUE;
}

bool wxPMBuildService::ProcessSection( Record& rec )
{
	if ( rec.GetName() == "template" )
	{
		BuildTemplateInfo* info = new BuildTemplateInfo();

		info->mName = rec[ "name" ];

		ParseFilters( rec[ "error_filters" ], info->mErrorFilters );
		ParseFilters( rec[ "warning_filters" ], info->mWarningFilters );
		//GetStringListVar( rec, "error_filter", info->mErrorFilters );
		//GetStringListVar( rec, "warning_filter", info->mWarningFilters );

		info->mContent.SetParentTemplate( mpTopTempl );
		info->mOutputFile.SetParentTemplate( mpTopTempl );
		info->mBuildCmdLine.SetParentTemplate( mpTopTempl );
		info->mCleanCmdLine.SetParentTemplate( mpTopTempl );
		info->mExecCmdLine.SetParentTemplate( mpTopTempl );

		info->mContent.SetScript( rec[ "content" ] );
		info->mOutputFile.SetScript( rec[ "output_file" ] );
		info->mBuildCmdLine.SetScript( rec[ "build_cmdline" ] );
		info->mCleanCmdLine.SetScript( rec[ "clean_cmdline" ] );
		info->mExecCmdLine.SetScript( rec[ "exec_cmdline" ] );

		Record* args = (Record*)rec.FindField( "args" );

		if ( args )
		{
            size_t i;
			for ( i = 0; i != args->GetFieldCount(); ++i )
			{
				Record& arg = args->GetRecordByIndex(i);

				BuildArgInfo* argInfo = new BuildArgInfo();

				argInfo->mName = arg[0u];

				if ( arg.GetFieldCount() > 1 )
				{
					Record& values = arg.GetRecordByIndex(1);

                    size_t k;
					for ( k = 0; k != values.GetFieldCount(); ++k )
						argInfo->mValues.Add( values[k] );
     			}

				info->mArgs.Add( argInfo );
			}
		}

		mTemplates.Add( info );

	} // end of if( .. )

	return TRUE;
}

void wxPMBuildService::CleanUp()
{
	ScriptProcessor::CleanUp();

    size_t i, cnt;
    cnt = mOutputWindows.Count();
    for (i = 0; i < cnt; i++)
        mOutputWindows.RemoveAt(0);
    mOutputWindows.Clear();

	WX_CLEAR_ARRAY( mTemplates );

    wxDELETE(mpInfo);
}

BuildTemplateInfo* wxPMBuildService::FindTemplateByName( const wxString& name )
{
    size_t i;
	for ( i = 0; i != mTemplates.Count(); ++i )
    {
		if ( mTemplates[i]->mName == name )
			return mTemplates[i];
    }
	return NULL;
}

/*** command handlers ***/

void wxPMBuildService::OnBuild( wxCommandEvent& event )
{
	mNeedToExecute = FALSE;
	DoExecute( TRUE );
}

void wxPMBuildService::OnClean( wxCommandEvent& event )
{
	DoExecute( FALSE );
}

void wxPMBuildService::OnBuildAndExecute( wxCommandEvent& event )
{
	mNeedToExecute = TRUE;
	DoExecute( TRUE );
}

void wxPMBuildService::OnExecute( wxCommandEvent& event )
{
	if ( mpBuildProcess )
	{
		wxMessageBox( "Building is already in progress..." );
		return;
	}

	// Do this with ASSERTs because I feel that the button should be
	// Disabled to prevent the assert to occur.
	wxASSERT( &wxPMBootSvc() != NULL );					// DBG::Ugly code... 
			// Can we make IsInit or IsEmpty function to do it more elegant?
	
	wxPMProject* prj = wxPMBootSvc().GetActiveProject();

	PMProjectConfigSet* set = 
		DoGetConfigForProject( *prj );

	BuildConfiguration* cfg = set->GetActiveConfig();

    // Give the user information if we can't continue with his request.
	if ( !cfg )
	{
		wxMessageBox(wxT("No active build configuration is set up\nUse \"Build | Configurations\" menu to create one"),
                     wxT("Config error."),
                     wxOK | wxCENTRE | wxICON_EXCLAMATION,
                     wxPMBootSvc().GetMainWindow()
                    );

		return; // Don't continue!
	}
	
	ReduceBuildTimeVars( *cfg, *prj );

	long ret = ::wxExecute( cfg->mExecCmdLine, FALSE, NULL );

	if ( !ret )
	{
		wxMessageDialog dlg( wxPMBootSvc().GetMainWindow(),
							 "Executable " + cfg->mExecCmdLine + " does not exist.\n\n Would you like to build it?",
							 "Error",
							 wxYES_NO );

		if ( dlg.ShowModal() == wxID_YES )
		
			OnBuild( event );
	}
	else
		wxPMBootSvc().SetStatusText( "Executing '" + cfg->mExecCmdLine + "'" );
}

void wxPMBuildService::OnSwitchToOutputWindow( wxCommandEvent& event )
{
	
}

void wxPMBuildService::OnToggleOutputWindow( wxCommandEvent& event )
{
	if ( this->IsShown() )
		wxPMBootSvc().HidePane( this );
	else
		wxPMBootSvc().ShowPane( this );
}

void wxPMBuildService::OnNextMessage( wxCommandEvent& event )
{
	OutputWindowInfo* info = FindOutputWindow( BUILD_WND_NAME );
	if ( !info || !info->mMessages.Count() )
		return;

	TPosition pos = info->mpView->GetModel()->GetCursor();

	if ( pos.mRow > info->mMessages.Count() )
		pos.mRow = info->mMessages.Count()-1;

	OutputMsgInfo* msg = info->mMessages[ pos.mRow ];
	OutputMsgInfo* err = FindError( *info, pos.mRow+1, info->mMessages.Count(), FALSE );

	if ( !err )
		err = FindError( *info, 0, pos.mRow, FALSE );

	if ( err )
	{
		int idx = info->mMessages.Index( err );
		info->mpView->GetModel()->SetCursor( TPosition( idx, 0 ) );

		GotoReferedDocument( *err );
	}
}

void wxPMBuildService::OnPreviousMessage( wxCommandEvent& event )
{
	OutputWindowInfo* info = FindOutputWindow( BUILD_WND_NAME );
	if ( !info || !info->mMessages.Count() )
		return;

	TPosition pos = info->mpView->GetModel()->GetCursor();

	if ( pos.mRow > info->mMessages.Count() )
		pos.mRow = info->mMessages.Count()-1;

	OutputMsgInfo* msg = info->mMessages[ pos.mRow ];
	OutputMsgInfo* err = FindError( *info, pos.mRow-1, -1, TRUE );

	if ( !err )
		err = FindError( *info, info->mMessages.Count()-1, pos.mRow, TRUE );

	if ( err )
	{
		int idx = info->mMessages.Index( err );
		info->mpView->GetModel()->SetCursor( TPosition( idx, 0 ) );

		GotoReferedDocument( *err );
	}
}

void wxPMBuildService::UpdateFiles()
{
	PMProjectConfigSet* set = 
		DoGetConfigForProject( *wxPMBootSvc().GetActiveProject() );

	DoUpdateOutputFiles( set, TRUE );
	DoUpdateOutputFiles( set, FALSE );
}

void wxPMBuildService::DetectMessageInfo( wxString line, OutputMsgInfo& info )
{
	// FIXME:: this whole detection is little too slow

	info.mText = line;
	info.mType = OutputMsgInfo::MSG_NORMAL;

	if ( !mpBuildingCfg )
		return;

	line += "\n";
	int len = line.Length();
	bool finished = false;

    int i;
	for ( i = 0; i != 2; ++i )
	{
		StringListListT* slst = ( i == 0 ) ? &mpBuildingCfg->mpTempl->mErrorFilters
									       : &mpBuildingCfg->mpTempl->mWarningFilters;

		size_t n;
		for ( n = 0; n != slst->Count(); ++n )
		{
			wxStringList* lst = (*slst)[n];

			info.mFile = "";
			info.mLine = -1;

			int cur = 0;

            size_t k;
			for ( k = 0; k != lst->GetCount() && cur < len; ++k )
			{
				wxString expect = (*lst)[k];

				if ( expect.Length() > 2 &&
					 expect[0u] == '_' &&
					 expect[1u] == '_' )
				{
					// extracting not-fixed value string until
					// the occurance of next fixed-value string is encountered

					wxString value;
					wxString next; 
					if ( k < lst->GetCount() )
						next = (*lst)[k+1];

					if ( next.Length() == 0 )
					{
						value = line.Mid( cur );
						cur = len;
					}
					else
					{
						int start = cur;

						while( cur < len )
						{
							if ( line[(unsigned int)cur] == next[0u] )
							{
								int j = 1, pos = cur+1;
								while( j < (int)next.Length() && 
									   pos < len &&
									   line[(unsigned int)pos] == next[(unsigned int)j] )
								{
									++pos; 
									++j;
								}

								if ( j == (int)next.Length() )
								{
									value = line.Mid( start, cur - start );
									cur = pos; 
									++k;
									break; // to outter loop
								}
								else
									++cur; // hasn't matched completely
							}
							else
								++cur;
						}
					}

					if ( value.Length() ) // value found
					{		 
						if ( expect == "__FILE__" )
							info.mFile = value;
						if ( expect == "__LINE__" )
							info.mLine = ::atoi( value );
					}
				}
				else
					// otherwise, just scan for the 'next' as fixed-value 
					// literal which is expected
				{
					while( cur < len )
					{
						if ( line[(unsigned int)cur] == expect[0u] )
						{
							int j = 1, pos = cur+1;
							while( j < (int)expect.Length() && 
								   pos < len &&
								   line[(unsigned int)pos] == expect[(unsigned int)pos] )
							{
								++pos; 
								++j;
							}

							if ( j == (int)expect.Length() )
							{
								cur = pos;
								break;
							}
						}
						else
							++cur;
					}
				}

			}  // end of for(...)

			// finish detection if "expectation-squence" was completed

			if ( info.mFile.Length() &&
				 info.mLine != -1 )
			{
				if ( i == 0 )
					info.mType = OutputMsgInfo::MSG_ERROR;
				else
					info.mType = OutputMsgInfo::MSG_WARNING;

				finished = true;
				break;
			}

		} // end of for( ..n != slst->Count()...)

		if ( finished )
			break;

	} // end of for( 0..1 )
}

void wxPMBuildService::OnConfigure( wxCommandEvent& event )
{
	// Do this with ASSERTs because I feel that the button should be
	// Disabled to prevent the assert to occur.
	wxASSERT( &wxPMBootSvc() != NULL );					// DBG::Ugly code... 
			// Can we make IsInit or IsEmpty function to do it more elegant?
	
    // Give the user information if we can't continue with his request.
    if (!wxPMBootSvc().GetWorkplace())
    {
        wxMessageBox(wxT("No Workspace loaded, unable to continue."), 
                     wxT("Config error."),
                     wxOK | wxCENTRE | wxICON_EXCLAMATION,
                     wxPMBootSvc().GetMainWindow());
        
        return; // Don't continue!
    }

	PMProjectConfigSet* cfgSet = 
		DoGetConfigForProject( *wxPMBootSvc().GetActiveProject() );

	BuildConfigurationsDialog dlg( &cfgSet->GetConfigs(), &mTemplates, this );

	wxTheXmlResource->
		LoadDialog( &dlg, wxPMBootSvc().GetMainWindow(), wxT("BuildSvc_ConfigDialog") );

	dlg.Layout();
	dlg.ShowModal();
}

// notification commands received from build-controller thread

void wxPMBuildService::OnBuildStarted( wxCommandEvent& event )
{
}

void wxPMBuildService::OnBuildMessageListUpdated( wxCommandEvent& event )
{
	mpBuildProcess = NULL;
}

void wxPMBuildService::OnBuildFinished( wxCommandEvent& event )
{
}

void wxPMBuildService::OnProcessTerminated( wxProcessEvent& event )
{
	if ( mpBuildProcess )
	{
		mTimer.Stop();

		SuckBuildOutput( FALSE );
		mpBuildProcess = NULL;
		mpBuildingCfg = NULL;
		int errs = 0, wngs = 0;
		
		// calc statistics

		OutputWindowInfo* info = FindOutputWindow( BUILD_WND_NAME );

        size_t i;
		for ( i = 0; i != info->mMessages.Count(); ++i )
		{
			OutputMsgInfo* msg = info->mMessages[i];

			if ( msg->mType == OutputMsgInfo::MSG_WARNING )
				++wngs;
			else
			if ( msg->mType == OutputMsgInfo::MSG_ERROR )
				++errs;
		}

		wxString s;
		s.Printf( "* Finished - %d error(s), %d warning(s) *", errs, wngs );
		AddMessageLine( "", BUILD_WND_NAME, FALSE );
		AddMessageLine( s, BUILD_WND_NAME, FALSE );

		if ( mNeedToExecute && errs == 0 )
		{
			wxCommandEvent evt;
			OnExecute( evt );
		}
	}
}

void wxPMBuildService::OnBuildTimer( wxTimerEvent& event )
{
	SuckBuildOutput( TRUE );
}

	// implementation of wxTextEditorKeyInterceptor interface 

bool wxPMBuildService::OnEditorKeyEvent( wxKeyEvent& key, bool isChar )
	{ return FALSE; }

bool wxPMBuildService::OnDblClick( wxMouseEvent& event )
{
	OutputWindowInfo* info = FindOutputWindow( BUILD_WND_NAME );

	if ( !info )
		return FALSE;

	TPosition pos = info->mpView->GetModel()->GetCursor();
	OutputMsgInfo* msg = info->mMessages[ pos.mRow ];

	if ( msg->mType == OutputMsgInfo::MSG_WARNING ||
		 msg->mType == OutputMsgInfo::MSG_ERROR )
	
		GotoReferedDocument( *msg );

	// TBD::
	return TRUE;
}

// management of output-windows

void wxPMBuildService::ClearOutputWindow( const wxString& name )
{
	OutputWindowInfo* ownd = FindOutputWindow( name );
	if ( !ownd )
		return;

	WX_CLEAR_ARRAY( ownd->mMessages );

	wxTextEditorModel* model = ownd->mpView->GetModel();
	model->DeleteAllText();

	ownd->mpView->Refresh();
	// TBD::
}

void wxPMBuildService::AddMessageLine( const wxString& line, 
									   const wxString& outWndName,
									   bool detect )
{
	OutputWindowInfo* ownd = FindOutputWindow( outWndName );
	if ( !ownd )
		return;

	OutputMsgInfo* msg = new OutputMsgInfo();

	if ( detect )

		DetectMessageInfo( line, *msg );
	else
	{
		msg->mText = line;
		msg->mLine = -1;
		msg->mType = OutputMsgInfo::MSG_NORMAL;
	}

	// TBD::...
	wxTextEditorModel* model = ownd->mpView->GetModel();
	model->AppendText( msg->mText + "\n" );

	if ( msg->mType == OutputMsgInfo::MSG_ERROR ||
		 msg->mType == OutputMsgInfo::MSG_WARNING )
	{
		int n = model->GetTotalRowCount() - 2;
		model->AddPin( new TPinBase( msg->mType == OutputMsgInfo::MSG_ERROR ? ERRROR_PIN_TC
																		: WARINING_PIN_TC,
									 n ) );
	}

	model->NotifyAllViews();

	ownd->mMessages.Add( msg );

	int last = model->GetTotalRowCount();
	if ( last > 0 )
		model->SetCursor( TPosition( last, 0 ) );
	//ownd->mpView->Refresh();
}

OutputWindowInfo* wxPMBuildService::AddOutputWindow( const wxString& name, const wxString& displayName )
{
	if ( FindOutputWindow( name ) )
		return FALSE;

	OutputWindowInfo* info = new OutputWindowInfo();

	info->mName = name;
	info->mDisplayName = displayName;

	mOutputWindows.Add( info );

	// TBD::...
	wxTextEditorView* view = new wxTextEditorView( mpPageCtrl, -1, NULL, 0 );
	view->SetControlWindow( mpPageCtrl );
	info->mpView = view;
	view->AddPinPainter( new ErrorPinPainter() );
	view->AddPinPainter( new WarningPainter() );
	wxTextEditorModel* model = view->GetModel();
	view->SetLineTrackingMode( TRUE );
	view->EnableCursor( TRUE );
	model->SetReadOnly( TRUE );
	//model->AppendText( "Hello\nWorld" );
	//model->AddPin( new TPinBase( ERRROR_PIN_TC, 0 ) );
	//model->AddPin( new TPinBase( WARINING_PIN_TC, 1 ) );
	model->NotifyAllViews();

	mpPageCtrl->AddTab( info->mpView, displayName, -1, (wxBitmap*)NULL );

	return info;
}

bool wxPMBuildService::DeleteOutputWindow( const wxString& name )
{
	OutputWindowInfo* ownd = FindOutputWindow( name );

	if ( !ownd )
		return FALSE;

	// TBD::...

	wxDELETE(ownd);
	mOutputWindows.Remove( ownd );

	return TRUE;
}

OutputWindowInfo* wxPMBuildService::FindOutputWindow( const wxString& outWndName )
{
    size_t i;
	for ( i = 0; i != mOutputWindows.Count(); ++i )
    {
		if ( mOutputWindows[i]->mName == outWndName )
			return mOutputWindows[i];
    }
	return NULL;
}

/*** protected methods ***/

void wxPMBuildService::SuckBuildOutput( bool fromTiemr )
{
	if ( !mpBuildProcess )
		return;

	wxInputStream* istm = mpBuildProcess->GetInputStream();
	wxInputStream* estm = mpBuildProcess->GetErrorStream();

    int i;
	for ( i =0; i != 2; ++i )
	{
		wxInputStream* stm = i == 1 ? istm : estm;

		while ( stm && !stm->Eof() )
		{
			wxTextInputStream ts( *stm );

			wxString line = ts.ReadLine();

			if ( line.Length() )
				AddMessageLine( line, BUILD_WND_NAME, TRUE );
		}
	}

	// FIXME:: FOR NOW::
	wxView* view = wxPMBootSvc().GetDocManager()->GetCurrentView();
	if ( view )
		view->GetFrame()->SetFocus();
}

void wxPMBuildService::DoExecute( bool build )
{
	if ( mpBuildProcess )
	{
		wxMessageBox( "Building is already in progress..." );
		return;
	}

	// Do this with ASSERTs because I feel that the button should be
	// Disabled to prevent one of the 2 asserts to occur.
	wxASSERT( &wxPMBootSvc() != NULL );					// DBG::Ugly code... 
			// Can we make IsInit or IsEmpty function to do it more elegant?
	wxASSERT( wxPMBootSvc().GetDocManager() != NULL );	// DBG::
	//wxASSERT( wxPMBootSvc().GetWorkplace() != NULL );	// DBG::

    // Give the user information if we can't continue with his request.
    if (!wxPMBootSvc().GetWorkplace())
    {
        wxMessageBox(wxT("No Workspace loaded, unable to continue."), 
                     wxT("Config error."),
                     wxOK | wxCENTRE | wxICON_EXCLAMATION,
                     wxPMBootSvc().GetMainWindow());
        
        return; // Don't continue!
    }

	// save all opened documents before building (FIXME:: crude?)
	if ( build )
	{
		wxPMBootSvc().GetDocManager()->SaveDocument(
				wxPMBootSvc().GetActiveProject() );
	}

	wxPMProject* prj = wxPMBootSvc().GetActiveProject();

	PMProjectConfigSet* set = 
		DoGetConfigForProject( *prj );

	BuildConfiguration* cfg = set->GetActiveConfig();
	mpBuildingCfg = cfg;

    // Give the user information if we can't continue with his request.
	if ( !cfg )
	{
        wxMessageBox(wxT("No active build configuration is set up\nUse \"Build | Configurations\" menu to create one"),
                     wxT("Config error."),
                     wxOK | wxCENTRE | wxICON_EXCLAMATION,
                     wxPMBootSvc().GetMainWindow()
                    );

		return; // Don't continue!
	}

	ReduceBuildTimeVars( *cfg, *prj );
	ClearOutputWindow( BUILD_WND_NAME );
	OutputWindowInfo* ownd = FindOutputWindow( BUILD_WND_NAME );
	//ownd->mpView->SetFocus();

	wxString cmd = build ? cfg->mBuildCmdLine : cfg->mCleanCmdLine;

	//AddMessageLine( "D:\\projects\\wxWorkshop\\ws\\buildsvc.cpp(1193) : error C2001: newline in constant",
	//				  BUILD_WND_NAME, TRUE );

	AddMessageLine( "--------------------Configuration: " + cfg->mName + " --------------------", BUILD_WND_NAME, FALSE );
	AddMessageLine( cmd , BUILD_WND_NAME, FALSE );

	wxString savedCwd = wxGetCwd();
	::wxSetWorkingDirectory( prj->GetFileInfo()->GetPathOnly() );

	mpBuildProcess = new wxProcess( this, XMLID( "on_process_terminated" ) );
	mpBuildProcess->Redirect();

	long ret = ::wxExecute( cmd, FALSE, mpBuildProcess );

	mTimer.Start( 100 );

	::wxSetWorkingDirectory( savedCwd );
}

/*** protected methods ***/

void wxPMBuildService::GetStringListVar( Record& fromRec, 
										 const wxString& name, 
										 wxStringList& toLst )
{
	Record& r = fromRec.GetRecordByName( name );

    size_t i;
	for ( i = 0; i != r.GetFieldCount(); ++i )
    {
		toLst.Add( r.GetValueByIndex( i ) );
    }
}

PMProjectConfigSet* wxPMBuildService::DoGetConfigForProject( wxPMDocument& project )
{
	PMProjectConfigSet* set = (PMProjectConfigSet*)GetAssociatedData( project );

	if ( !set )
	{
		set = new PMProjectConfigSet();
		AssociateData( project, set );
	}

	return set;
}

bool wxPMBuildService::GenerateOutputFile( BuildConfiguration& cfg, wxPMProject& project )
{
	ScriptTemplate::SetVarPrefixChar( '%' );

	ReduceBuildTimeVars( cfg, project );

	if ( cfg.mOutputFile.Length() == 0 )
		return TRUE;

	wxString content = cfg.mpTempl->mContent.SubstitueVariables();

	ScriptTemplate::SetVarPrefixChar();

	wxString savedCwd = wxGetCwd();

	::wxSetWorkingDirectory( project.GetFileInfo()->GetPathOnly() );

	wxFileOutputStream fstm( cfg.mOutputFile );
	wxTextOutputStream tstm( fstm );

	tstm << content;

	::wxSetWorkingDirectory( savedCwd );

	return TRUE;
}

void wxPMBuildService::ReduceBuildTimeVars( BuildConfiguration& cfg, wxPMProject& project )
{
	ScriptTemplate::SetVarPrefixChar( '%' );

	ClearTmpStrings();

	ScriptTemplate* templ = GetProcessorTemplate();

	templ->SetVar( "PROJECT_NAME", MakeTmpString( project.GetTitle() ) );
	templ->SetVar( "DATE",		   MakeTmpString( wxDateTime::Now().Format() ) );
	templ->SetVar( "TEMPLATE",	   MakeTmpString( "\"" + cfg.mpTempl->mName + "\"" ) );

    size_t i;
	for ( i = 0; i != cfg.mFields.GetCount(); ++i )
    {
		templ->SetVar( cfg.mFields[i], MakeTmpString( cfg.mValues[i] ) );
    }

	cfg.mOutputFile = cfg.mpTempl->mOutputFile.SubstitueVariables();

	templ->SetVar( "OUTPUT_FILE", MakeTmpString( cfg.mOutputFile ) );

	cfg.mBuildCmdLine = cfg.mpTempl->mBuildCmdLine.SubstitueVariables();

	templ->SetVar( "BUILD_CMDLINE", MakeTmpString( cfg.mBuildCmdLine ) );

	cfg.mCleanCmdLine = cfg.mpTempl->mCleanCmdLine.SubstitueVariables();

	templ->SetVar( "CLEAN_CMDLINE", MakeTmpString( cfg.mCleanCmdLine ) );

	cfg.mExecCmdLine = cfg.mpTempl->mExecCmdLine.SubstitueVariables();

	templ->SetVar( "EXEC_CMDLINE", MakeTmpString( cfg.mExecCmdLine ) );

	ScriptTemplate::SetVarPrefixChar();
}

bool wxPMBuildService::DoUpdateOutputFiles( PMProjectConfigSet* set, bool clean )
{
	// TBD:: error handling 

	BuildConfigurationListT& lst = set->GetConfigs();

	wxPMProject* prj = wxPMBootSvc().GetActiveProject();

    size_t i;
	for ( i = 0; i != lst.Count(); ++i )
	{
		BuildConfiguration& cfg = *lst[i];

		ReduceBuildTimeVars( cfg, *prj );

		if ( cfg.mOutputFile.Length() )
		{
			if ( wxFileExists( cfg.mOutputFile ) && clean )
			
				wxRemoveFile( cfg.mOutputFile );
			else
			if ( !wxFileExists( cfg.mOutputFile ) && !clean )
				
				GenerateOutputFile( cfg, *prj );
		}
	}

	return TRUE;
}

OutputMsgInfo* wxPMBuildService::FindError( OutputWindowInfo& ownd, int from, int till, bool backwards  )
{
	if ( !backwards )
	{
		if ( (unsigned int)from >= ownd.mMessages.Count() )
			return NULL;

		int i;
		for ( i = from; i != till; ++i )
		{
			OutputMsgInfo* msg = ownd.mMessages[i];

			if ( msg->mType == OutputMsgInfo::MSG_ERROR )
				return msg;
		}
	}
	else
	{
		int i;
		for ( i = from; i != till; --i )
		{
			OutputMsgInfo* msg = ownd.mMessages[i];

			if ( msg->mType == OutputMsgInfo::MSG_ERROR )
				return msg;
		}
	}

	return NULL;
}

bool wxPMBuildService::GotoReferedDocument( OutputMsgInfo& ofMsg, bool promptFailure  )
{
	wxString file = wxPMToUnixPath( ofMsg.mFile );

	if ( !wxIsAbsolutePath( file ) )
	{
		wxString prjDir = wxPMBootSvc().GetActiveProject()->GetFileInfo()->GetPathOnly();

		file = wxPMGetAbsolutePath( prjDir, file );
	}

	wxPMDocument* doc = wxPMBootSvc().GetDocManager()
		->OpenDocument( wxPMFileInfo( file ), NULL );

	if ( !doc )
	{
		if ( promptFailure )
			wxMessageBox( "Refered file " + ofMsg.mFile + " is not included into\n"
						  "any of the projects of the current workplace" );

		return TRUE;
	}

	I_EditableText* itext = (I_EditableText*)doc->GetInterface( "I_EditableText" );

	if ( itext )
	{
		itext->SetCursorPosition( ofMsg.mLine-1, 0 );
		return TRUE;
	}
	else
		return FALSE;
}

static bool IS_BLANK( char x ) 
{
	return ( x == ' ' || x == '\t' || x == '\n' || x == '\r' );
}

static wxString get_next_word( const char*& cur )
{
	while( *cur && IS_BLANK( *cur ) )
		++cur;

	const char* start = cur;

	if ( *cur == '\'' )
	{
		++cur;
		start = cur;
		while( *cur && *cur != '\'' )
			++cur;

		++cur; // skip '''

		return wxString( start, size_t( cur - start - 1 ) );
	}

	while( *cur && !IS_BLANK( *cur ) )
		++cur;

	return wxString( start, size_t( cur - start ) );
}

void wxPMBuildService::ParseFilters( const wxString& content, StringListListT& toLst )
{
	ScriptTemplate templ;
	templ.SetScript( content );
	templ.SetParentTemplate( mpTopTempl );
	wxString s = templ.SubstitueVariables();
	const char* cur = s.c_str();
	wxStringList* curLst = new wxStringList();

    wxString word;
	for (;;)
	{
		word = get_next_word( cur );

		if ( word == "|" )
		{
			toLst.Add( curLst );
			curLst = new wxStringList();
		}
		else
		if ( !word.Length() )
			break;
		else
			curLst->Add( word );
	}

	if ( curLst->GetCount() )
		toLst.Add( curLst );
	else
		delete curLst;
}

/***** Implementation for class BuildControllerThread *****/

class BuildProcess : public wxProcess
{
public:
	BuildProcess( BuildControllerThread* cth )
		: 
		  mpCth( cth ) 
	{}

	virtual void OnTerminate(int pid, int status)
	{
		//mpCth->OnTerminated();

		//delete this;
	}

protected:
	BuildControllerThread* mpCth;
};

/*
BuildControllerThread::BuildControllerThread( wxPMBuildService* buildSvc, 
											  BuildConfiguration* cfg )

	: mpBuildSvc( buildSvc ),
	  mpCfg( cfg )
{
}

void* BuildControllerThread::Entry()
{
	mIsTerminated = FALSE;

	return NULL;
}

void BuildControllerThread::OnTerminated()
{
	mIsTerminated = TRUE;
}
*/
