/////////////////////////////////////////////////////////////////////////////
// Name:        No names yet.
// Purpose:     Contrib. demo
// Author:      Aleksandras Gluchovas
// Modified by:
// Created:     2001/03/28
// RCS-ID:      $Id: classfinder_svc.cpp,v 1.18 2001/11/17 11:25:01 hansvl Exp $
// Copyright:   (c) Aleksandras Gluchovas (alex@soften.ktu.lt)
// Licence:   	GPL (version 2.0 or later)
/////////////////////////////////////////////////////////////////////////////

#ifdef __GNUG__
    #pragma implementation "classfinder_svc.h"
#endif

// For compilers that support precompilation, includes "wx/wx.h".
#include "wx/wxprec.h"

#ifdef __BORLANDC__
    #pragma hdrstop
#endif

#ifndef WX_PRECOMP
    #include "wx/wx.h"
#endif

#include "classfinder_svc.h"
#include "mimetypes.h"

#include "wx/pmf/inf/textediting.h"
#include "parserbase.h"
#include "cjparser.h"

#include <string.h>
#include "strlist.h"

#include "wx/tokenzr.h"

/***** Implementation for class wxPMClassFinderService *****/

static inline bool is_blank( char ch )
{
	return ch == ' '  || ch == '\t' || 
		   ch == '\n' || ch == '\r';
}

static inline bool is_alpha_num( char ch )
{
	return (ch >= 'a' && ch <= 'z') ||
		   (ch >= 'A' && ch <= 'Z') ||
		   (ch >= '0' && ch <= '9') ||
		   (ch == '_');
}

static inline void skip_blanks( char*& cur, char* end )
{
	while( cur < end && is_blank( *cur ) )
		++cur;
}

struct spArgInfo
{
	char* type;
	char* name;
	char* defValue;
};

struct spMtdInfo
{
	char* name;
	char* type;
	spArgInfo* args;
};

class spWxDefineArrayMacroPlugin : public SourceParserPlugin
{
	// overriden methods

	virtual SourceParserPlugin* CanAcceptMacro( char* cur, char* end, spContext* outtterCtx )
	{
		if ( memcmp( cur, "WX_DEFINE_ARRAY", 15 ) == 0 )
			return this;
		else
			return SourceParserPlugin::CanAcceptMacro( cur, end, outtterCtx );
	}

	virtual void ParseMacro( char* start, char*& cur, char* end, spContext* outtterCtx )
	{
		size_t srcOfs = size_t( cur - start );

		while( cur < end && *cur != '(' )
			++cur;

		++cur; // skip '('

		skip_blanks( cur, end );
		char* typeStart = cur;

		while( cur < end && !is_blank( *cur ) && *cur != ',' )
			++cur;

		wxString type( typeStart, size_t( cur - typeStart ) );

		while( cur < end && *cur != ',' )
			++cur;

		++cur; // skip ','

		skip_blanks( cur, end );
		char* nameStart = cur;

		while( cur < end && is_alpha_num( *cur ) )
			++cur;

		wxString name( nameStart, size_t( cur - nameStart ) );

		// "virtualize" new array type

		spArgInfo argInfos[] = {	/* 0  */ { "_T","item","" }, {"","",""},// Add
									/* 2  */ { "size_t","count","" }, {"","",""}, // Alloc
									/* 4  */ { "","","" }, // Clear
									/* 5  */ { "","","" }, // Count
									/* 6  */ { "size_t","index","" }, { "","","" }, // Detach
									/* 8  */ { "","","" }, // Empty
									/* 9  */ { "","","" }, // GetCount
									/* 10 */ { "_T&","item","" }, { "bool","searchFromEnd","FALSE" }, {"","",""},// Index
									/* 13 */ { "_T&","item","" }, {"","",""}, // Index
									/* 15 */ { "_T","item","" },  { "size_t","n","" }, {"","",""},// Insert
									/* 18 */ { "_T*","item","" }, { "size_t","n","" }, {"","",""},// Insert
									/* 21 */ { "_T&","item","" }, { "size_t","n","" }, {"","",""},// Insert
									/* 24 */ { "","","" }, // EsEmpty
									/* 25 */ { "size_t","index","" }, {"","",""}, // Item
									/* 27 */ { "","","" }, // Last
									/* 28 */ { "_T","item","" }, {"","",""}, // Remove
									/* 30 */ { "size_t","index","" }, {"","",""}, // RemoveAt
									/* 32 */ { "","","" }, // Shrink
									/* 33 */ { "CMPFUNCT<_T>","compareFunction","" }, {"","",""}, // Sort
									/* 36 */ { "size_t","index","" }, {"","",""}, // operator[]
								};

		spMtdInfo mtds[] = {	{ "Add",		"void", argInfos+0 },
								{ "Alloc",		"void", argInfos+2},
								{ "Clear",		"void", argInfos+4},
								{ "Count",		"size_t", argInfos+5},
								{ "Detach",		"_T*",  argInfos+6 },
								{ "Empty",		"void", argInfos+8 },
								{ "GetCount",	"size_t", argInfos+9 },
								{ "Index",		"int",  argInfos+10 },
								{ "Index",		"int",  argInfos+13 },
								{ "Insert",		"void", argInfos+15 },
								{ "Insert",		"void", argInfos+18 },
								{ "Insert",		"void", argInfos+21 },
								{ "IsEmpty",	"bool", argInfos+24 },
								{ "Item",		"_T&",  argInfos+25 },
								{ "Last",		"_T&",  argInfos+27 },
								{ "Remove",		"bool", argInfos+28 },
								{ "RemoveAt",	"bool", argInfos+30 },
								{ "Shrink",		"void", argInfos+32 },
								{ "Sort",		"void", argInfos+33 },
								{ "operator[]", "_T&",  argInfos+36 },
								{ "","",NULL },
							};

		spClass* cl = new spClass();
		cl->mName = name;
		cl->mSrcOffset = srcOfs;
		cl->mContextLength = 1;

		for( int i = 0;; ++i )
		{
			spMtdInfo& inf = mtds[i];

			if ( inf.name[0] == '\0' )
				break;

			spOperation* op = new spOperation();
			op->mRetType = inf.type;
			op->mName = inf.name;
			op->mRetType.Replace( "_T", type, FALSE );
			op->mSrcOffset = srcOfs;
			op->mContextLength = 1;

			int k = 0;
			for(;;)
			{
				spArgInfo& ainf = inf.args[k];

				if ( ainf.type[0] == '\0' )
					break;

				spParameter* par = new spParameter();
				par->mName = ainf.name;
				par->mType = ainf.type;
				par->mInitVal = ainf.defValue;

				par->mType.Replace( "_T", type, FALSE );

				op->AddParameter( par );
				++k;
			}

			op->mFlags = SP_FLAG_VIS_PUBLIC;

			if ( inf.name[0] == 'o' ) // "operator[]"
				op->mFlags |= SP_FLAG_IS_OPERATOR | SP_FLAG_IS_ARRAY;

			cl->AddMember( op );
		}

		outtterCtx->AddMember( cl );
	}
};

/***** Implementation for class wxPMClassFinderService *****/

IMPLEMENT_DYNAMIC_CLASS( wxPMClassFinderService, wxPMSimpleService )

BEGIN_EVENT_TABLE( wxPMClassFinderService, wxPMSimpleService )

	EVT_MENU( ID_DUMP_CLASS_INFO,   wxPMClassFinderService::OnDumpClassInfo )
	EVT_MENU( ID_TEST_EXPR_PARSING, wxPMClassFinderService::OnTestExprParsing )

END_EVENT_TABLE()

wxPMClassFinderService::wxPMClassFinderService()

	: mpParser( NULL ),
	  mpWxMacroParser( NULL ),
	  mpTmpBuf( NULL ),
	  mDocHash( wxKEY_INTEGER )
{}

wxPMClassFinderService::~wxPMClassFinderService()
{
	wxDELETEA(mpTmpBuf);
	wxDELETE(mpParser);
	// mpWxMacroParser is deleted by the parser

	mDocHash.BeginFind();

    wxNode* node;
	while( (node = mDocHash.Next()) != NULL )
    {
        delete (spContext*)node->GetData();
    }
}

bool wxPMClassFinderService::Start()
{
	mpIndexSvc = (wxPMIndexingService*)wxPMBootSvc().
					FindUniqueServiceInstance( "wxPMIndexingService" );

	if ( !mpIndexSvc )
		return FALSE;

	mpIndexSvc->RegisterIndexHandler( (I_IndexHandler*)this );

	wxPMBootSvc().RegisterUniqueServiceInstance( this, "wxPMClassFinderService" );

	return TRUE; // started ok
}

bool wxPMClassFinderService::Stop()
{
	return TRUE;
}

static wxCommandInfo _gClassFinderCommands[] = {

	wxCommandInfo( -1, "SEPARATOR" ),

	wxCommandInfo( wxPMClassFinderService::ID_DUMP_CLASS_INFO, "&Dump all class information", "", "" ),
	wxCommandInfo( wxPMClassFinderService::ID_TEST_EXPR_PARSING, "&Test class information", "", "" ),

	wxCommandInfo( 0, "" ) // terminator
};

static wxPMMergableMenuChain _gClassFinderChains[] = {

	wxPMMergableMenuChain( "/&Tools/", _gClassFinderCommands ),
};

wxPMMergableMenuChain* wxPMClassFinderService::GetMergableMenuChain()
{
	return _gClassFinderChains;
}

/*
 * some higher-level algorithms
 */

// global helpers

static inline bool IsAlpha( char ch )
{
	return ( ch >= 'A' && ch <= 'Z' ) ||
	       ( ch >= 'a' && ch <= 'z' ) ||
	       ch == '_';
}

static inline bool IsAlphaNum( char ch )
{
	return ( ch >= 'A' && ch <= 'Z' ) ||
	       ( ch >= 'a' && ch <= 'z' ) ||
	       ( ch >= '0' && ch <= '9' ) ||
	       ch == '_';
}

static bool IsName( const wxString& token )
{
	if ( !IsAlpha( token[0] ) )
		return FALSE;
		
	for( size_t i = 1; i != token.Length(); ++i )
	{
		char ch = token[i];
		
		if ( !IsAlphaNum( ch ) )
			return FALSE;
	}
	
	return TRUE;	
}

static int FindLeftBrace( StrListT& lst, int pos )
{
	int depth = 1; --pos;
	
	while( depth > 0 && pos >= 0 )
	{
		if ( lst[pos] == ")" )
			++depth;
		else
		if ( lst[pos] == "(" )
			--depth;

		if ( depth == 0 )
			break;

		--pos;
	}
	
	if ( depth != 0 )
		return -1;
		
	return pos;
}

static int FindLeftBracket( StrListT& lst, int pos )
{
	int depth = 1; --pos;
	
	while( depth > 0 && pos >= 0 )
	{
		if ( lst[pos] == "]" )
			++depth;
		else
		if ( lst[pos] == "[" )
			--depth;

		if ( depth == 0 )
			break;

		--pos;
	}
	
	if ( depth != 0 )
		return -1;
		
	return pos;
}

static bool AnalyzeExpr( StrListT& lst, StrListT& outLst )
{
	size_t len = lst.GetCount();
	int cur = len - 1;

	if ( lst[cur] == "(" || lst[cur] == "[" || lst[cur] == "." )
		--cur;

	wxString s;
	
	for(;cur > 0;)
	{
		s=lst[cur]; // DBG::

		bool nameFound = FALSE;
		if ( IsName( lst[cur] ) )
		{
			outLst.Insert( lst[cur], 0 );
			nameFound = TRUE;
			--cur;
		}
//		else
//			break;
			
		if( cur < 0 )
			break;
			
//		if ( lst[cur] != "." ) // !
//			break;

		if ( lst[cur] == "." )
			--cur;

		if ( cur < 0 )
			break;
			
		// expecting preceding braces now

		s=lst[cur]; // DBG::

		if ( lst[cur]==";" ||
			 lst[cur]=="}" ||
			 lst[cur]=="{" )
			 break;
		
		if ( lst[cur] == "]" )
		{
			int lbracket = FindLeftBracket( lst, cur );

			if ( lbracket == -1 )
				break;

			cur = lbracket - 1;

			outLst.Insert( "[]", 0 );
			continue;
		}

		if ( lst[cur] != ")" )
		{
			if ( !nameFound )
				break;

			continue;
		}
			
		int lbrace = FindLeftBrace( lst, cur );

		s=lst[cur]; // DBG::
		
		if ( lbrace == -1 )
			break;
			
		if ( lbrace != 0 &&
		     IsName( lst[lbrace-1]  ) )
		{
			cur = lbrace-1;
			s=lst[cur]; // DBG::
			continue;
		}
		
		cur = lbrace;
		
		// expecting left-most cast or a name
		
		while( cur < lst.Count() &&
		       lst[cur] == "(" )
			++cur;

		if ( cur >= lst.Count() )
			return FALSE; // ERROR?

		s=lst[cur]; // DBG::

		if ( !IsName( lst[cur] ) )
			return FALSE; // ERORR?
		
		s=lst[cur+1]; // DBG::

		if ( lst[cur+1] == "(" )
			return FALSE; // ERORR? (of rare +/- kind)
		
		s=lst[cur]; // DBG::

		outLst.Insert( lst[cur], 0 );
		
		break; // finished!
	}
	
	return TRUE;
}

static const char* _gDelLst = " \n\r\t!@#$%^&*(){}[]<>.,?=-+\";'~/\\";
static unsigned int _gDelLstLen = strlen( _gDelLst );

static wxString GetNextToken( char*& cur )
{
	char* start = cur;

	while( *cur )
	{
		char ch = *cur;

		for( size_t i = 0; i != _gDelLstLen; ++i )
		{
			if ( _gDelLst[i] == ch )
			{
				if ( start == cur )
				{
					++cur;
					return wxString(ch);
				}
				else
				{
					*cur = '\0';
					wxString tok = start;
					*cur = ch;
					return tok;
				}
			}
		}

		++cur;
	}

	return start;
}

static wxString GetQuotedToken( char*& cur, char ch )
{
	char* start = cur;
	++cur; // skip leading quote

	while( *cur )
	{
		if ( *cur == ch )
		
			break;

		if ( *cur == '\\' &&
			 ( *(cur+1) == '"' ||
			   *(cur+1) == '\'' ) 
		   )
			cur += 2;
		else
			++cur;
	}

	wxString tok;

	if ( *cur )
	{
		++cur;
		char ch = *cur;
		*cur = '\0';
		tok = start;
		*cur = ch;
	}
	else
		tok = start;

	return tok;
}

static void TokenizeStatement( const wxString& stmt, StrListT& lst )
{
	char* cur = (char*)(const char*)stmt.c_str();

	while ( *cur )
	{
		wxString tk = GetNextToken( cur );

		if ( tk.Length() == 1 )
		{
			char ch = tk[0u];

			if ( ch == ' '  || 
				 ch == '\n' ||
				 ch == '\r' ||
				 ch == '\t' ) 
			
				 continue;

			// (special hack) convert all '->'s to  '.'s

			if ( ch == '-' &&
				 *cur == '>' )
			{
				cur += 1;
				lst.Add( "." );
				continue;
			}

			if ( ch == '\'' || ch == '"' )
			
				tk = GetQuotedToken( cur, ch );
		}

		lst.Add( tk );
	}
}

bool wxPMClassFinderService::ExtractSourceContextInfo( wxPMDocument& fromDoc, 
													   SourceEditingContextInfo& toInfo )
{
	I_ContainingText* inf = (I_ContainingText*)fromDoc.GetInterface( "I_ContainingText" );

	if ( !inf )
		return FALSE;

	int line = 0, col = 0;
	inf->GetCursorPosition( &line, &col );

	int line0 = line - 3;
	if ( line0 < 0 ) line0 = 0;

	wxString text;
	inf->GetTextRange( &text, line0, 0, line, 500 );

	int len = text.Length();
	int i = len - 1;
	// now truncate obtained text to the last word at the
	// cursor position

	while( i >= 0 && text[(unsigned int)i] != '\n' && text[(unsigned int)i] != '\r' )
		--i;
	++i;
	unsigned int start = i;
	i = start + col;

	while( i < len && IsAlphaNum( text[(unsigned int)i] ) )
		++i;

	toInfo.mEditedText = text.Mid( 0, i );

	// now resolve surrounding context

	inf->GetAllText( &text );

	SourceParserBase* parser = GetParser();
	spFile* sfile = parser->Parse( text );
	parser->ResolveLineNumbers( sfile, text );

	spContext* inCtx = parser->FindSurroundingContext( sfile, line, col );
	
	if ( inCtx->GetType() == SP_CTX_OPERATION )
	{
		MParamListT& args = ((spOperation*)inCtx)->GetParams();
		toInfo.mScope = ((spOperation*)inCtx)->mScope;

		// add arguments of the current methods

		size_t j;
		for( j = 0; j != args.Count(); ++j )
		{
			toInfo.mCtxTypes.Add( args[j]->mType );
			toInfo.mCtxIdentifiers.Add( args[j]->mName );
		}

		// add member varaibles if we are in a class method

		if ( toInfo.mScope.Length() )
		{
			spClass* found = FindClass( ::spGetTypeClass( toInfo.mScope ) );

			if ( found )
			{
				toInfo.mCtxTypes.Add( found->mName );
				toInfo.mCtxIdentifiers.Add( "this" );

				MMemberListT lst;
				FindMembers( found, lst, NULL, SP_CTX_ATTRIBUTE, TRUE, NULL, FALSE );
				
				for( j = 0; j != lst.Count(); ++j )
				{
					spAttribute* attr = (spAttribute*)lst[j];

					toInfo.mCtxTypes.Add( attr->mType );
					toInfo.mCtxIdentifiers.Add( attr->mName );
				}

				found->mMembers.Clear();

				WX_CLEAR_ARRAY( lst );
				// FIXME:: leaking lst

				delete found;
			}
		}
	}

	char* from = (char*)(const char*)text.c_str() + inCtx->mSrcOffset;
	parser->AnalyzeFunctionBody( from, from + inCtx->mContextLength, 
								 toInfo.mCtxTypes, toInfo.mCtxIdentifiers  );

	delete sfile;

	return TRUE;
}

spContext* wxPMClassFinderService::DetectLastContext( SourceEditingContextInfo& ctx )
{

	int j; int len = ctx.mEditedText.Length();

	if ( !len )
		return NULL;

//	bool isMethod = ctx.mEditedText[ len ] == '('; ==> HVL::Code is not in use...

	StrListT tokens, identifiers;
	TokenizeStatement( ctx.mEditedText, tokens );
	AnalyzeExpr( tokens, identifiers );

	if ( identifiers.Count() == 0 )
		return NULL;

	// find out the type of left-most identifier

	wxString first = identifiers[0];
	wxString curType = first;

	size_t i = 0;
	for( ; i != ctx.mCtxIdentifiers.GetCount(); ++i )
	{
		wxString r = ctx.mCtxIdentifiers[i];
		wxString p = ctx.mCtxTypes[i];

		if ( ctx.mCtxIdentifiers[i] == first )
		{
			curType = ctx.mCtxTypes[i];
			break;
		}
	}

	// (if the above loop didn't find the match, then
	//  the name of the identifier is assumed as name of the type)

	for( j = 1; // start from the second one
		 j < identifiers.Count(); ++j )
	{
		spClass* curClass = FindClass( spGetTypeClass( curType ) );

		if ( !curClass )
		{
			ctx.mLastUnrssolvedCtx = curType;
			ctx.mLastUnresolvedName = identifiers[j];
			return NULL;
		}

		wxString nextId = identifiers[j];

		// try to find this identifier among members
		// of the class or superclasses of curClass

		if ( nextId == "[]" )
		{
			spOperation* op = FindArrayOperator( curClass );

			if ( !op )
				break;

			curType = ::spGetTypeClass( op->mRetType );
			wxDELETE(curClass);
			continue;
		}
				 
		//spContext* found = FindMember( curClass, nextId, TRUE );

		MMemberListT lst;
		FindMembers( curClass, lst, &nextId, SP_CTX_ATTRIBUTE|SP_CTX_OPERATION, TRUE, NULL, FALSE );

		if ( lst.Count() == 0 )
		{
			ctx.mLastUnrssolvedCtx = ::spGetTypeClass( curType );
			ctx.mLastUnresolvedName = nextId;

			return NULL;
		}

		if ( j == identifiers.Count() - 1 && ctx.mHelpNeeded )
			return lst[0];

		spContext* found = lst[0];

		if ( found->GetType() == SP_CTX_OPERATION )
		
			curType = ::spGetTypeClass( ((spOperation*)found)->mRetType );
		else
		if ( found->GetType() == SP_CTX_ATTRIBUTE )

			curType = ::spGetTypeClass( ((spAttribute*)found)->mType );

		if ( j == identifiers.Count() - 1 )
			break;

		wxDELETE(found);

		// next iteration goes deeper towards right-hand-side of expression
	}

	spContext* found = FindClass( ::spGetTypeClass( curType ) );

	if ( !found )
		ctx.mLastUnrssolvedCtx = curType;

	return found;
}
						
void wxPMClassFinderService::FindMembers( spClass* cl, MMemberListT& toLst, wxString* matchPattern, int filterMask, 
										  bool searchBases, wxStringList* baseClassNames, bool sort )
{
	MMemberListT& lst = cl->GetMembers();
	int ptnLen = matchPattern ? matchPattern->Length() : -1;
	wxString ptnLower;
	if (matchPattern)
		ptnLower = matchPattern->Lower();
	else
		ptnLower = wxEmptyString;

	size_t i = 0;
	for( ; i != lst.Count(); ++i )
	{
		spContext* ctx = lst[i];
		int nameLen = ctx->mName.Length();

		if ( ctx->GetType() & filterMask )
		{
			if ( matchPattern && ptnLen == nameLen &&
				 ptnLower == ctx->mName.Lower() )
			{
				toLst.Add( ctx ); // asked for a single match, we found and return it
				
				if ( ctx->GetType() == SP_CTX_OPERATION )
					((spOperation*)ctx)->mScope = cl->mName; // FOR NOW:: FIXME:: ...

				return;
			}
			else
			if ( !matchPattern )
			{
				if ( ctx->GetType() == SP_CTX_OPERATION )
					((spOperation*)ctx)->mScope = cl->mName; // FOR NOW:: FIXME:: ...

				toLst.Add( ctx );
			}
		}
	}

	if ( baseClassNames )
	
		for( i = 0; i != cl->mSuperClassNames.GetCount(); ++i )
			baseClassNames->Add( cl->mSuperClassNames[i] );

	if ( searchBases )
	{
		cl->mMembers.Clear(); // FIMXE:: leak!

		for( i = 0; i != cl->mSuperClassNames.GetCount(); ++i )
		{
			spClass* base = FindClass( cl->mSuperClassNames[i] );

			if ( base )
			{
				FindMembers( base, toLst, matchPattern, filterMask, searchBases, baseClassNames, FALSE ); // recursion
				base->mMembers.Clear(); // FIMXE:: leak!
				delete base;

				if ( matchPattern && toLst.Count() )
					return;
			}
		}
	}

	if ( sort )
		toLst.Sort( spContextCompareFunction );
}

spClass* wxPMClassFinderService::FindClass( const wxString& name )
{
	ItemTable& it = mpIndexSvc->GetItemTable();

	if ( it.FindByName( name, "", SP_CTX_CLASS, FALSE ) )
		return ReadClass();

	if ( it.FindByName( name, "", SP_CTX_TYPEDEF, FALSE ) )
	{
		spTypeDef* td = ReadTypeDef();

		wxString origType = ::spGetTypeClass( td->mOriginalType );
		delete td;

		return FindClass( origType ); // recursion!
	}
	
	return NULL;
}

spOperation* wxPMClassFinderService::FindArrayOperator( spClass* ofClass )
{
	size_t i = 0;
	for( ; i != ofClass->mMembers.Count(); ++i )
	{
		spContext* ctx = ofClass->mMembers[i];

		if ( ctx->mFlags & ( SP_FLAG_IS_ARRAY | SP_FLAG_IS_OPERATOR ) )
			return (spOperation*)ctx;
	}

	return NULL;
}

spContext* wxPMClassFinderService::FindMember( spClass* cl,
											   const wxString& name,
											   bool searchSuperClasses )
{
	size_t i = 0;
	for( ; i != cl->mMembers.Count(); ++i )
	{
		spContext* ctx = cl->mMembers[i];

		if ( ctx->mName == name )
			return ctx;
	}

	if ( searchSuperClasses )
	{
		for (i = 0; i != cl->mSuperClassNames.GetCount(); ++i)
		{
			spClass* super = FindClass( cl->mSuperClassNames[i] );

			if ( super )
			{
				// recursive call
				spContext* found = FindMember( super, name );

				if ( found )
					return found;
			}
		}
	}

	wxDELETE(cl); // !!!

	return NULL;
}

void wxPMClassFinderService::DumpAllItems( wxTextOutputStream& stm )
{
	ItemTable& it = mpIndexSvc->GetItemTable();

	int types[] = { SP_CTX_CLASS, SP_CTX_OPERATION };

    int i;
	for (i = 0; i != 2; ++i)
	{
		if ( it.FindByType( types[i] ) )
		{
            spContext* ctx = NULL;
			while (it.Ok() && it.GetType() == types[i])
			{
				ctx = ReadContext();

				if ( ctx )
				{
					ctx->Dump( stm, wxEmptyString );
					wxDELETE(ctx);
				}

				stm << "\n";
				it.Next();
			}
		}

		it.End();
	}
}

void wxPMClassFinderService::DumpAllItemsToFile( const wxString& fname )
{
	wxFFileOutputStream fs( fname );
	wxTextOutputStream ts( fs );

  	DumpAllItems( ts );
}

/*** implementation of I_IndexHandler interface ***/

const char* wxPMClassFinderService::GetIndexHandlerName()
{
	return "cpp_class_info_handler";
}

const char* wxPMClassFinderService::GetIndexHandlerDescription()
{
	return "C++ Class-Information index handler";
}

bool wxPMClassFinderService::OnClearIndexItems( int indexId )
{
	ItemTable& it = mpIndexSvc->GetItemTable();

	it.DeleteItemsByType( SP_CTX_CLASS, indexId );
	it.DeleteItemsByType( SP_CTX_OPERATION, indexId );
	it.DeleteItemsByType( SP_CTX_ATTRIBUTE, indexId );
	it.DeleteItemsByType( SP_CTX_TYPEDEF, indexId );

	return TRUE;
}

bool wxPMClassFinderService::OnBeginBuild( int indexId )
{
	if ( !mpTmpBuf )
	{
		mTmpBufLen = 10000;
		mpTmpBuf = new char[mTmpBufLen+1];
	}

	GetParser();

	return TRUE; // nothing
}

bool wxPMClassFinderService::OnIndexFile( const wxPMFileInfo& file, int fileId, int indexId )
{
	wxFile fl( file.GetPath() );
	if ( !fl.IsOpened() )
		return FALSE;

	int len = fl.Length();

	if ( len > mTmpBufLen )
	{
		wxDELETEA(mpTmpBuf);
		mTmpBufLen = len;
		mpTmpBuf = new char[ mTmpBufLen + 1 ];
	}

	char* buf = mpTmpBuf;

	// this must be fast, yeahh...!
	int n = fl.Read( buf, len );

	buf[n] = '\0'; 

	spFile* fileCtx = GetParser()->Parse( buf, buf + n, FALSE );

	if ( !fileCtx )
		return FALSE;

	//mpParser->ResolveLineNumbers( fileCtx, buf, buf + n );

	ItemTable& it = mpIndexSvc->GetItemTable();
	it.SetIndexId( indexId );
	it.SetFileId( fileId );

	WriteFile( fileCtx );

	delete fileCtx;

	return TRUE;
}

void wxPMClassFinderService::OnDumpIndexForFile( wxTextOutputStream& stm, int fileId, int indexId )
{
	spFile* file = LoadInfoForFile( fileId, indexId );

	file->Dump( stm, "\t" );

	delete file;
}

bool wxPMClassFinderService::OnEndBuild( int indexId )
{
	// FIXME:: performance vc. mem usage?
//	wxDELETEA(mpTmpBuf);
//	wxDELETE(mpParser);

	return TRUE; // nothing
}

bool wxPMClassFinderService::DocumentMonitoringNeeded()
{
	return TRUE;
}

bool wxPMClassFinderService::CanAcceptFile( const wxPMFileInfo& file, int indexId )
{
	const char* type = file.GetType().c_str();
// TODO: Need to replace these hard coded checks with the values specified in MIMETYPES.H
	// performance..
	return strcmp( type, wxwMIME_C_HEADER )   == 0 ||
		   strcmp( type, wxwMIME_CPP_SOURCE ) == 0 ||
		   strcmp( type, wxwMIME_C_SOURCE ) == 0;
}

void wxPMClassFinderService::OnDocumentsSwitched( wxPMDocument* fromDoc, wxPMDocument* toDoc, int fileId, int indexId )
{
	spFile* oldInfo = (spFile*)mDocHash.Get( long(fromDoc) );

	if ( !oldInfo )
		oldInfo = LoadInfoForFile( fileId, indexId );

	spFile* newInfo = ParseDocContent( *fromDoc );
	SyncIndexWithContent( oldInfo, newInfo, fileId, indexId );

	// substitue cached info with updated one
	mDocHash.Delete( long(fromDoc) );
	mDocHash.Put( long(fromDoc), newInfo ); 
	delete oldInfo;
}

void wxPMClassFinderService::OnDocumentSaved( wxPMDocument& doc, int fileId, int indexId )
{
	OnDocumentsSwitched( &doc, NULL, fileId, indexId );

	RemoveCachedInfoForDoc( doc );
}

void wxPMClassFinderService::OnDocumentChangesDiscarded( wxPMDocument& doc, int fileId, int indexId )
{
	if ( !mDocHash.Get( long(&doc) ) ) 
		return; // wasn't tracked, thus no changes to discard

	// load and parse file from disk, 
	spFile* fl = GetParser()->ParseFile( doc.GetFileInfo()->GetPath() );

	// index it's content anew

	ItemTable& it = mpIndexSvc->GetItemTable();
	it.DeleteItemsByFile( fileId, indexId );
	it.SetFileId( fileId );
	it.SetIndexId( indexId );

	WriteContext( fl );

	RemoveCachedInfoForDoc( doc );
}

SourceParserBase* wxPMClassFinderService::GetParser()
{
	// FIXME:: multiple languages...

	if ( !mpParser )
	{
		if ( !mpWxMacroParser )
			mpWxMacroParser = new spWxDefineArrayMacroPlugin();

		mpParser = new CJSourceParser();
		mpParser->SetPlugin( mpWxMacroParser );

		// NOTE:: more plugins could be chained using SetNextPlugin(..)
	}

	return mpParser;
}

// command handlers 

void wxPMClassFinderService::OnDumpClassInfo( wxCommandEvent& event )
{  // TODO: Remove hardcoded path
	DumpAllItemsToFile( "d:\\projects\\sample\\dump.txt" );
}

void wxPMClassFinderService::OnTestExprParsing( wxCommandEvent& event )
{
	SourceEditingContextInfo ctx;
	ctx.mEditedText = "{ z.getFirst()->mtd1(";
	ctx.mCtxTypes.Add( "Zero" );
	ctx.mCtxIdentifiers.Add( "z" );

	spContext* found = DetectLastContext( ctx );
	// TBD::
}

/*** protected methods ***/

void wxPMClassFinderService::WriteContext( spContext* ctx, bool update )
{
		switch( ctx->GetType() )
		{
			case SP_CTX_CLASS : 
				WriteClass( (spClass*)ctx, update ); 
				break;

			case SP_CTX_OPERATION : 
				WriteOperation( (spOperation*)ctx, update ); 
				break;

			case SP_CTX_FILE : 
				WriteFile( (spFile*)ctx, update ); 
				break;

			case SP_CTX_TYPEDEF :
				WriteTypeDef( (spTypeDef*)ctx, update );
				break;

			default: break;
		}
}

void wxPMClassFinderService::RemoveContext( spContext* ctx )
{
	mpIndexSvc->GetItemTable().
		DeleteByIdAndType( ctx->GetType(), (int)ctx->GetUserData() );
}

void wxPMClassFinderService::WriteFile( spFile* file, bool update )
{
	MMemberListT& lst = file->GetMembers();

	for( size_t i = 0; i != lst.Count(); ++i )
	
		WriteContext( lst[i], update );
}

void wxPMClassFinderService::WriteTypeDef( spTypeDef* td, bool update )
{
	mTBuf.Clear();
	mTBuf.WriteString( td->mOriginalType );

	ItemTable& it = mpIndexSvc->GetItemTable();

	it.SetName( td->mName );
	it.SetDataStr( mTBuf.GetContent() );
	it.SetContext( "" );
	it.SetType( SP_CTX_TYPEDEF );

	if ( update ) 
		it.SetId( (int)td->GetUserData() );

	if ( update )
		it.Write();
	else
		it.Add();
}

void wxPMClassFinderService::WriteOperation( spOperation* op, bool update )
{
	mTBuf.Clear();
	mTBuf.WriteShort( op->mFlags );
	mTBuf.WriteString( op->GetArgListDigest() );
	WriteOpProperties( mTBuf, op );

	ItemTable& it = mpIndexSvc->GetItemTable();

	if ( update ) 
		it.SetId( (int)op->GetUserData() );

	it.SetName( op->mName );
	it.SetContext( op->mScope );
	it.SetDataStr( mTBuf.GetContent() );
	it.SetType( SP_CTX_OPERATION );

	if ( update )
        it.Write();
    else
        it.Add();
}

void wxPMClassFinderService::WriteClass( spClass* cl, bool update )
{
	mTBuf.Clear();

	mTBuf.WriteShort( cl->mFlags );
	mTBuf.WriteByte( cl->mSuperClassNames.GetCount() );
	
	size_t i = 0;
	for( ; i != cl->mSuperClassNames.GetCount(); ++i )
	
		mTBuf.WriteString( cl->mSuperClassNames[i] );
	
	mTBuf.WriteString( cl->mTemplateTypes );

	mTBuf.WriteShort( cl->CountMembers( SP_CTX_OPERATION ) );

	for( i = 0; i != cl->mMembers.Count(); ++i )
	{
		spContext* ctx = cl->mMembers[i];

		if ( ctx->GetType() == SP_CTX_OPERATION )
		{
			spOperation* op = (spOperation*)ctx;

			mTBuf.WriteShort( op->mFlags );
			mTBuf.WriteString( op->mName );
			WriteOpProperties( mTBuf, op );
		}
	}

	mTBuf.WriteShort( cl->CountMembers( SP_CTX_ATTRIBUTE ) );

	for( i = 0; i != cl->mMembers.Count(); ++i )
	{
		spContext* ctx = cl->mMembers[i];

		if ( ctx->GetType() == SP_CTX_ATTRIBUTE )
		{
			spAttribute* attr = (spAttribute*)ctx;

			mTBuf.WriteShort( attr->mFlags );
			mTBuf.WriteString( attr->mName );
			mTBuf.WriteString( attr->mType );
			mTBuf.WriteString( attr->mInitVal );
		}
	}

	ItemTable& it = mpIndexSvc->GetItemTable();

	if ( update ) 
		it.SetId( (int)cl->GetUserData() );

	it.SetName( cl->mName );
	it.SetContext( "" );
	it.SetDataStr( mTBuf.GetContent() );
	it.SetType( SP_CTX_CLASS );

	if ( update )
        it.Write();
    else
        it.Add();
}

spContext* wxPMClassFinderService::ReadContext()
{
	ItemTable& it = mpIndexSvc->GetItemTable();

	wxString nm = it.GetName();
	wxString ct = it.GetContext();
	wxString ds = it.GetDataStr();
	int type = it.GetType();

	if ( type == SP_CTX_CLASS )
		return ReadClass();
	else 
	if ( type == SP_CTX_OPERATION )
		return ReadOperation();
	else
	if ( type == SP_CTX_TYPEDEF )
		return ReadTypeDef();
	else
		return NULL;
}

spClass* wxPMClassFinderService::ReadClass()
{
	ItemTable& it = mpIndexSvc->GetItemTable();
	spClass* cl = new spClass();

	cl->SetUserData( (void*)it.GetId() );

	cl->mName = it.GetName();
	mTBuf.SetContent( it.GetDataStr() );

	cl->mFlags = mTBuf.ReadShort();
	size_t n = (size_t)mTBuf.ReadByte();

	size_t i = 0; wxString s;

	for( ; i != n; ++i )
	{
		s = mTBuf.ReadString();
		cl->mSuperClassNames.Add( s );
	}

	cl->mTemplateTypes = mTBuf.ReadString();

	n = (size_t)mTBuf.ReadShort();

	for( i = 0; i != n; ++i )
	{
		spOperation* op = new spOperation();

		op->mFlags = mTBuf.ReadShort();
		op->mName = mTBuf.ReadString();
		ReadOpProperties( mTBuf, op );

		cl->AddMember( op );
	}

	n = (size_t)mTBuf.ReadShort();

	for ( i = 0; i != n; ++i )
	{
		spAttribute* at = new spAttribute();

		at->mFlags   = mTBuf.ReadShort();
		at->mName    = mTBuf.ReadString();
		at->mType    = mTBuf.ReadString();
		at->mInitVal = mTBuf.ReadString();

		cl->AddMember( at );
	}

	return cl;
}

spOperation* wxPMClassFinderService::ReadOperation()
{
	ItemTable& it = mpIndexSvc->GetItemTable();
	spOperation* op = new spOperation();

	op->SetUserData( (void*)it.GetId() );

	op->mName = it.GetName();
	op->mScope = it.GetContext();

	mTBuf.SetContent( it.GetDataStr() );
	op->mFlags = mTBuf.ReadShort();
	op->mArgsDigest = mTBuf.ReadString();

	ReadOpProperties( mTBuf, op );

	return op;
}

spTypeDef* wxPMClassFinderService::ReadTypeDef()
{
	ItemTable& it = mpIndexSvc->GetItemTable();
	spTypeDef* td = new spTypeDef();

	td->SetUserData( (void*)it.GetId() );
	td->mName = it.GetName();

	mTBuf.SetContent( it.GetDataStr() );
	td->mOriginalType = mTBuf.ReadString();

	return td;
}

void wxPMClassFinderService::WriteOpProperties( TextBuffer& buf, spOperation* op )
{
	buf.WriteString( op->mRetType );
	buf.WriteByte( op->mParams.Count() );

    size_t i;
	for ( i = 0; i != op->mParams.Count(); ++i )
	{
		spParameter* param = op->mParams[i];

		buf.WriteString( param->mName );
		buf.WriteString( param->mType );
		buf.WriteString( param->mInitVal );
	}
}

void wxPMClassFinderService::ReadOpProperties( TextBuffer& buf, spOperation* op )
{
	op->mRetType = buf.ReadString();
	size_t n = (size_t)buf.ReadByte();

    size_t i;
	for ( i = 0; i != n; ++i )
	{
		spParameter* param = new spParameter();

		param->mName    = buf.ReadString();
		param->mType    = buf.ReadString();
		param->mInitVal = buf.ReadString();

		op->mParams.Add( param );
	}
}

spFile* wxPMClassFinderService::ParseDocContent( wxPMDocument& doc )
{
	I_ContainingText* ctinf = 
		(I_ContainingText*)doc.GetInterface( "I_ContainingText" );

	if ( !ctinf )
		return NULL;

	wxString txt;
	ctinf->GetAllText( &txt );
	char* start = (char*)(const char*)txt;
	char* end = start + txt.Length();

	return GetParser()->Parse( start, end, FALSE );
}

spFile* wxPMClassFinderService::LoadInfoForFile( int fileId, int indexId )
{
	spFile* info = new spFile();

	ItemTable& it = mpIndexSvc->GetItemTable();

	it.FindByFileAndIndex( fileId, indexId );

	while( it.Ok() && 
		   indexId == it.GetIndexId() &&
		   fileId == it.GetFileId() )
	{
		spContext* ctx = ReadContext();
		info->AddMember( ctx );

		it.Next();
	}

	it.End();

	return info;
}

void wxPMClassFinderService::SyncIndexWithContent( spFile* oldInfo, spFile* newInfo, int fileId, int indexId )
{
	ItemTable& it = mpIndexSvc->GetItemTable();
	it.Begin();
	it.SetFileId( fileId );
	it.SetIndexId( indexId );

	size_t i;
	MMemberListT& oldLst = oldInfo->GetMembers();
	MMemberListT& newLst = newInfo->GetMembers();

	wxHashTable oldHash( wxKEY_STRING );
	wxHashTable newHash( wxKEY_STRING );

	for ( i = 0; i != oldLst.Count(); ++i )
		oldHash.Put( oldLst[i]->GetFullDigest(), oldLst[i] );

	// find added and updated contexts

	for ( i = 0; i != newLst.Count(); ++i )
	{
		spContext* newCtx = newLst[i];
		wxString dig = newCtx->GetFullDigest();
		newHash.Put( dig, newLst[i] );

		spContext* oldCtx = (spContext*)oldHash.Get( dig );

		if ( !oldCtx )
			WriteContext( newCtx ); // add
		else 
		if ( !ContextsEqual( oldCtx, newCtx ) )
        {
#if 0
			newCtx->SetUserData( oldCtx->GetUserData() );
			// FIMXE:: update causes db-index coruption
			WriteContext( newCtx, TRUE ); // update
#else
			// do it brutally, but safely...
			RemoveContext( oldCtx );
			WriteContext( newCtx );
#endif

        }
		else
        {
			// for unchanged objects,
			// transfer original indexing ids to new objects
			newCtx->SetUserData( oldCtx->GetUserData() );
        }
	}

	it.End();

	// find removed contexts

	for ( i = 0; i != oldLst.Count(); ++i )
	{
		spContext* oldCtx = oldLst[i];
		spContext* newCtx = (spContext*)newHash.Get( oldCtx->GetFullDigest() );

		if ( !newCtx )
			RemoveContext( oldCtx );
	}
}

void wxPMClassFinderService::RemoveCachedInfoForDoc( wxPMDocument& doc )
{
	spFile* fl = (spFile*)mDocHash.Get( (long)&doc );

	if ( !fl )
		return;

	mDocHash.Delete( (long)&doc );
	wxDELETE(fl);
}

bool wxPMClassFinderService::ContextsEqual( spContext* ctx1, spContext* ctx2 )
{
	int type1 = ctx1->GetType();
	int type2 = ctx2->GetType();

	if ( type1 != type2 )
		return TRUE; // FIXME:: what to do with contexts having same
					 //			digest but different types?

	if ( type1 == SP_CTX_CLASS )
	{
		MMemberListT& lstA = ctx1->GetMembers();
		MMemberListT& lstB = ctx2->GetMembers();

		if ( lstA.Count() != lstB.Count() )
			return FALSE;

		for( size_t i = 0; i != lstA.Count(); ++i )
		{
			spContext* c1 = lstA[i];
			spContext* c2 = lstB[i];
			int t1 = c1->GetType();

			if ( t1 != c1->GetType() )
				return FALSE;

			if ( c1->GetFullDigest() !=
				 c2->GetFullDigest() )
				 return FALSE;
		}
	}

	return TRUE;
}

/***** Implementation for class TextBuffer *****/

// some helpers

// static char _gConvBuf[5]; //HVL:: Not in use

#define BITS_TO_HEX( x ) ( x >= 10 ? (char)((int)'A' + x - 10 ) \
								   : (char)((int)'0' + x ) )

#define HEX_TO_BITS( c ) ( c >= 'A' ? 10 + c - 'A' \
									: c - '0' )

void byte_to_hex( int byte, char* buf )
{
	int p1 = byte & 0xF;
	int p2 = ( byte >> 4 ) & 0xF;

	buf[ 1 ] = BITS_TO_HEX( p1 );
	buf[ 0 ] = BITS_TO_HEX( p2 );
}

void short_to_hex( int s, char* buf )
{
	byte_to_hex( s & 0xFF, buf + 2 );
	byte_to_hex( ( s >> 8 ) & 0xFF, buf );
}

int hex_to_byte( char* hex )
{
	char c1 = hex[0], c2 = hex[1];

	int p1 = HEX_TO_BITS( c1 );
	int p2 = HEX_TO_BITS( c2 );

	return ( ( p1 << 4 ) & 0xF0 ) | p2;
}

int hex_to_short( char* hex )
{
	int p1 = hex_to_byte( hex );
	int p2 = hex_to_byte( hex + 2 );

	return ( ( p1 << 8 ) & 0xFF00 ) | p2;
}

TextBuffer::TextBuffer()

	: mpBuf( NULL ),
	  mLen(0),
	  mBufLen(0)
{}

TextBuffer::~TextBuffer()
{
	CleanUp();
}

void TextBuffer::SetContent( const wxString& text )
{
	CleanUp();
	mReadHead = 0;

	int len = text.Length() + 1;
	mpBuf = new char[ len ];

	memcpy( mpBuf, text.c_str(), len - 1 );
	mpBuf[ len - 1 ] = '\0';

	mLen = len - 1;
	mBufLen = len;
}

const char* TextBuffer::GetContent()
{
	mpBuf[mLen] = '\0';
	return mpBuf;
}

void TextBuffer::WriteString( const wxString& s )
{
	size_t len = s.Length();

	if ( len > 254 )
	{
		//int avoidCompilerWarning = 0;
		//wxASSERT(avoidCompilerWarning); // FOR NOW::

		len = 254;
	}
	
	Resize( mLen + 1 + len );

	mpBuf[mLen] = (unsigned char)(len + 1); // add one to avoid zero value (terminator)
	
	const char* st = s.c_str();

	size_t i;
	for ( i = 0; i < len; ++i )
	{
		if ( st[i] == '\0' )
			wxFAIL_MSG( "stopppp" ); // DBG::
	}
	memcpy( mpBuf + mLen + 1, st, len );

	mLen += len + 1;
}

void TextBuffer::WriteShort( int s )
{
	Resize( mLen + 4 );

	short_to_hex( s, mpBuf + mLen );

	mLen += 4;
}

void TextBuffer::WriteByte( int byte )
{
	Resize( mLen + 1 );

	if ( byte > 254 )
	{
		int avoidCompilerWarning = 0;
		wxASSERT(avoidCompilerWarning); // FOR NOW::
	}

	mpBuf[mLen] = (unsigned char)(byte + 1);

	++mLen;
}

int TextBuffer::ReadShort()
{
	wxASSERT( mReadHead < mLen ); 

	int s = hex_to_short( mpBuf + mReadHead );

	mReadHead += 4;

	return s;
}

int TextBuffer::ReadByte()
{
	wxASSERT( mReadHead < mLen );

	int b = (unsigned char)mpBuf[ mReadHead ] - 1; // remove '\0' padding

	++mReadHead;

	return b;
}

wxString TextBuffer::ReadString()
{
	wxASSERT( mReadHead < mLen );

	int len = (unsigned char)mpBuf[ mReadHead ] - 1;
	++mReadHead;

	char tmp = mpBuf[ mReadHead + len ];
	mpBuf[ mReadHead + len ] = '\0';

	wxString s = mpBuf + mReadHead;

	mpBuf[ mReadHead + len ] = tmp;

	mReadHead += len;

	return s;
}

void TextBuffer::CleanUp()
{
	if ( mpBuf )
	{
		wxDELETEA(mpBuf);
		mLen = 0;
		mBufLen = 0;
	}
}

void TextBuffer::Clear()
{
	if ( !mpBuf )
		SetContent( "" );

	mLen = 0;
}

void TextBuffer::Resize( int newLen )
{
	if ( mBufLen >= newLen )
		return;

	mBufLen = ( newLen / 16 ) * 32 + 32;

	char* newBuf = new char[ mBufLen ];

	memcpy( newBuf, mpBuf, mLen );

	wxDELETEA(mpBuf);

	mpBuf = newBuf;
}
