/////////////////////////////////////////////////////////////////////////////
// Name:        No names yet.
// Purpose:     Contrib. demo
// Author:      Aleksandras Gluchovas
// Modified by:
// Created:     2000/02/08
// RCS-ID:      $Id: classfinder_svc.h,v 1.11 2001/11/15 06:12:59 aleksandras Exp $
// Copyright:   (c) Aleksandras Gluchovas
// Licence:   	GPL (version 2.0 or later)
/////////////////////////////////////////////////////////////////////////////

#ifndef __CLASSFINDER_SVC_G__
#define __CLASSFINDER_SVC_G__

#ifdef __GNUG__
    #pragma interface "classfinder_svc.h"
#endif

#include "indexing_svc.h"
#include "wx/txtstrm.h"
#include "wx/hash.h"

#include "wx/pmf/inf/textediting.h"

class SourceParserBase;
class spContext;
class spFile;
class spOperation;
class spTypeDef;
class spClass;
class TextBuffer;
class MMemberListT;
class spWxDefineArrayMacroPlugin;

class TextBuffer : public wxObject
{
public:
	TextBuffer();
	~TextBuffer();

	void SetContent( const wxString& text );
	const char* GetContent();

	void WriteString( const wxString& s );
	void WriteShort( int s );
	void WriteByte( int byte );
	int ReadShort();
	int ReadByte();
	wxString ReadString();

	void CleanUp();
	void Clear();

protected:
	void Resize( int newLen );

protected:
	char* mpBuf;
	int mReadHead;
	int mLen;
	int mBufLen;
};

struct SourceEditingContextInfo
{
	SourceEditingContextInfo() : mHelpNeeded( FALSE )
		{}

	wxString mEditedText;

	wxStringList mCtxTypes;
	wxStringList mCtxIdentifiers;
	wxString mScope;
	bool mHelpNeeded; // FOR NOW::...

	// usefull for help-system, when
	// complete resolution is not necessery
	wxString mLastUnrssolvedCtx;
	wxString mLastUnresolvedName;
};

class wxPMClassFinderService : public wxPMSimpleService,
							   protected I_IndexHandler
{
public:
	wxPMClassFinderService();
	~wxPMClassFinderService();

	// overriden methods of wxPMService

	virtual bool Start();
	virtual bool Stop();

	virtual wxPMMergableMenuChain* GetMergableMenuChain();

	// some higher-level algorithms

	spClass* FindClass( const wxString& name );

	void FindMembers( spClass* cl, MMemberListT& toLst, wxString* matchPattern = NULL, int filterMask = 0xFFFFFF, 
					  bool searchBases = TRUE, wxStringList* baseClassNames = NULL, bool sort = TRUE );

	spContext* FindMember( spClass* cl, const wxString& name, 
						   bool searchSuperClasses = TRUE );

	spOperation* FindArrayOperator( spClass* ofClass );

	void DumpAllItems( wxTextOutputStream& stm );
	void DumpAllItemsToFile( const wxString& fname );

	// in-text parsing and condext detection

	bool ExtractSourceContextInfo( wxPMDocument& fromDoc, SourceEditingContextInfo& toInfo );
	spContext* DetectLastContext( SourceEditingContextInfo& ctx );

	// implementation of I_IndexHandler interface

	virtual const char* GetIndexHandlerName();
	virtual const char* GetIndexHandlerDescription();

	virtual bool OnClearIndexItems( int indexId );
	virtual bool OnBeginBuild( int indexId );
	virtual bool OnIndexFile( const wxPMFileInfo& file, int fileId, int indexId );
	virtual void OnDumpIndexForFile( wxTextOutputStream& stm, int fileId, int indexId );
	virtual bool OnEndBuild( int indexId );

	virtual bool DocumentMonitoringNeeded(); // default imp. returns FALSE
	virtual bool CanAcceptFile( const wxPMFileInfo& file, int indexId );
	virtual void OnDocumentsSwitched( wxPMDocument* fromDoc, wxPMDocument* toDoc, int fileId, int indexId );
	virtual void OnDocumentSaved( wxPMDocument& doc, int fileId, int indexId );
	virtual void OnDocumentChangesDiscarded( wxPMDocument& doc, int fileId, int indexId );

	virtual SourceParserBase* GetParser();

	/*** command handlers ***/

	enum
	{
		ID_DUMP_CLASS_INFO = 100,
		ID_TEST_EXPR_PARSING
	};

	void OnDumpClassInfo( wxCommandEvent& event );
	void OnTestExprParsing( wxCommandEvent& event );

	DECLARE_EVENT_TABLE()
	DECLARE_DYNAMIC_CLASS( wxPMClassFinderService )

protected:

	// methods for (de)serialization of collected 
	// class/operation info in and out of wxString object

	void WriteContext( spContext* ctx, bool update = FALSE );
	void RemoveContext( spContext* ctx );

	void WriteFile( spFile* file, bool update = FALSE );
	void WriteOperation( spOperation* op, bool update = FALSE );
	void WriteClass( spClass* cl, bool update = FALSE );
	void WriteTypeDef( spTypeDef* td, bool update = FALSE );

	spClass* ReadClass();
	spOperation* ReadOperation();
	spContext* ReadContext();
	spTypeDef* ReadTypeDef();

	void WriteOpProperties( TextBuffer& buf, spOperation* op );
	void ReadOpProperties( TextBuffer& buf, spOperation* op );

	spFile* ParseDocContent( wxPMDocument& doc );
	spFile* LoadInfoForFile( int fileId, int indexId );
	void SyncIndexWithContent( spFile* oldInfo, spFile* newInfo, int fileId, int indexId );
	void RemoveCachedInfoForDoc( wxPMDocument& doc );

	bool ContextsEqual( spContext* ctx1, spContext* ctx2 );

protected:
	wxPMIndexingService* mpIndexSvc;
	SourceParserBase* mpParser;
	spWxDefineArrayMacroPlugin* mpWxMacroParser;
	char* mpTmpBuf;
	int mTmpBufLen;
	TextBuffer mTBuf;

	wxHashTable mDocHash; // doc <=> cached info
};
							   

#endif /* __CLASSFINDER_SVC_G__ */

