/////////////////////////////////////////////////////////////////////////////
// Name:        No names yet.
// Purpose:     Contrib. demo
// Author:      Aleksandras Gluchovas
// Modified by:
// Created:     2001/03/28
// RCS-ID:      $Id: codecompl_svc.cpp,v 1.15 2001/11/20 21:31:53 georgetasker Exp $
// Copyright:   (c) Aleksandras Gluchovas (alex@soften.ktu.lt)
// Licence:   	GPL (version 2.0 or later)
/////////////////////////////////////////////////////////////////////////////

#ifdef __GNUG__
    #pragma implementation "codecompl_svc.h"
#endif

// For compilers that support precompilation, includes "wx/wx.h"
#include "wx/wxprec.h"

#ifdef __BORLANDC__
    #pragma hdrstop
#endif

#ifndef WX_PRECOMP
    #include "wx/wx.h"
#endif
#include "wx/wspp/rescrolw.h"
#include "codecompl_svc.h"
#include "classfinder_svc.h"

#include "wx/log.h"
#include "myutils.h"

static const wxSize _gComplPanelSize( 270, 150 ); // default

/***** Implementation for class wxPMCodeCompletionService *****/

IMPLEMENT_DYNAMIC_CLASS( wxPMCodeCompletionService, wxPMSimpleService )

BEGIN_EVENT_TABLE( wxPMCodeCompletionService, wxPMSimpleService )

END_EVENT_TABLE()

wxPMCodeCompletionService::wxPMCodeCompletionService()
	: mpClassFinderSvc( NULL )
{}

wxPMCodeCompletionService::~wxPMCodeCompletionService()
{}

// overriden methods of wxPMService

bool wxPMCodeCompletionService::Start()
{
	mpClassFinderSvc = (wxPMClassFinderService*)wxPMBootSvc().
					FindUniqueServiceInstance( "wxPMClassFinderService" );

	if ( !mpClassFinderSvc || !mpClassFinderSvc->IsStarted() )
	{
		wxMessageBox( "WARNING:\n\n"
			          "CodeCompletion-service failed locating active instance of mpClassFinderSvc,\n"
					  "code completion will not be available" );
		return FALSE;
	}

	wxPMBootSvc().AddListener( this );

	return TRUE; // started ok
}

bool wxPMCodeCompletionService::Stop()
{
	return TRUE;
}

wxPMMergableMenuChain* wxPMCodeCompletionService::GetMergableMenuChain()
{
	return NULL;
}

void wxPMCodeCompletionService::OnServiceStarted( wxPMService& service )
{
	wxObject* obj = service.GetWxObject();

	if ( obj->IsKindOf( CLASSINFO(wxPMDocument) ) )
	{
		wxPMDocument* doc = (wxPMDocument*)obj;

		I_EditableText* inf = (I_EditableText*)doc->GetInterface( "I_EditableText" );

		if ( !inf )
			return;

		inf->AddKeyInterceptor( this, 0 );
	}
}

bool wxPMCodeCompletionService::OnEditorKeyEvent( wxKeyEvent& key, 
												  bool isChar, wxPMService* fromSvc )
{
	if ( key.m_keyCode == ' ' && key.ControlDown() && isChar )
	{
		wxPMDocument* doc = (wxPMDocument*)fromSvc->GetWxObject();

		SourceEditingContextInfo ctxInfo;
		if ( !mpClassFinderSvc->ExtractSourceContextInfo( *doc, ctxInfo ) )
			return FALSE;

		spContext* found = mpClassFinderSvc->DetectLastContext( ctxInfo );

		if ( found || ctxInfo.mLastUnrssolvedCtx.Length() == 0 )
			return FALSE;

		found = mpClassFinderSvc->FindClass( ctxInfo.mLastUnrssolvedCtx );
		if ( !found )
			return FALSE;

		///////// TBD:: comments

		wxPoint pos; wxSize size;
		AutoComplPanel::CalcPanelPosition( FALSE, size, pos, *doc );

		MMemberListT lst;
		mpClassFinderSvc->FindMembers( (spClass*)found, lst, NULL, SP_CTX_OPERATION | SP_CTX_ATTRIBUTE );

		spContext* match = NULL;
		match = FindUniqueMatch( lst, ctxInfo.mLastUnresolvedName );

		if ( match )
		{
			AutoComplPanel* panel = new AutoComplPanel( doc, lst, ctxInfo.mLastUnresolvedName, found, TRUE, pos, size );
			panel->StoreScreenPos();
			panel->CompleteName( match, ctxInfo.mLastUnresolvedName );
			return TRUE;
		}			  
		else
		{
			AutoComplPanel* panel = new AutoComplPanel( doc, lst, ctxInfo.mLastUnresolvedName, found, FALSE, pos, size );
			panel->StoreScreenPos();
		}

		return TRUE;
	}
	else
	if ( isChar && 
		 ( key.m_keyCode == '.' || key.m_keyCode == '>' ) )
	{
		// FOR NOW::

		wxLogMessage( "Got '.' char" );

		wxPMDocument* doc = (wxPMDocument*)fromSvc->GetWxObject();

		SourceEditingContextInfo ctxInfo;
		if ( !mpClassFinderSvc->ExtractSourceContextInfo( *doc, ctxInfo ) )
			return FALSE;

		unsigned int len = ctxInfo.mEditedText.Length();
		if ( key.m_keyCode == '>' && len && ctxInfo.mEditedText[len-1] != '-' )
			return false;

		if ( key.m_keyCode == '>' && len )
			ctxInfo.mEditedText.Remove( len-1 ); // remove trailing "-"

		spContext* found = mpClassFinderSvc->DetectLastContext( ctxInfo );

		if ( !found || found->GetType() != SP_CTX_CLASS )
			return FALSE;


		wxPoint pos; wxSize size;
		AutoComplPanel::CalcPanelPosition( FALSE, size, pos, *doc );


		/*
		MMemberListT lst;

		for( int i = 0; i!=5;++i )
		{
			spOperation* op = new spOperation();
			op->SetName( IntToStr( i ) + "_Foo" );

			if ( i != 0 )
			{
				op->AddParameter( new spParameter( "x", "int" ) );
				op->AddParameter( new spParameter( "y", "int" ) );
			}

			op->mRetType = "int";
			lst->Add( op );
		}
		*/

		///////// TBD:: comments

		MMemberListT lst;
		mpClassFinderSvc->FindMembers( (spClass*)found, lst, NULL, SP_CTX_OPERATION | SP_CTX_ATTRIBUTE );

		AutoComplPanel* panel = new AutoComplPanel( doc, lst, "", found, FALSE, pos, size );

		panel->StoreScreenPos();

		return FALSE;
	}

	return FALSE;
}

spContext*  wxPMCodeCompletionService::FindUniqueMatch( MMemberListT& inLst, wxString name )
{
	name = name.Lower();
	wxString found;
	spContext* foundCtx = NULL;

	for( size_t i = 0; i != inLst.Count(); ++i )
	{
		spContext* ctx = inLst[i];
		wxString ctxName = ctx->mName.Lower();

		if ( ctxName.StartsWith( name ) )
		{
			if ( found.Length() && found != ctxName )
				return NULL;

			found = ctxName;

			if ( !foundCtx )
				foundCtx = ctx;
		}
	}

	return foundCtx;
}

/***** Implementation for class TCtxListPinPainter *****/

wxBitmap TCtxListPinPainter::mBmps[11];

#define WXS_ICON( name ) wxPMBmpMgr().FindBitmap( "code##name" )

TCtxListPinPainter::TCtxListPinPainter()

	: TPinPainterBase( CTX_LIST_PIN_TC ),
	  mBkBrush( wxSystemSettings::GetSystemColour( wxSYS_COLOUR_WINDOW ), wxSOLID )
{
	if ( mBmps[0].Ok() ) return;

	wxPMBitmapManager& bm = wxPMBmpMgr();
	mBmps[CLASS_IMAGE]		  = bm.FindBitmap( "code/class" );
	mBmps[PRIV_MTD_IMAGE]	  = bm.FindBitmap( "code/priv_mtd" );
	mBmps[PROT_MTD_IMAGE]	  = bm.FindBitmap( "code/prot_mtd" );
	mBmps[PUB_MTD_IMAGE]	  = bm.FindBitmap( "code/pub_mtd" );
	mBmps[PRIV_VAR_IMAGE]	  = bm.FindBitmap( "code/priv_var" );
	mBmps[PROT_VAR_IMAGE]	  = bm.FindBitmap( "code/prot_var" );
	mBmps[PUB_VAR_IMAGE]	  = bm.FindBitmap( "code/pub_var" );
	mBmps[PUB_PURE_MDT_IMAGE] = bm.FindBitmap( "code/pub_pure_mtd" );
	mBmps[PRIV_MTD_DEF_IMAGE] = bm.FindBitmap( "code/priv_mtd_def" );
	mBmps[PROT_MTD_DEF_IMAGE] = bm.FindBitmap( "code/prot_mtd_def" );
	mBmps[PUB_MTD_DEF_IMAGE]  = bm.FindBitmap( "code/pub_mtd_def" );

	for( int i = 0; i != 11; ++i )
	{
		wxMask* mask = new wxMask( mBmps[i], *wxRED );
		mBmps[i].SetMask( mask );
	}
}

void TCtxListPinPainter::DrawPin( TPinBase* pPin, wxTextEditorView& view, wxDC& dc, 
								  const wxPoint& pos, const wxSize& dim )
{
	CtxListPin* pMyPin = (CtxListPin*)pPin;

	wxBitmap& bmp = mBmps[ pMyPin->mIconNo ];

	dc.DrawBitmap( bmp, pos.x, pos.y, TRUE );

	int w = bmp.GetWidth();
	int h = bmp.GetHeight();

	// fill in gaps around image

	if ( w < dim.x )
	{
		dc.SetPen( *wxTRANSPARENT_PEN );
		dc.SetBrush( mBkBrush );

		dc.DrawRectangle( pos.x + w, pos.y,
			              dim.x - w + 1, dim.y + 1 );
	}
	
	if ( h < dim.y )
	{
		if ( !(w < dim.x) )
		{
			dc.SetPen( *wxTRANSPARENT_PEN );
			dc.SetBrush( mBkBrush );
		}

		dc.DrawRectangle( pos.x, pos.y + h,
			              w + 1, dim.y - h + 1);
	}
}

/***** Implementation for class ContextListBox *****/

BEGIN_EVENT_TABLE( ContextListBox, wxTextEditorView )

	EVT_CHAR    ( ContextListBox::OnChar    )
	EVT_KEY_DOWN( ContextListBox::OnKeyDown )
	EVT_KILL_FOCUS( ContextListBox::OnKillFocus )
	EVT_LEFT_DCLICK( ContextListBox::OnDblClick )

END_EVENT_TABLE()

static int ctx_visibility_to_var_image_no( int flags, int type, bool isDef = TRUE )
{
	if ( flags & SP_FLAG_VIS_PRIVATE )
	{
		if ( type == SP_CTX_ATTRIBUTE )
			return TCtxListPinPainter::PRIV_VAR_IMAGE;
		else
			return isDef ? TCtxListPinPainter::PRIV_MTD_DEF_IMAGE
						 : TCtxListPinPainter::PRIV_MTD_IMAGE;
	}			
	else
	if ( flags & SP_FLAG_VIS_PROTECTED )
	{
		if ( type == SP_CTX_ATTRIBUTE )
			return TCtxListPinPainter::PROT_VAR_IMAGE;
		else
			return isDef ? TCtxListPinPainter::PROT_MTD_DEF_IMAGE
						 : TCtxListPinPainter::PROT_MTD_IMAGE;
	}
	else
	{
		if ( type == SP_CTX_ATTRIBUTE )
			return TCtxListPinPainter::PUB_VAR_IMAGE;
		else
			return isDef ? TCtxListPinPainter::PUB_MTD_DEF_IMAGE
						 : TCtxListPinPainter::PUB_MTD_IMAGE;
	}
}

ContextListBox::ContextListBox( wxWindow*     parent, 
							    wxWindowID    id,
								MMemberListT& ctxList,
								const wxString& nameTyped,
								spFile*       pFile,
								const wxSize& size )

	: wxTextEditorView( parent, id, new wxTextEditorModel(), 0 ),
	  mpFile( pFile ),
	  mPopupMode( FALSE ),
	  mNameTyped( nameTyped )
{
	SetSize(0,0, size.x, size.y );
	mCtxList = ctxList;
	wxString text;

	mpModel->SetReadOnly( TRUE );
	SetLineTrackingMode( TRUE, wxColour( 0,255,255 ) );

	mpModel->AddView( this );
	mpModel->SetActiveView( this );

	AddPinPainter( new TCtxListPinPainter() );


	size_t i = 0;
	for ( ; i != mCtxList.Count(); ++i)
	{
		if ( i != 0 )
            text += '\n';

		spContext* pCtx = mCtxList[i];
		
		if ( pCtx->GetType() == SP_CTX_CLASS )
			text += pCtx->GetName();
		else
		if ( pCtx->GetType() == SP_CTX_ATTRIBUTE )
		{
			spAttribute* pAttr = (spAttribute*)pCtx;
	
			text += pCtx->GetName() + " : " + pAttr->mType;
		}
		else
		if ( pCtx->GetType() == SP_CTX_OPERATION )
		{
			spOperation* pOp = (spOperation*)pCtx;

			text += pOp->FormatSignature( TRUE, TRUE, FALSE, FALSE, TRUE );

			if ( pOp->mRetType.length() )

				text += " : " + pOp->mRetType;
		}
	}

	mpModel->InsertText( TPosition( 0,0 ), text.c_str(), text.length() );

	for (i = 0; i != mCtxList.Count(); ++i)
	{
		if ( i != 0 ) text += '\n';

		spContext* pCtx = mCtxList[i];
		int iconNo = -1;
		
		if ( pCtx->GetType() == SP_CTX_CLASS )
			iconNo = TCtxListPinPainter::CLASS_IMAGE;
		else
		if ( pCtx->GetType() == SP_CTX_ATTRIBUTE )
			iconNo = ctx_visibility_to_var_image_no( pCtx->mFlags, pCtx->GetType() );
		else
		if ( pCtx->GetType() == SP_CTX_OPERATION )
		{
			spOperation* pOp = (spOperation*)pCtx;

			if ( pOp->mFlags & SP_FLAG_IS_DEFINITION  ||
				 pOp->mScope.length() != 0 )
				iconNo = ctx_visibility_to_var_image_no( pCtx->mFlags, pCtx->GetType(), TRUE );
			else
				iconNo = ctx_visibility_to_var_image_no( pCtx->mFlags, pCtx->GetType(), FALSE );
		}

		mpModel->AddPin( new CtxListPin( i, iconNo ) );
	}

	SetMaxColumns( 250 );

	// FIXME:: wxTextEditorView's bug??
	SetRowsPerPage( GetRowsPerPage() - 1 );

	RecalcPagingInfo();
	mpModel->OnGotoLine( FindClosestMatch(), 0 );
}

ContextListBox::~ContextListBox()
{
	/*
	if ( mpFile ) 
		wxDELETE(mpFile);
	else
	{
        size_t i;
		for (i = 0; i != mCtxList.Count(); ++i)
			wxDELETE(mCtxList[i]);
	}
	*/
}

void ContextListBox::SetPopupMode( bool on )
{
	mPopupMode = on;
}

#define ALPHA_NUM( ch ) ( ( ch >= 'a' && ch <= 'z' ) || \
	                      ( ch >= 'A' && ch <= 'Z' ) || \
						  ( ch >= '0' && ch <= '9' ) || \
						  ( ch == '_' ) )

#define UPPER_CHAR( ch ) ( ( ch >= 'a' && ch <='z' ) ? ch - 'a' + 'A' : ch )

void ContextListBox::OnChar( wxKeyEvent& event )
{
	if ( IsClipboardCmd( event ) )
	{
		event.Skip();
		return;
	}

	if ( event.m_keyCode == WXK_DOWN ||
		 event.m_keyCode == WXK_UP ||
		 event.m_keyCode == WXK_END ||
		 event.m_keyCode == WXK_HOME )
	{
		event.Skip();
		return;
	}

	if ( !event.ControlDown() &&
		 !event.AltDown() )
	{
			if ( event.m_keyCode == WXK_BACK )
			{
				if ( mNameTyped.length() != 0 ) 
				
					mNameTyped.erase( mNameTyped.length() -1, 1 );
				else
				{
					OnCloseList( event );
					return;
				}

				OnKeyReceived( event );
			}
			else
			{
				if ( ALPHA_NUM( event.m_keyCode ) )

					mNameTyped += event.m_keyCode;

				OnKeyReceived( event );
			}

			if ( ALPHA_NUM( event.m_keyCode ) )

				mpModel->OnGotoLine( FindClosestMatch(), 0 );
	}

	// do not let it propagate to wxTextEditorView
}

void ContextListBox::OnKeyDown( wxKeyEvent& event )
{
	if ( IsClipboardCmd( event ) )
	{
		event.Skip();
		return;
	}

	if ( !event.ControlDown() &&
		 !event.AltDown()     &&
		 !event.ShiftDown() )
	{
		if ( event.m_keyCode == WXK_LEFT ) 
		
			OnCloseList( event );
		else
		if ( event.m_keyCode == WXK_RIGHT ) 
		
			OnCloseList( event );
		else
		if ( event.m_keyCode == WXK_UP ) 
		
			mpModel->OnMoveUp();
		else
		if ( event.m_keyCode == WXK_DOWN ) 
		
			mpModel->OnMoveDown();
		else
		if ( event.m_keyCode == WXK_HOME ) 
		
			mpModel->OnStartOfText();
		else
		if ( event.m_keyCode == WXK_END ) 
		
			mpModel->OnEndOfText();
		else
		if ( event.m_keyCode == WXK_NEXT ) 
		
			mpModel->OnPageDown();
		else
		if ( event.m_keyCode == WXK_PRIOR ) 
		
			mpModel->OnPageUp();
		else
		if ( event.m_keyCode == WXK_DELETE )

			OnCloseList( event );
		else
		if ( event.m_keyCode == WXK_ESCAPE )

			OnCloseList( event );	
		else
		if ( ( event.m_keyCode >= WXK_START && 
			   event.m_keyCode != WXK_SHIFT )
			  && mPopupMode
		   )
		   OnCloseList( event );
	}
	else
	{
		if ( event.m_keyCode == WXK_INSERT && mPopupMode )

			OnCloseList( event );	
		else
		if ( !event.ShiftDown() )

			OnCloseList( event );
	}
}

void ContextListBox::OnKillFocus( wxFocusEvent& event )
{
	OnFocusKilled();
}

void ContextListBox::OnDblClick( wxMouseEvent& event )
{
	wxKeyEvent key;
	key.m_keyCode = WXK_RETURN;
	key.m_altDown     = FALSE;
	key.m_controlDown = FALSE;
	key.m_shiftDown   = FALSE;

	OnChar( key );
}

int ContextListBox::FindClosestMatch()
{
	size_t closest = 0, maxMatches = 0;

	wxString& s2 = mNameTyped;

	for( size_t i = 0; i != mCtxList.Count(); ++i )
	{
		wxString& s1 = mCtxList[i]->GetName();
		
		size_t len = wxMin( s1.length(), s2.length() );

		size_t n;
		for( n = 0; n != len && 
			 UPPER_CHAR(s1[n]) == UPPER_CHAR(s2[n]) ; ++n );

		if ( n > maxMatches )
		{
			closest = i;
			maxMatches = n;
		}
	}

	return closest;
}

spContext* ContextListBox::GetMatchingContext()
{
	return mCtxList[ mpModel->GetCursor().mRow ];
}

/***** Implementation for class AutoComplPanel *****/

BEGIN_EVENT_TABLE( AutoComplPanel, wxPanel )

END_EVENT_TABLE()

wxSize AutoComplPanel::mPanelSize( 270, 150 ); // default

AutoComplPanel::AutoComplPanel( wxPMDocument* doc,
							    MMemberListT& ctxList,
								const wxString& nameTyped,
								spContext*    ctxAtCursor,
								bool completeNameNow,
							    const wxPoint& pos,
							    const wxSize&  size )

	: wxPanel( ((I_EditableText*)doc->GetInterface( "I_EditableText" ))->GetTextWidgetWindow(),
			   -1, 
			   pos, size, wxSUNKEN_BORDER ),
	  mpCtxBox( NULL  ),
	  mpArgBox( NULL ),
	  mpDoc( doc ),
	  mpCtxAtCursor( ctxAtCursor ),
	  mNameTyped( nameTyped )
{
	SetBackgroundColour( wxSystemSettings::GetSystemColour( wxSYS_COLOUR_WINDOW ) );

	mpInfTextEdit = (I_EditableText*)doc->GetInterface( "I_EditableText" );
	mpInfContText = (I_ContainingText*)doc->GetInterface( "I_ContainingText" );
	mpInfTextEdit->HoldCursor(TRUE);

	//monitor->Enable( FALSE );

	mCtxList = ctxList;
	int w = 0, h = 0;
	GetClientSize( &w, &h );
	mpArgBox     = NULL;
	mArgHintMode = FALSE;

	if ( !completeNameNow )
	{
		mpCtxBox = new AutoComplBox( this, mCtxList, mNameTyped, NULL, wxSize( w,h ) );

		mpCtxBox->SetPopupMode( TRUE );
		mpCtxBox->SetFocus();

		mpArgBox     = NULL;
		mArgHintMode = FALSE;
	}
}

AutoComplPanel::~AutoComplPanel()
{
	if ( mpCtxAtCursor )
		delete mpCtxAtCursor;

	//wxMessageBox("AutoComplPanel destoryed!");

	//mpMonitor->Enable( TRUE );

	//mpFile->GetEditor().HoldCursor( FALSE );
}

bool AutoComplPanel::CalcPanelPosition( bool      above,
									    wxSize&   size,
									    wxPoint&  pos,
										wxPMDocument& doc )
{
	if ( size.x == 0 )

		size = AutoComplPanel::mPanelSize;

	I_EditableText* infEditText   = (I_EditableText*)doc.GetInterface( "I_EditableText" );
	I_ContainingText* infContText = (I_ContainingText*)doc.GetInterface( "I_ContainingText" );
	wxSize charSz = infEditText->GetCharacterSize();

	int w = 0, h = 0;
	infEditText->GetTextWidgetWindow()->GetClientSize( &w, &h );

	int line=0, col=0, x=0, y=0;
	infContText->GetCursorPosition( &line, &col );

	infEditText->TextPosToScreenPos( line, col, &x, &y );
	pos.x = x;
	pos.y = y + charSz.y;

	// first, decide whether it's enough space above or
	// below (as requested), and change this setting if
	// it's really necessary

	int spcAbove = pos.y - charSz.y;
	int spcBelow = h - pos.y;


	if ( above && 
		 spcAbove < size.y && spcBelow > spcAbove )

		 above = FALSE;
	else
	if ( !above &&
		 spcBelow < size.y && spcAbove > spcBelow )

		 above = TRUE;
		
	// clip child's dimensions if necessary

	if ( above &&
		 spcAbove < size.y )

		 size.y -= size.y - spcAbove;
	else
	if ( !above &&
		 spcBelow < size.y )
		 
		 size.y -= size.y - spcBelow;

	// position vertically

	if ( above )
	
		pos.y -= charSz.y + size.y; 

	// position horizontally

	if ( pos.x + size.x > w )
	{
		if ( size.x > w )
		{
			size.x = w;
			pos.x = 0;
		}
		else
			pos.x = w - size.x;
	}

	int cut = 0; // FOR NOW::

	if ( pos.x + size.x >= h ) 
	
		size.x -= cut;

	if ( pos.y == 0 )
	{
		pos.y  += cut;
		size.y -= cut;
	}

	if ( pos.y + size.y >= h )
	
		size.y -= cut;

	return above;
}

void AutoComplPanel::StoreScreenPos()
{
	int line=0, col=0;
	mpInfContText->GetCursorPosition( &line, &col );

	mPrevCursorLine = line;
	mpInfTextEdit->GetPagePosition( &mPrevPageLine, &mPrevPageCol );
}

// overriden methods of ContextListBox

void AutoComplPanel::OnCloseList( wxKeyEvent& key )
{
	mpInfTextEdit->GetTextWidgetWindow()->ProcessEvent( key );
	mpInfTextEdit->HoldCursor( FALSE );

	if ( mpCtxBox ) 
		mpCtxBox->SetEnabled( FALSE );

	Show(FALSE);

	wxWindow* win = mpInfTextEdit->GetTextWidgetWindow();

	if ( mpCtxBox || mpArgBox )
		((wxWindow*)GetChildren().First()->GetData())->Destroy();

	Destroy();
	
	win->SetFocus();
}

void AutoComplPanel::OnKeyReceived( wxKeyEvent& key )
{
	if ( key.m_keyCode == WXK_ESCAPE )
	{
		wxKeyEvent dummy;
		OnCloseList( dummy );
		return;
	}

	if ( !mArgHintMode && ( key.m_keyCode == WXK_RETURN || key.m_keyCode == ' ' ) )
	{
		CompleteName();
		return;
	}

	if ( !mArgHintMode && !ALPHA_NUM( key.m_keyCode ) )
	{
		if ( key.m_keyCode == '(' &&
			 MemberPresent( mpCtxBox->GetTypedName(), FALSE, SP_CTX_OPERATION ) )
		{
			CompleteName();
			return;

			/*
			if ( !TrySwitching() ) 
			{
				mpCtxBox->SetEnabled( FALSE );
				Destroy();
				return;
			}
			*/
		}
		else
		if ( key.m_keyCode != WXK_BACK )
		{
			mpInfTextEdit->GetTextWidgetWindow()
				->GetEventHandler()
					->ProcessEvent( key );

			mpCtxBox->SetEnabled( FALSE );
			Destroy();
			return;
		}
	}

	mpInfTextEdit->GetTextWidgetWindow()
		->GetEventHandler()
			->ProcessEvent( key );

	// popup-window refreshing logic

	// FIXME:: messy...

	{
		int line=0, col=0, pgLine, pgCol;
		mpInfContText->GetCursorPosition( &line, &col );
		mpInfTextEdit->GetPagePosition( &pgLine, &pgCol );

		if ( mPrevCursorLine != line || key.m_keyCode == WXK_RETURN ||
			 pgLine != mPrevPageLine ||
			 pgCol  != mPrevPageCol  )
		{
			if ( !( mIsAbove && 
					pgLine == mPrevPageLine &&
					pgCol  == mPrevPageCol )
			   )
			{
				Show( FALSE );
				RepositionPanel();
				Show( TRUE );
			}
		}

	}

	StoreScreenPos();
}

void AutoComplPanel::OnFocusKilled()
{
	Destroy();
}

void AutoComplPanel::OnSize( wxSizeEvent& event )
{
	return;  // TODO: ??? Why is this here?

	if ( !mpCtxBox ) return;

	int w = 0, h = 0;
	GetClientSize( &w, &h );

	if ( mArgHintMode )

		mpArgBox->SetSize( 0,0, w,h );
	else
		mpCtxBox->SetSize( 0,0, w,h );
}

/*** protected methods ***/

spContext* AutoComplPanel::MemberPresent( const wxString& name, bool matchExactly, int mask )
{
	wxString lowerPtn = name.Lower();

    size_t i;
	for (i = 0; i != mCtxList.Count(); ++i)
    {
		if ( mCtxList[i]->GetType() & mask )
		{
			wxString& ctxName = mCtxList[i]->mName;
			if ( matchExactly )
			{
				if ( ctxName == name )
						return mCtxList[i];
			}
			else
			if ( ctxName.Lower().StartsWith( lowerPtn ))
				return mCtxList[i];
		}
    }
	return FALSE;
}

void AutoComplPanel::FindOverloads( MMemberListT& ops, const wxString& name )
{
    size_t i;
	for (i = 0; i != mCtxList.Count(); ++i)
    {
		if ( mCtxList[i]->GetType() == SP_CTX_OPERATION && 
		     mCtxList[i]->GetName() == name ) 
        {
			ops.Add( mCtxList[i] );
        }
    }
}

void AutoComplPanel::CompleteName()
{
	spContext* member = mpCtxBox->GetMatchingContext();
	CompleteName( member, mpCtxBox->GetTypedName() );
}

void AutoComplPanel::CompleteName( spContext* member, const wxString& typedName )
{
	wxString fullName = member->GetName();

	size_t i = 0;
	size_t min = wxMin( fullName.length(), typedName.length() );

	while ( i < min && typedName[i] == fullName[i] )
		++i;

	int line = 0, col = 0;
	mpInfContText->GetCursorPosition( &line, &col );

	if ( i < typedName.length() )
	{
		// truncate unmatched trailing characters

		int dif = typedName.length() - i;

		mpInfTextEdit->DeleteRenage( line, col - dif, line, col );
		col -= dif;
	}

	wxString toInsert;

	if ( i < fullName.length() )
	{
		// append missing characters

		int dif = fullName.length() - i;

		toInsert = wxString( fullName, i, dif );
	}

	if ( member->GetType() == SP_CTX_OPERATION )
	{
		// QUCIK-N-DIRTY::  use InsertText() instead of ProcessEvent()

		wxEvtHandler* pHandler = mpInfTextEdit->GetTextWidgetWindow()->GetEventHandler();

		/*
		wxKeyEvent key( wxEVT_CHAR );
		key.m_altDown     = FALSE;key.m_controlDown = FALSE;
		key.m_metaDown    = FALSE;key.m_shiftDown   = FALSE;
		key.m_keyCode     = '(';

		pHandler->ProcessEvent( key );
		*/

		MMemberListT ops;
		FindOverloads( ops, fullName );

		// either we have overloads or method has arguments to be listed

		if ( member->GetType() == SP_CTX_OPERATION &&
		     ( ((spOperation*)member)->GetParams().Count() > 0 && ops.Count() >= 1 )
		   )
		{
			toInsert += "(";
			mpInfTextEdit->InsertText( toInsert, line, col );
			mpInfTextEdit->SetCursorPosition( line, col+toInsert.Length() );
			SwitchToHintMode( ops );
			return;
		}

		toInsert += "()";
	}

	if ( toInsert.Length() )
		mpInfTextEdit->InsertText( toInsert, line, col );

	col += toInsert.Length();
	mpInfTextEdit->SetCursorPosition( line, col );

	if ( mpCtxBox )
		mpCtxBox->SetEnabled( FALSE );

	wxKeyEvent dummy;
	OnCloseList( dummy );
}

void AutoComplPanel::RepositionPanel()
{
	wxPoint pos;
	wxSize  size = ( mArgHintMode ) ? mpArgBox->GetBoxSize()
									: _gComplPanelSize;

	if ( mArgHintMode )
	{
		size.y += 10;
		size.x += 10;
	}

	mIsAbove = AutoComplPanel::CalcPanelPosition( TRUE, size, pos, *mpDoc );

	this->SetSize( pos.x, pos.y, size.x, size.y );

	StoreScreenPos();
}

void AutoComplPanel::SwitchToHintMode( MMemberListT& ops )
{
	// FOR NOW:: FIXME:: reuse...
	wxClientDC dc(GetParent());
	dc.SetFont( GetParent()->GetFont() );
	long x,y;
	dc.GetTextExtent( "X", &x, &y );
	wxSize charDim(x,y);

	// FOR NOW:: HACK:: ...
	wxASSERT( GetParent()->IsKindOf( CLASSINFO( wxTextEditorView ) ) );
	charDim = ((wxTextEditorView*)GetParent())->GetCharacterSize();

	if ( mpCtxBox )
	{
		mpCtxBox->SetEnabled( FALSE );
		mpCtxBox->Destroy();
		mpCtxBox = NULL;
	}

	mArgHintMode = TRUE;

	int w = 0, h = 0;
	mpInfTextEdit->GetTextWidgetWindow()->GetClientSize( &w, &h );

	mpArgBox = new ArgListBox( this, ops, w, charDim );

	RepositionPanel();

	GetClientSize( &w, &h );
	mpArgBox->SetSize( 0,0, w,h );
	mpArgBox->SetFocus();

	mpInfTextEdit->HoldCursor( TRUE );
}

bool AutoComplPanel::TrySwitching()
{
	MMemberListT ops;
	FindOverloads( ops, mpCtxBox->GetTypedName() );

	if ( ops.Count() > 0 || 
		 ops[0]->GetMembers().Count() != 0 )
	{
		 SwitchToHintMode( ops );
		 return TRUE;
	}
	else
		return FALSE;
}

/***** Implementation for class AutoComplBox *****/

AutoComplBox::AutoComplBox( AutoComplPanel* panel, 
							MMemberListT&   ctxList,
							const wxString& nameTyped,
							spFile*         file,
							const wxSize& size )

	: ContextListBox( panel, -1, ctxList, nameTyped, file, size ),
	  mpPanel( panel ),
	  mEnabled( TRUE )
{}

void AutoComplBox::OnCloseList( wxKeyEvent& key ) 
{ 
	if ( mEnabled ) mpPanel->OnCloseList( key ); 
}

void AutoComplBox::OnKeyReceived( wxKeyEvent& key ) 
{ 
	if ( mEnabled ) mpPanel->OnKeyReceived( key ); 
}

void AutoComplBox::OnFocusKilled() 
{ 
	if ( mEnabled ) mpPanel->OnFocusKilled(); 
}

void AutoComplBox::SetEnabled( bool enabled ) 
{ 
	mEnabled = enabled; 
}

/***** Implementation for class ArgListBox *****/

BEGIN_EVENT_TABLE( ArgListBox, wxTextEditorView )

	EVT_CHAR      ( ArgListBox::OnChar      )
	EVT_KEY_DOWN  ( ArgListBox::OnKeyDown   )
	EVT_KILL_FOCUS( ArgListBox::OnKillFocus )

END_EVENT_TABLE()


ArgListBox::ArgListBox( AutoComplPanel* pPanel,
						MMemberListT&      mtdList,
						int                maxWidth,
						const wxSize&      charDim )

	: wxTextEditorView( pPanel, -1, new wxTextEditorModel(), 0 ),
	  mpPanel( pPanel ),
	  mName( "(" )
{
	wxString text;

	maxWidth /= charDim.x;

	int actualMaxWidth = 0;

	int curLine =  0;
	int curWidth = 0;

    size_t i;
	for (i = 0; i != mtdList.Count(); ++i)
	{
		spOperation* pOp = (spOperation*)mtdList[i];

		if ( i != 0 )
		{
			text += '\n';
			++curLine;
		}

		curWidth = 0;

		++curWidth;

		wxString sg = pOp->FormatSignature( TRUE, TRUE, TRUE, TRUE, FALSE );
		text += sg;

		/*
      size_t k;

		for (k = 0; k != pOp->GetMembers().Count(); ++k)
		{
			wxString parText;

			spParameter* pPar = (spParameter*)pOp->GetParams()[k];

			parText += pPar->mType;
			parText += ' ';
			parText += pPar->mName;
			
			if ( pPar->mInitVal.length() )
			{
				parText += " = ";
				parText += pPar->mInitVal;
			}

			if ( (int)(k > 0 && curWidth + parText.length() + 2) >= maxWidth )
			{
				actualMaxWidth = wxMax( curWidth + 1, actualMaxWidth );

				curWidth = parText.length();
				++curLine;

				text += ",\n" + parText;
			}
			else
			{
				if ( k > 0 ) 
				{
					text += ", ";
					curWidth += 2;
				}

				text += parText;
				curWidth += parText.length();
			}
		}
		*/

		actualMaxWidth = wxMax( (int)sg.Length(), actualMaxWidth );
	}

	SetMargins( 0,0, 0,0 );

/*
	SetBackgroundColour( wxColour( 224,224,224 ) );
	SetColours( wxColour( 224,224,224 ), // normalBkCol 
				wxColour(   0,  0,128 ), // selectedBkCol
				wxColour( 255,255,255 ) // selectedTextCol
			  );
			  */

	EnableScrollbars( FALSE );
	EnableCursor( FALSE );

	mBoxDim.y = ( curLine + 1 ) * charDim.y;
	mBoxDim.x = actualMaxWidth * charDim.x;

	mpModel->InsertText( TPosition(0,0), text.c_str(), text.length() );

	Activate();
}

void ArgListBox::OnChar( wxKeyEvent& event )
{
	if ( IsClipboardCmd( event ) )
	{
		event.Skip();
		return;
	}

	if ( !event.ControlDown() &&
		 !event.AltDown() )
	{
		if ( event.m_keyCode == WXK_BACK )
		{
			if ( mName.length() != 0 ) 
			
				mName.erase( mName.length() -1, 1 );

			mpPanel->OnKeyReceived( event );
		}
		else
		{
			mName += event.m_keyCode;
		}

		int level = 0;

        size_t i;
		for (i = 0; i != mName.length(); ++i)
		{
			if ( mName[i] == '(' )
                ++level;
			if ( mName[i] == ')' )
                --level;
		}

		if ( level == 0 ) 
			mpPanel->OnCloseList( event );
		else
		{
			if ( event.m_keyCode != WXK_BACK )
				mpPanel->OnKeyReceived( event );
		}
	}
	else
		mpPanel->OnCloseList( event );
}

void ArgListBox::OnKeyDown( wxKeyEvent& event )
{
	if ( IsClipboardCmd( event ) )
	{
		event.Skip();
		return;
	}

	if ( event.ControlDown() ||
		 event.AltDown() )
		 mpPanel->OnCloseList( event );
	else
	if ( event.m_keyCode == WXK_ESCAPE )
		mpPanel->OnCloseList( event );
	else
	if ( event.m_keyCode >= WXK_START && 
	     event.m_keyCode != WXK_SHIFT &&
		 event.m_keyCode != WXK_RETURN 
	   )

		mpPanel->OnCloseList( event );
}

void ArgListBox::OnKillFocus( wxFocusEvent& event )
{
	mpPanel->OnFocusKilled();
}
