/////////////////////////////////////////////////////////////////////////////
// Name:        No names yet.
// Purpose:     Contrib. demo
// Author:      Aleksandras Gluchovas
// Modified by:
// Created:     2000/02/08
// RCS-ID:      $Id: codecompl_svc.h,v 1.5 2001/11/15 06:12:59 aleksandras Exp $
// Copyright:   (c) Aleksandras Gluchovas
// Licence:   	GPL (version 2.0 or later)
/////////////////////////////////////////////////////////////////////////////

#ifndef __CODECOMPL_SVC_G__
#define __CODECOMPL_SVC_G__

#ifdef __GNUG__
    #pragma interface "codecompl_svc.h"
#endif

#include "wx/panel.h"
#include "wx/pmf/lseditor.h"
#include "wx/pmf/pm.h"
#include "wx/pmf/inf/textediting.h"

#include "parserbase.h"

class wxsTypingMonitor;
class wxsSourceEditorPlugin;
class wxPMClassFinderService;

class AutoComplBox;
class ArgListBox;
class AutoComplPanel;

class wxPMCodeCompletionService : public wxPMSimpleService,
								  protected wxPMBootstrapServiceListener,
								  protected I_KeyEventInterceptor
{
public:
	wxPMCodeCompletionService();
	~wxPMCodeCompletionService();

	// overriden methods of wxPMService

	virtual bool Start();
	virtual bool Stop();

	virtual wxPMMergableMenuChain* GetMergableMenuChain();

	// overriden method of wxPMBootstartServiceListener interface

	virtual void OnServiceStarted( wxPMService& service );

	// overriden method of I_KeyInterceptor interface

	virtual bool OnEditorKeyEvent( wxKeyEvent& key, bool isChar, wxPMService* fromSvc );

	static spContext* FindUniqueMatch( MMemberListT& inLst, wxString name );

	DECLARE_DYNAMIC_CLASS( wxPMCodeCompletionService )
	DECLARE_EVENT_TABLE()

protected:
	wxPMClassFinderService* mpClassFinderSvc;
};

class AutoComplPanel : public wxPanel
{
public:
	AutoComplPanel( wxPMDocument* doc,
				    MMemberListT& ctxList,
					const wxString& nameTyped,
					spContext* ctxAtCursor,
					bool completeNameNow,
				    const wxPoint& pos,
				    const wxSize&  size );

	~AutoComplPanel();

	static bool CalcPanelPosition( bool      above,
								   wxSize&   size,
								   wxPoint&  pos,
								   wxPMDocument& doc );

	void StoreScreenPos();

	// overriden methods of ContextListBox

	virtual void OnCloseList( wxKeyEvent& key );
	virtual void OnKeyReceived( wxKeyEvent& key );
	virtual void OnFocusKilled();

	void OnSize( wxSizeEvent& event );

	// utilities
	void CompleteName();
	void CompleteName( spContext* member, const wxString& typedName );

	DECLARE_EVENT_TABLE()

protected:
	spContext* MemberPresent( const wxString& name, bool matchExactly, int mask );
	void FindOverloads( MMemberListT& ops, const wxString& name );

	void RepositionPanel();
	void SwitchToHintMode( MMemberListT& ops );
	bool TrySwitching();

protected:
	AutoComplBox*	   mpCtxBox;
	ArgListBox*        mpArgBox;
	wxPMDocument*	   mpDoc;

	// obtained interfaces
	I_EditableText*    mpInfTextEdit;
	I_ContainingText*  mpInfContText;

	wxString		   mNameTyped;

	spContext*		   mpCtxAtCursor;
	MMemberListT       mCtxList;
	wxSize             mCharDim;

	bool               mArgHintMode; // TRUE, if argument list is being shown now,
									 // otherwise method-list pop-up

	bool               mIsAbove;
	int                mPrevCursorLine;
	int                mPrevPageLine;
	int                mPrevPageCol;
	static wxSize	   mPanelSize; // default: (270, 150)
};

#define CTX_LIST_PIN_TC (BOOKMARK_PIN_TC + 100)

class CtxListPin : public TPinBase
{
public:
	int mIconNo;

public:
	CtxListPin( size_t row, int iconNo )
		: TPinBase( CTX_LIST_PIN_TC, row ),
		  mIconNo( iconNo )
	{}

	static int GetPinTypeCode() { return CTX_LIST_PIN_TC; }
};

class TCtxListPinPainter : public TPinPainterBase
{
protected:
	wxBrush mBkBrush;
	static wxBitmap mBmps[11];
public:
	TCtxListPinPainter();

	virtual void DrawPin( TPinBase* pPin, wxTextEditorView& view, wxDC& dc, 
						  const wxPoint& pos, const wxSize& dim );

	enum { CLASS_IMAGE = 0,
		   PRIV_MTD_IMAGE,
		   PROT_MTD_IMAGE,
		   PUB_MTD_IMAGE,
		   PRIV_VAR_IMAGE,
		   PROT_VAR_IMAGE,
		   PUB_VAR_IMAGE,
		   PUB_PURE_MDT_IMAGE,
		   PRIV_MTD_DEF_IMAGE,
		   PROT_MTD_DEF_IMAGE,
		   PUB_MTD_DEF_IMAGE
		 };
};

class ContextListBox : public wxTextEditorView 
{
public:
	ContextListBox( wxWindow*     parent, 
		               wxWindowID    id,
					   MMemberListT& ctxList,
					   const wxString& nameTyped,
					   spFile*       pFile,
					   const wxSize& size );

	~ContextListBox();

	void OnChar( wxKeyEvent& event );
	void OnKeyDown( wxKeyEvent& event );
	void OnKillFocus( wxFocusEvent& event );
	void OnDblClick( wxMouseEvent& event );

	void SetPopupMode( bool on );

	wxString GetTypedName() { return mNameTyped; }
	void SetTypedName( const wxString& name ) { mNameTyped = name; }

	spContext* GetMatchingContext();

	// notifications sent to the derived class

	virtual void OnCloseList( wxKeyEvent& key )   {}
	virtual void OnKeyReceived( wxKeyEvent& key ) {}
	virtual void OnFocusKilled() {}

	DECLARE_EVENT_TABLE()

protected:
	int FindClosestMatch();

protected:
	MMemberListT mCtxList;
	wxString	 mNameTyped;
	spFile*      mpFile;
	bool         mPopupMode;
};

class AutoComplBox : public ContextListBox
{
public:
	AutoComplBox( AutoComplPanel* panel, 
				  MMemberListT&   ctxList,
				  const wxString& nameTyped,
				  spFile*         file,
				  const wxSize& size );

	virtual void OnCloseList( wxKeyEvent& key );

	virtual void OnKeyReceived( wxKeyEvent& key );

	virtual void OnFocusKilled();

	void SetEnabled( bool enabled );

protected:
	AutoComplPanel* mpPanel;
	bool     mEnabled;
};

class ArgListBox : public wxTextEditorView
{
	AutoComplPanel* mpPanel;

	wxSize mBoxDim;
	wxString mName;   // current arg.list wxString

public:
	ArgListBox( AutoComplPanel* pPanel,
			    MMemberListT&      mtdList,
			    int                maxWidth, 
			    const wxSize&      charDim );

	void OnChar( wxKeyEvent& event );
	void OnKeyDown( wxKeyEvent& event );
	void OnKillFocus( wxFocusEvent& event );

	wxSize& GetBoxSize() { return mBoxDim; }

	DECLARE_EVENT_TABLE()
};

#endif /* __CODECOMPL_SVC_G__ */

