/////////////////////////////////////////////////////////////////////////////
// Name:        No names yet.
// Purpose:     Contrib. demo
// Author:      Aleksandras Gluchovas
// Modified by:
// Created:     2001/03/23
// RCS-ID:      $Id: helpfinder_svc.cpp,v 1.15 2001/11/15 20:07:12 georgetasker Exp $
// Copyright:   (c) Aleksandras Gluchovas (alex@soften.ktu.lt)
// Licence:   	GPL (version 2.0 or later)
/////////////////////////////////////////////////////////////////////////////

#ifdef __GNUG__
    #pragma implementation "helpfinder_svc.h"
#endif

// For compilers that support precompilation, includes "wx/wx.h".
#include "wx/wxprec.h"

#ifdef __BORLANDC__
    #pragma hdrstop
#endif

#ifndef WX_PRECOMP
    #include "wx/wx.h"
#endif

#include "helpfinder_svc.h"
#include "classfinder_svc.h"
#include "parserbase.h"
#include "wx/xrc/xmlres.h"
#include "wx/mstream.h"

#include "wx/pmf/pm_flalg.h"
#include "wx/fl/controlbar.h"

#include "wx/html/htmlwin.h"
#include "wx/bmpbuttn.h"
#include "wx/icon.h"

// helper class

class HelpFinderMDIChildFrame : public wxMDIChildFrame
{
public:
	HelpFinderMDIChildFrame(wxMDIParentFrame* parent, wxWindowID id, const wxString& title, 
							const wxPoint& pos = wxDefaultPosition, const wxSize& size = wxDefaultSize, 
							long style = wxDEFAULT_FRAME_STYLE, const wxString& name = "frame")
		: wxMDIChildFrame( parent, id, title, pos, size, style, name )
	{}

	void OnActivate( wxActivateEvent& event );

	DECLARE_EVENT_TABLE()
};

BEGIN_EVENT_TABLE( HelpFinderMDIChildFrame, wxMDIChildFrame )

	EVT_ACTIVATE( HelpFinderMDIChildFrame::OnActivate )

END_EVENT_TABLE()

void HelpFinderMDIChildFrame::OnActivate( wxActivateEvent& event )
{
	if ( !wxPMFStarted() )
		return;

	wxView* curView = wxPMBootSvc().GetDocManager()->GetCurrentView();

	if ( curView )
		curView->OnActivateView( TRUE, NULL, curView );

	wxNode* node = GetChildren().First();

	if ( node )
		((wxWindow*)node->GetData())->SetFocus();
}

/***** Implemnetation for class FinderHtmlWindow *****/

class FinderHtmlWindow : public wxHtmlWindow
{
public:
	FinderHtmlWindow( wxWindow *parent, wxWindowID id = -1, 
					  const wxPoint& pos = wxDefaultPosition, 
					  const wxSize& size = wxDefaultSize, 
					  long style = wxHW_SCROLLBAR_AUTO, 
					  const wxString& name = "htmlWindow" );

	// overriden method of wxHtmlWindow
	virtual void OnLinkClicked(const wxHtmlLinkInfo& link);
};

FinderHtmlWindow::FinderHtmlWindow( wxWindow *parent, wxWindowID id, const wxPoint& pos, 
								    const wxSize& size, long style, 
								    const wxString& name )
	: wxHtmlWindow( parent, id, pos, size, style, name )
{}

void FinderHtmlWindow::OnLinkClicked(const wxHtmlLinkInfo& link)
{
	/*
	if ( mPath.Length() )
		LoadPage( mPath + wxString('#') + link.GetHref() );
	else
	{
	*/
	LoadPage( link.GetHref() );
}

/***** Implementation for class wxPMHelpFinderService *****/

static wxCommandInfo _gHelpCommands[] = {

	wxCommandInfo( wxPMHelpFinderService::ID_CONFIGURE, "Co&nfigure...", "", "" ),
	wxCommandInfo( wxPMHelpFinderService::ID_OPTIONS, "&Options...", "", "" ),
	wxCommandInfo( wxPMHelpFinderService::ID_TOGGLE_FINDER_PANE, "&Finder window", "", "Alt+3", TRUE, FALSE ),
	wxCommandInfo( -1, "SEPARATOR" ),

	wxCommandInfo( wxPMHelpFinderService::ID_CONTENT, "&Content", "", "" ),
	wxCommandInfo( -1, "SEPARATOR" ),
	
	wxCommandInfo( wxPMHelpFinderService::ID_USE_MDI_WINDOW, "&Use MDI window", "", "", TRUE, TRUE ),
	wxCommandInfo( wxPMHelpFinderService::ID_AUTO_HELP,		 "&Auto-help on source", "", "" ),

	wxCommandInfo( 0, "" ) // list-terminator
};

static wxCommandInfo _gHelpCommands1[] = {

	wxCommandInfo( wxPMHelpFinderService::ID_TOGGLE_FINDER_PANE, "&Finder window", "", "Alt+3", TRUE, FALSE ),

	wxCommandInfo( 0, "" ) // terminator
};

static wxPMMergableMenuChain _gHelpChains[] = {

	wxPMMergableMenuChain( "/&Help/", _gHelpCommands ),
	wxPMMergableMenuChain( "/&View/",  _gHelpCommands1 ),

	wxPMMergableMenuChain( "", NULL ) // list-terminator
};

IMPLEMENT_DYNAMIC_CLASS( wxPMHelpFinderService, wxPMSimpleService )

BEGIN_EVENT_TABLE( wxPMHelpFinderService, wxPMSimpleService )

	EVT_MENU( ID_CONFIGURE,			 wxPMHelpFinderService::OnConfigure )
	EVT_MENU( ID_CONTENT,			 wxPMHelpFinderService::OnContent )
	EVT_MENU( ID_OPTIONS,			 wxPMHelpFinderService::OnOptions )
	EVT_MENU( ID_FIND_HELP,			 wxPMHelpFinderService::OnFindHelp )
	EVT_MENU( ID_USE_MDI_WINDOW,	 wxPMHelpFinderService::OnUseMDIWindow )
	EVT_MENU( ID_TOGGLE_FINDER_PANE, wxPMHelpFinderService::OnToggleFinderPane )

	EVT_KEY_DOWN( wxPMHelpFinderService::OnKeyDown )
	EVT_CHAR( wxPMHelpFinderService::OnKeyDown )

END_EVENT_TABLE()

wxPMHelpFinderService::wxPMHelpFinderService()

	: mpIndexSvc( NULL ),
	  mpClassFinder( NULL ),  
      mpFinderPanel( NULL ),
      mpPane( NULL ),	  
	  mpMdiChild( NULL ),
	  mpTmpBuf( NULL ),
	  mTmpBufLen( 0 ),
	  mUseMDI( TRUE )
{}

wxPMHelpFinderService::~wxPMHelpFinderService()
{
	if ( mpFinderPanel )
	{
		mpFinderPanel->SetFinder( NULL );

		wxNode* node = mpFinderPanel->GetChildren().First();
		while( node )
		{
			((wxWindow*)node->GetData())->SetNextHandler( NULL );

			node = node->GetNext();
		}
	}
}


wxPMMergableMenuChain* wxPMHelpFinderService::GetMergableMenuChain()
{
	return _gHelpChains;
}

bool wxPMHelpFinderService::Start()
{
	mpIndexSvc = (wxPMIndexingService*)wxPMBootSvc().
					FindUniqueServiceInstance( "wxPMIndexingService" );

	if ( !mpIndexSvc || !mpIndexSvc->IsStarted() )
	{
		wxMessageBox( "WARNING:\n\n"
			          "HelpFinder-service failed locating active instance of wxPMIndexingService,\n"
					  "help will not be available" );
		return FALSE;
	}

	mpIndexSvc->RegisterIndexHandler( (I_IndexHandler*)this );

	wxPMBootSvc().AddListener( this );

	return TRUE; // started ok
}

bool wxPMHelpFinderService::Stop()
{
	return TRUE;
}

// command event handlers

void wxPMHelpFinderService::OnConfigure( wxCommandEvent& event )
{
}

void wxPMHelpFinderService::OnContent( wxCommandEvent& event )
{
	wxMessageBox( "OnContent" );
}

void wxPMHelpFinderService::OnOptions( wxCommandEvent& event )
{
}

void wxPMHelpFinderService::OnFindHelp( wxCommandEvent& event )
{
}

void wxPMHelpFinderService::OnKeyDown( wxKeyEvent& event )
{
	wxCommandEvent dummy;
	wxWindow* focus = wxWindow::FindFocus();

	if ( !event.ControlDown() && 
		 ( event.m_keyCode == 'G' ||
		   event.m_keyCode == 'g' ) && 
		 focus && !focus->IsKindOf( CLASSINFO(wxTextCtrl) ) )

		 mpFinderPanel->OnGotoAnchor( dummy );
	else
	if ( event.ControlDown() && event.m_keyCode == 'G' )
			   
		mpFinderPanel->OnGotoAnchor( dummy );
	else
	if ( !event.ControlDown() && 
		 ( event.m_keyCode == 'f' ||
		   event.m_keyCode == 'F' ) && 
		 focus && !focus->IsKindOf( CLASSINFO(wxTextCtrl) ) )

		 mpFinderPanel->FocusName();
	else
	if ( event.ControlDown() && event.m_keyCode == 'F' )
			   
		mpFinderPanel->FocusName();
	else
	if ( event.AltDown() && event.m_keyCode == WXK_LEFT )

		mpFinderPanel->GetHtmlWindow()->HistoryBack();
	else
	if ( event.AltDown() && event.m_keyCode == WXK_RIGHT )

		mpFinderPanel->GetHtmlWindow()->HistoryForward();
	else
	if ( event.m_keyCode == WXK_F1 )
	
		this->OnContent( dummy );
	else
	if ( event.m_keyCode == WXK_F5 )
	{
		wxWindow* wnd = mpFinderPanel->GetParent();
		if ( wnd->IsKindOf( CLASSINFO(wxMDIChildFrame) ) )
			((wxMDIChildFrame*)wnd)->Maximize(TRUE);
	}
	else
		event.Skip();
}

static wxPMFrameLayoutPreference _gFinderLayPref( FL_ALIGN_BOTTOM, 400,1, 400, 180, FALSE );

void wxPMHelpFinderService::OnToggleFinderPane( wxCommandEvent& event )
{
	ShowPanel( !PanelShown() );         
}

void wxPMHelpFinderService::OnUseMDIWindow( wxCommandEvent& event )
{
	if ( PanelShown() )
		ShowPanel( FALSE ); // reparenting will happen

	mUseMDI = !mUseMDI;
	ShowPanel( TRUE );

	wxMenuBar* pBar = wxPMBootSvc().GetMainWindow()->GetMenuBar() ;
	pBar->Check( event.m_extraLong, mUseMDI );
}

void wxPMHelpFinderService::OnAutohelpSource( wxCommandEvent& event )
{
	// TBD::
}

// implementation of wxPMBootstartServiceListener interface

void wxPMHelpFinderService::OnServiceStarted( wxPMService& service )
{
	wxObject* obj = service.GetWxObject();

	if ( obj->IsKindOf( CLASSINFO(wxPMDocument) ) )
	{
		wxPMDocument* doc = (wxPMDocument*)obj;

		I_EditableText* inf = (I_EditableText*)doc->GetInterface( "I_EditableText" );

		if ( !inf )
			return;

		inf->AddKeyInterceptor( this, 0 );
	}
}

// implementation of I_KeyInterceptor interface

bool wxPMHelpFinderService::OnEditorKeyEvent( wxKeyEvent& key, bool isChar, wxPMService* fromSvc )
{
	if ( key.m_keyCode == WXK_F1 ||
		 ( key.ControlDown() &&
		   key.m_keyCode == WXK_BACK ) )
	{
		if ( !PanelShown() )
			ShowPanel( TRUE );

		if ( !mpClassFinder )
		{
			mpClassFinder = (wxPMClassFinderService*)wxPMBootSvc().
								FindUniqueServiceInstance( "wxPMClassFinderService" );
			if ( !mpClassFinder )
				return FALSE;
		}

		wxPMDocument* doc = wxDynamicCast( fromSvc->GetWxObject(), wxPMDocument );
		SourceEditingContextInfo info;
		if ( !mpClassFinder->ExtractSourceContextInfo( *doc, info ) )
			return FALSE;

		info.mHelpNeeded = TRUE;
		spContext* found = mpClassFinder->DetectLastContext( info );

		wxString ctxName;
		wxString itemName;

		if ( found )
		{
			if ( found->GetType() == SP_CTX_ATTRIBUTE )
			{
				spAttribute* attr = (spAttribute*)found;

				ctxName = attr->mName;
				itemName = attr->mType;

				delete found->GetParent();
			}
			else
			if ( found->GetType() == SP_CTX_CLASS )
			{
				ctxName = found->GetName();

				delete found;
			}
			else
			if ( found->GetType() == SP_CTX_OPERATION )
			{
				itemName = found->GetName();
				//ctxName = found->GetParent()->GetName();
				//	delete found->GetParent();
				ctxName = ((spOperation*)found)->mScope;

				// FOR NOW:: FIXME::
				if ( ctxName == "wxWindowBase" )
					ctxName = "wxWindow";
			}
		}
		else
		{
			itemName = info.mLastUnresolvedName;
			ctxName = info.mLastUnrssolvedCtx;
		}

		bool exact = FALSE;

		RaisePanel();

		if ( ( ctxName.Length() || itemName.Length() ) &&
			 mpFinderPanel->Search( itemName, ctxName, &exact, TRUE ) )
		{
			//if ( !exact )
		
			//	mpFinderPanel->Search( "", ctxName, NULL, TRUE );

			mpFinderPanel->LoadItemContent();
		}
		else
			wxMessageBox( "No help found for item" );

		return TRUE;
	}

	return FALSE;
}

// implementation of I_IndexHandler interface

const char* wxPMHelpFinderService::GetIndexHandlerName()
{
	return "wxWin_html_help_handler";
}

const char* wxPMHelpFinderService::GetIndexHandlerDescription()
{
	return "wxWindows Html-help index handler";
}

bool wxPMHelpFinderService::OnClearIndexItems( int indexId )
{
	mpIndexSvc->GetItemTable().DeleteItemsByType( WX_HELP_ENTRY_ITEM, indexId );

	return TRUE;
}

bool wxPMHelpFinderService::OnBeginBuild( int indexId )
{
	mTmpBufLen = 10000;
	mpTmpBuf = new char[mTmpBufLen+1];

	return TRUE; // nothing
}

bool wxPMHelpFinderService::OnIndexFile( const wxPMFileInfo& file, int fileId, int indexId )
{
	wxFile fl( file.GetPath() );
	if ( !fl.IsOpened() )
		return FALSE;

	int len = fl.Length();

	if ( len > mTmpBufLen || !mpTmpBuf )
	{
		wxDELETEA(mpTmpBuf);
		mTmpBufLen = len;
		mpTmpBuf = new char[ mTmpBufLen + 1 ];
	}

	char* buf = mpTmpBuf;

	// this must be fast, yeahh...!
	int n = fl.Read( buf, len );

	buf[n] = '\0'; 
	
	char* cur = buf;
	char* found = strstr( buf, "<A NAME" );

	ItemTable& itemTbl = mpIndexSvc->GetItemTable();

	// those are common for all items in a given file
	itemTbl.SetFileId( fileId );
	itemTbl.SetIndexId( indexId );

	while ( found )
	{
		cur = found + 7;

		while ( *cur && *cur != '"' )
        {
			++cur;
        }

		if ( !*cur )
			break;

		++cur; // skip leading '"'

		char* start = cur;

		while ( *cur && *cur != '"' )
        {
			++cur;
        }

		if ( !*cur )
			break;

		*cur = '\0';
		wxString name = start;
		wxString hrefName = start;
		*cur = '"';

		wxString context;

		// now scan fromward to see if there's any heading ( <Hx> ... </Hx> ) ahead,
		// if so, take name from it's content, otherwise the name
		// is that found in <a NAME="..value.." >

		while ( *cur && *cur != '/' )
        {
			++cur;
        }

		while ( *cur && *cur != '<' )
        {
			++cur;
        }

		char* sectionStart = cur;

		if ( *cur && *(cur+1) == 'H' )
		{
			while ( *cur && *cur != '>' )
            {
				++cur;
            }

			if ( !*cur )
				break;

			++cur; // skip '<'

			start = cur;

			while ( *cur && *cur != '<' )
            {
				++cur;
            }

			*cur = '\0';

			wxString s= start;

			int pos = s.Find( "::" );

			*cur = '<'; // restore

			if ( pos != -1 )
			{
				context = s.Mid( 0, pos );
				name = s.Mid( pos + 2 );
			}
			else
				name = s;
		}
		else
		{
			context = name;
			name = "";
		}

		found = strstr( cur, "<A NAME" );

		int secitonSize = (int)(found - sectionStart);

		if ( found == NULL )

			secitonSize = (int)( (buf+n) - sectionStart );

		itemTbl.SetName( name.Lower() );
		itemTbl.SetContext( context.Lower() );
		itemTbl.SetLineNo( (int)( sectionStart - buf ) );
		itemTbl.SetLineCount( secitonSize );
		itemTbl.SetDataStr( hrefName + ";" + name );

		itemTbl.Add();
	}

	return TRUE;
}

bool wxPMHelpFinderService::PanelShown()
{
	if ( !mpPane && !mpMdiChild )
		return FALSE;

	if ( mUseMDI )
		return mpMdiChild->IsShown();
	else
		return mpPane->IsShown();
}

void wxPMHelpFinderService::OnDumpIndexForFile( wxTextOutputStream& stm, int fileId, int indexId )
{
	ItemTable& it = mpIndexSvc->GetItemTable();

	it.SetId(-1);
	it.SetIndexId( indexId );
	it.SetFileId( fileId );

	it.Begin( ItemTable::BY_FILE );

	while( it.Next() && it.Ok() )
	
		stm << it.GetContext() << "::" << it.GetName() << " at " << it.GetLineNo() << "\n";

	it.End();
}

bool wxPMHelpFinderService::OnEndBuild( int indexId )
{
	wxDELETEA(mpTmpBuf);

	return TRUE; // nothing
}

wxPMIndexingService* wxPMHelpFinderService::GetIndexingService()
{
	return mpIndexSvc;
}

void wxPMHelpFinderService::OnFinderPanelClosed()
{
	if ( mpFinderPanel->GetParent() == mpMdiChild )
		mpMdiChild = NULL;
	else
		mpPane = NULL;

	mpFinderPanel = NULL;
}

/* protected methods */

void wxPMHelpFinderService::CreateHelpPanel()
{
	wxWindow* parent = wxPMBootSvc().GetMainWindow();
	if (mUseMDI) 
		parent = mpMdiChild;

	/*
	wxWindow* panel = wxTheXmlResource->
		LoadWindow( parent, "HelpFinder_Panel1" );
	//mpFinderPanel = wxDynamicCast( mpFinderPanel, HelpFinderPanel );
		*/

	wxWindow* home     = new wxBitmapButton( parent, XMLID("home"), wxPMBmpMgr().FindMaskedBitmap( "browsers/16x16_home" ) );
	wxWindow* back     = new wxBitmapButton( parent, XMLID("back"), wxPMBmpMgr().FindMaskedBitmap( "browsers/16x16_back" ) );
	wxWindow* forward  = new wxBitmapButton( parent, XMLID("forward"), wxPMBmpMgr().FindMaskedBitmap( "browsers/16x16_forward" ) );
	wxWindow* search   = new wxBitmapButton( parent, XMLID("search"), wxPMBmpMgr().FindMaskedBitmap( "browsers/16x16_search" ) );
	wxWindow* gotoAnch = new wxBitmapButton( parent, XMLID("gotoAnchor"), wxPMBmpMgr().FindMaskedBitmap( "browsers/16x16_goto_anchor" ) );

	int id = home->GetId();


	home->SetToolTip( "Home (F1)" );
	back->SetToolTip( "Back (Ctr+Left Arrow)" );
	forward->SetToolTip( "Forward (Ctrl+Right Arrow)" );
	search->SetToolTip( "Search (Ctrl+F or F)" );
	gotoAnch->SetToolTip( "Goto Subsection (Ctrl+G or G)" );

	FinderHtmlWindow* htmlWin =
		new FinderHtmlWindow( parent, XMLID("htmlWindow"), wxDefaultPosition, wxDefaultSize, wxHW_SCROLLBAR_AUTO|wxSUNKEN_BORDER );

	home->SetName( "home" );
	back->SetName( "back" );
	forward->SetName( "forward" );
	search->SetName( "search" );
	gotoAnch->SetName( "gotoAnch" );
	htmlWin->SetName( "htmlWindow" );

	wxWindow* panel = wxTheXmlResource->
		LoadWindow( parent, "HelpFinder_Panel1" );
	mpFinderPanel = wxDynamicCast( panel, HelpFinderPanel );

	// connect all children to this handler intercept for "Ctrl+F" keystroke

	wxNode* node = mpFinderPanel->GetChildren().First();
	while( node )
	{
		((wxWindow*)node->GetData())->SetNextHandler( this );

		node = node->GetNext();
	}

	mpFinderPanel->mpHtmlWin = htmlWin;
	mpFinderPanel->SetFinder( this );
}

void wxPMHelpFinderService::CreatePane()
{
	if ( mpPane ) 
		return;

	bool hasFl = 
		wxPMBootSvc().FindUniqueServiceInstance( "I_PMPaneLayoutAlgorithm" ) != NULL;

	mpPane = new wxPMSimplePane( mpFinderPanel, hasFl ? &_gFinderLayPref : NULL );
	mpPane->SetPaneName( "HelpFinder Window" );

	wxPMBootSvc().AddService( mpPane );
	wxPMBootSvc().StartService( mpPane ); 
}

void wxPMHelpFinderService::CreateMdiChild()
{
	if ( mpMdiChild )
		return;
					 
	mpMdiChild = new HelpFinderMDIChildFrame( wxPMBootSvc().GetMainWindow(), -1, "HelpFinder Window" );

	// TBD:: portable frame icon
	wxIcon icon( "helpicon", wxBITMAP_TYPE_ICO_RESOURCE );

	mpMdiChild->SetIcon( icon );
}

void wxPMHelpFinderService::ShowPanel( bool show )
{
	if ( show )
	{
		if ( mUseMDI )
		{
			if ( !mpMdiChild )
				CreateMdiChild();

			if ( !mpFinderPanel ) 
				CreateHelpPanel();

			if ( mpFinderPanel->GetParent() != mpMdiChild )
				mpFinderPanel->Reparent( mpMdiChild );

			if ( mpPane && mpPane->IsShown() )
				wxPMBootSvc().HidePane( mpPane );

			if ( !mpMdiChild->IsShown() )
				mpMdiChild->Show();

			mpFinderPanel->Show();
			mpFinderPanel->Layout();
			mpMdiChild->Layout();
			mpMdiChild->Refresh();
			int w,h;
			mpMdiChild->GetClientSize( &w,&h );
			mpFinderPanel->SetSize( 0,0,w,h );
			mpMdiChild->Raise();
		}
		else // otherwise use pane docked around one of the edges of main frame
		{	 
			if ( !mpFinderPanel ) 
				CreateHelpPanel();

			if ( !mpPane )
				CreatePane();

			if ( mpFinderPanel->GetParent() != wxPMBootSvc().GetMainWindow() )
				mpFinderPanel->Reparent( wxPMBootSvc().GetMainWindow() );

			if ( !mpPane->IsShown() )
				wxPMBootSvc().ShowPane( mpPane );

			if ( mpMdiChild && mpMdiChild->IsShown() )
				mpMdiChild->Show( FALSE );
		}

		mpFinderPanel->FocusName();
	}
	else // hiding of helpfinder window
	{
		if ( mpMdiChild && mpMdiChild->IsShown() )
        {
            mpMdiChild->Show( FALSE );
        }
		else
        {
		    if ( mpPane && mpPane->IsShown() )
			    wxPMBootSvc().HidePane( mpPane );
        }
	}
}

void wxPMHelpFinderService::RaisePanel()
{
	if ( mUseMDI && mpMdiChild )
		mpMdiChild->Activate();

	if ( mpFinderPanel )
		mpFinderPanel->FocusContent();
}

/***** Implementation for class HelpFinderPanel *****/

IMPLEMENT_DYNAMIC_CLASS( HelpFinderPanel, wxPanel )

BEGIN_EVENT_TABLE( HelpFinderPanel, wxPanel )

	EVT_BUTTON( XMLID( "search" ), HelpFinderPanel::OnSearch )
	EVT_BUTTON( XMLID( "home" ), HelpFinderPanel::OnHome )
	EVT_BUTTON( XMLID( "top" ), HelpFinderPanel::OnTop )
	EVT_BUTTON( XMLID( "back" ), HelpFinderPanel::OnBack )
	EVT_BUTTON( XMLID( "forward" ), HelpFinderPanel::OnForward )
	EVT_BUTTON( XMLID( "gotoAnchor" ), HelpFinderPanel::OnGotoAnchor )

	EVT_INIT_DIALOG( HelpFinderPanel::OnInitDialog )

	EVT_TEXT_ENTER( XMLID( "name" ), HelpFinderPanel::OnEnter )
	EVT_TEXT_ENTER( XMLID( "context" ), HelpFinderPanel::OnEnter )

	EVT_SET_FOCUS( HelpFinderPanel::OnSetFocus )

END_EVENT_TABLE()

HelpFinderPanel::HelpFinderPanel()

	: mpFinder( NULL )
{}

void HelpFinderPanel::SetFinder( wxPMHelpFinderService* finderSvc )
{
	mpFinder = finderSvc;
}

void HelpFinderPanel::OnInitDialog( wxInitDialogEvent& event )
{
	wxTextCtrl* pName = (wxTextCtrl*)wxXmlFindWindow( "name", this );
	wxTextCtrl* pCtx  = (wxTextCtrl*)wxXmlFindWindow( "context", this );
	pName->SetValue("");
	pCtx->SetValue("");

	FocusName();
}

void HelpFinderPanel::OnSetFocus( wxFocusEvent& event )
{
}

void HelpFinderPanel::OnEnter( wxCommandEvent& event )
{
	OnSearch( event );
	FocusContent();
}

void HelpFinderPanel::OnSearch( wxCommandEvent& event )
{
	wxTextCtrl* name = (wxTextCtrl*)wxXmlFindWindow( "name", this );
	wxTextCtrl* ctx = (wxTextCtrl*)wxXmlFindWindow( "context", this );

	if ( Search( name->GetValue(), ctx->GetValue(), NULL ) )

		LoadItemContent();
}

void HelpFinderPanel::OnHome( wxCommandEvent& event )
{
	wxCommandEvent dummy;
	mpFinder->OnContent( dummy );
}

bool HelpFinderPanel::Search( wxString name, wxString context, bool* matchedExactly, bool setCtlValues )
{
	if ( setCtlValues )
	{
		((wxTextCtrl*)wxXmlFindWindow( "name", this ))->SetValue( name );
		((wxTextCtrl*)wxXmlFindWindow( "context", this ))->SetValue( context );
	}

	name = name.Lower();
	context = context.Lower();

	ItemTable& it = mpFinder->GetIndexingService()->GetItemTable();

	if ( it.FindByName( name, context, wxPMHelpFinderService::WX_HELP_ENTRY_ITEM, TRUE ) )
	{
		if ( matchedExactly )

			*matchedExactly = it.GetName() == name && 
							  it.GetContext() == context;

		return TRUE;
	}

	return FALSE;
}

bool HelpFinderPanel::LoadItemContent()
{
	wxPMIndexingService& isvc = *mpFinder->GetIndexingService();
	ItemTable& itbl = isvc.GetItemTable();
	FileTable& ftbl = isvc.GetFileTable();
	wxString ctx = itbl.GetContext();
	wxString name = itbl.GetName();

	if ( !ftbl.FindById( itbl.GetFileId() ) )
		return ReturnNotFound( name, ctx );

	wxString fname = ftbl.GetName();

	// remember ids
	mFileId = ftbl.GetId();
	mIndexId = ftbl.GetIndexId();

	wxFile fl( fname );
	if ( !fl.IsOpened() )
		return ReturnNotFound( name, ctx );

	int len = itbl.GetLineCount();
	int pos = itbl.GetLineNo();

	wxString anchor = itbl.GetDataStr().BeforeFirst(';');

	if ( anchor.Length() )
	{
		wxHtmlWindow* html = (wxHtmlWindow*)wxXmlFindWindow( "htmlWindow", this );
		html->LoadPage( ftbl.GetName() + wxString('#') + anchor );
	}
	else
	{
		char* buf = new char[ len + 1 ];

		if ( fl.Seek( pos ) != pos )
			return ReturnNotFound( name, ctx );

		if ( fl.Read( buf, len ) != len )
			return ReturnNotFound( name, ctx );

		buf[len] = '\0';

		SetPageContent( "<html><body>" + wxString(buf) + "</body></html>" );
		wxDELETEA(buf);
	}

	return TRUE;
}

void HelpFinderPanel::SetPageContent( const wxString& html )
{
	wxHtmlWindow* wnd = (wxHtmlWindow*)wxXmlFindWindow( "htmlWindow", this );

	wnd->SetPage( html );
}
void HelpFinderPanel::OnTop( wxCommandEvent& event )
{
}

void HelpFinderPanel::OnBack( wxCommandEvent& event )
{
	wxHtmlWindow* wnd = (wxHtmlWindow*)wxXmlFindWindow( "htmlWindow", this );
	wnd->HistoryBack();

}

void HelpFinderPanel::OnForward( wxCommandEvent& event )
{
	wxHtmlWindow* wnd = (wxHtmlWindow*)wxXmlFindWindow( "htmlWindow", this );
	wnd->HistoryForward();
}

FinderHtmlWindow* HelpFinderPanel::GetHtmlWindow()

	{ return (FinderHtmlWindow*)wxXmlFindWindow( "htmlWindow", this ); }
	
void HelpFinderPanel::OnGotoAnchor( wxCommandEvent& event )
{
	FinderHtmlWindow* wnd = (FinderHtmlWindow*)wxXmlFindWindow( "htmlWindow", this );

	if ( !wnd->GetOpenedPage().Length() )
	{
		wxMessageBox( "No page is loaded" );
		return;
	}

	wxPMIndexingService& isvc = *mpFinder->GetIndexingService();
	ItemTable& itbl = isvc.GetItemTable();
	FileTable& ftbl = isvc.GetFileTable();

	if ( !ftbl.FindByName( wnd->GetOpenedPage(), mIndexId ) )
	{
		wxMessageBox( "File " + wnd->GetOpenedPage() + " is not indexed" );
		return;
	}

	mFileId = ftbl.GetId();

	AnchorSelectionDlg* dlg = (AnchorSelectionDlg*)wxTheXmlResource->
		LoadWindow( this, "HelpFinder_SelectAnchorDlg" );

	wxListBox* box = (wxListBox*)wxXmlFindWindow( "anchorsList", dlg );

	itbl.FindByFileAndIndex( mFileId, mIndexId );
	wxStringList anchors;
	while( itbl.Ok() && 
		   mFileId == itbl.GetFileId() &&
		   mIndexId == itbl.GetIndexId() )
	{
		wxString name = itbl.GetDataStr().AfterFirst(';');
		if ( name.Length() )
		{
			box->Append( name );
			anchors.Add( itbl.GetDataStr().Before(';') );
		}

		itbl.Next();
	}

	itbl.End();

	if ( dlg->ShowModal() == wxID_OK )
	{
		wxString href = ftbl.GetName() + wxString('#') + anchors[ dlg->GetSelection() ];
		wnd->LoadPage( href );
	}

	dlg->Destroy();
}

HelpFinderPanel::~HelpFinderPanel()
{
	if ( mpFinder )
		mpFinder->OnFinderPanelClosed();
}

void HelpFinderPanel::FocusName()
{
	wxTextCtrl* name = (wxTextCtrl*)wxXmlFindWindow( "name", this );
	name->SetFocus();
	int len = name->GetValue().Length();
	name->SetInsertionPoint( len );
	name->SetSelection(0, len );
}

void HelpFinderPanel::FocusContent()
{
	wxHtmlWindow* wnd = (wxHtmlWindow*)wxXmlFindWindow( "htmlWindow", this );
	wnd->SetFocus();
}


bool HelpFinderPanel::ReturnNotFound( const wxString& name, const wxString& context )
{
	wxString s = context + "::" + name;

	SetPageContent( "<html><body>\"<b>" + s  + "</b>\" not found!</body></html>" );

	return FALSE;
}

/***** Implementation for class AnchorSelectorEntryCtrl *****/

class AnchorSelectorEntryCtrl : public wxTextCtrl
{
public:
	AnchorSelectorEntryCtrl()
		: wxTextCtrl()
	{}

	void OnKeyDown( wxKeyEvent& event );
	void OnChar( wxKeyEvent& event );

	DECLARE_EVENT_TABLE()
	DECLARE_DYNAMIC_CLASS( AnchorSelectorEntryCtrl );

protected:
};

IMPLEMENT_DYNAMIC_CLASS( AnchorSelectorEntryCtrl, wxTextCtrl )

BEGIN_EVENT_TABLE( AnchorSelectorEntryCtrl, wxTextCtrl )

	EVT_KEY_DOWN( AnchorSelectorEntryCtrl::OnKeyDown )

END_EVENT_TABLE()

void AnchorSelectorEntryCtrl::OnKeyDown( wxKeyEvent& event )
{
	AnchorSelectionDlg* dlg = (AnchorSelectionDlg*)GetParent();

	if ( event.m_keyCode == WXK_UP )
		dlg->OnMoveSelection( TRUE );
	else
	if ( event.m_keyCode == WXK_DOWN )
		dlg->OnMoveSelection( FALSE );
	else
	if ( event.m_keyCode == WXK_END )
		dlg->OnEnd( TRUE );
	else
	if ( event.m_keyCode == WXK_HOME )
		dlg->OnEnd( FALSE );
	else
		event.Skip();
}

void AnchorSelectorEntryCtrl::OnChar( wxKeyEvent& event )
{
	event.Skip();
}

/***** Implementation for class AnchorSelectionDlg *****/

IMPLEMENT_DYNAMIC_CLASS( AnchorSelectionDlg, wxDialog )

BEGIN_EVENT_TABLE( AnchorSelectionDlg, wxDialog )

	EVT_INIT_DIALOG( AnchorSelectionDlg::OnInitDialog )
	EVT_LISTBOX( XMLID( "anchorsList" ), AnchorSelectionDlg::OnListBox )
	EVT_LISTBOX_DCLICK( XMLID( "anchorsList" ), AnchorSelectionDlg::OnListDblClick )

	EVT_TEXT( XMLID( "entry" ), AnchorSelectionDlg::OnText )

END_EVENT_TABLE()

AnchorSelectionDlg::AnchorSelectionDlg()
	: wxDialog(),
	  mpList( NULL ),
	  mpEntry( NULL )
{}

void AnchorSelectionDlg::OnInitDialog( wxInitDialogEvent& event )
{
	mpEntry = (wxTextCtrl*)wxXmlFindWindow( "entry", this );
	mpList  = (wxListBox*)wxXmlFindWindow( "anchorsList", this );
	mpEntry->SetValue("");

//	wxWindow* fwd = (wxTextCtrl*)wxXmlFindWindow( "forward", this );
//	wxWindow* fwd = (wxTextCtrl*)wxXmlFindWindow( "forward", this );

}

void AnchorSelectionDlg::OnListBox( wxCommandEvent& event )
{
	SetSelectedEntry();
}

void AnchorSelectionDlg::OnListDblClick( wxCommandEvent& event )
{
	EndModal( wxID_OK );
}

void AnchorSelectionDlg::OnText( wxCommandEvent& event )
{
	GotoClosestMatch();
	event.Skip();
}

// notifications from entry control

void AnchorSelectionDlg::OnMoveSelection( bool up )
{
	int sel = mpList->GetSelection();

	if ( up && sel == 0 )
		return;

	if ( !up && sel == mpList->GetCount() - 1 )
		return;

	if ( up ) --sel;
		 else ++sel;

	mpList->SetSelection( sel );
	SetSelectedEntry();
}

void AnchorSelectionDlg::OnNameChanged()
{
}

void AnchorSelectionDlg::OnEnd( bool end )
{
	if ( mpList->GetCount() == 0 )
		return;

	if ( end ) mpList->SetSelection( mpList->GetCount() - 1 );
		  else mpList->SetSelection( 0 );

	SetSelectedEntry();
}

/*** protected methods ***/

void AnchorSelectionDlg::GotoClosestMatch()
{
	if ( !mpEntry || mpList->GetCount() == 0 )
		return;

	wxString entry = mpEntry->GetValue().Lower();

	size_t closest = 0, maxMatches = 0;

    int i;
	for (i = 0; i != mpList->GetCount(); ++i)
	{
		wxString item = mpList->GetString(i).Lower();
		
		size_t len = wxMin( item.length(), entry.length() );

		size_t n;
		for( n = 0; n != len && item[n] == entry[n] ; ++n );

		if ( n > maxMatches )
		{
			closest = i;
			maxMatches = n;
		}
	}

	mpList->SetSelection( closest );
}

void AnchorSelectionDlg::SetSelectedEntry()
{
	int len = mpList->GetStringSelection().Length();
	mpEntry->SetValue( mpList->GetStringSelection() );
	mpEntry->SetInsertionPoint( len );
	mpEntry->SetSelection( 0, len );
}

int AnchorSelectionDlg::GetSelection()
	{ return mpList->GetSelection(); }
