/////////////////////////////////////////////////////////////////////////////
// Name:        No names yet.
// Purpose:     Contrib. demo
// Author:      Aleksandras Gluchovas
// Modified by:
// Created:     2001/03/23
// RCS-ID:      $Id: helpfinder_svc.h,v 1.5 2001/11/15 06:12:59 aleksandras Exp $
// Copyright:   (c) Aleksandras Gluchovas (alex@soften.ktu.lt)
// Licence:   	GPL (version 2.0 or later)
/////////////////////////////////////////////////////////////////////////////

#ifndef __HELPFINDER_SVC_G__
#define __HELPFINDER_SVC_G__

#ifdef __GNUG__
    #pragma interface "helpfinder_svc.h"
#endif

#include "wx/pmf/pm.h"
#include "wx/list.h"
#include "wx/dynarray.h"

#include "wx/pmf/inf/textediting.h"

#include "indexing_svc.h"

class wxPMClassFinderService;
class HelpFinderPanel;
class FinderHtmlWindow;

class wxPMHelpFinderService : public wxPMSimpleService,
							  protected I_IndexHandler,
							  protected I_KeyEventInterceptor,
							  protected wxPMBootstrapServiceListener
{
public:
	wxPMHelpFinderService();
	~wxPMHelpFinderService();

	enum MY_MENU_COMMANDS { 

		ID_CONFIGURE = 100,
		ID_OPTIONS,
		ID_CONTENT,
		ID_FIND_HELP,
		ID_TOGGLE_FINDER_PANE,
		ID_USE_MDI_WINDOW,
		ID_AUTO_HELP
	};

	enum ITEM_TYPES
	{
		WX_HELP_ENTRY_ITEM = 201
	};

	// overriden methods of wxPMService

	virtual bool Start();
	virtual bool Stop();

	virtual wxPMMergableMenuChain* GetMergableMenuChain();

	// event handlers

	void OnConfigure( wxCommandEvent& event );
	void OnOptions( wxCommandEvent& event );
	void OnFindHelp( wxCommandEvent& event );
	void OnToggleFinderPane( wxCommandEvent& event );
	void OnUseMDIWindow( wxCommandEvent& event );
	void OnAutohelpSource( wxCommandEvent& event );
	void OnContent( wxCommandEvent& event );

	void OnKeyDown( wxKeyEvent& event );

	// overriden method of wxPMBootstartServiceListener interface

	virtual void OnServiceStarted( wxPMService& service );

	// overriden method of I_KeyInterceptor interface

	virtual bool OnEditorKeyEvent( wxKeyEvent& key, bool isChar, wxPMService* fromSvc );

	// implementation of I_IndexHandler interface

	virtual const char* GetIndexHandlerName();
	virtual const char* GetIndexHandlerDescription();

	virtual bool OnClearIndexItems( int indexId );
	virtual bool OnBeginBuild( int indexId );
	virtual bool OnIndexFile( const wxPMFileInfo& file, int fileId, int indexId );
	virtual void OnDumpIndexForFile( wxTextOutputStream& stm, int fileId, int indexId );
	virtual bool OnEndBuild( int indexId );

	// notification received from finder-panel
	void OnFinderPanelClosed();

	// other methods

	wxPMIndexingService* GetIndexingService();

	DECLARE_EVENT_TABLE()
	DECLARE_DYNAMIC_CLASS( wxPMHelpFinderService )

protected:
	void CreateHelpPanel();
	void CreatePane();
	void CreateMdiChild();
	void ShowPanel( bool show );
	void RaisePanel();
	bool PanelShown();

protected:

	wxPMIndexingService* mpIndexSvc;
	wxPMClassFinderService* mpClassFinder;
	HelpFinderPanel* mpFinderPanel;

	// two possible containers for mpFinderPanel :
	wxPMSimplePane* mpPane;
	wxMDIChildFrame* mpMdiChild;

	char* mpTmpBuf;
	int mTmpBufLen;
	bool mUseMDI;
};

class HelpFinderPanel : public wxPanel
{
public:
	HelpFinderPanel();
	~HelpFinderPanel();


	void SetFinder( wxPMHelpFinderService* finderSvc );

	void OnInitDialog( wxInitDialogEvent& event );
	void OnSetFocus( wxFocusEvent& event );
	void OnKeyDown( wxKeyEvent& event );

	// command events
	void OnEnter( wxCommandEvent& event );
	void OnSearch( wxCommandEvent& event );
	void OnHome( wxCommandEvent& event );
	void OnTop( wxCommandEvent& event );
	void OnBack( wxCommandEvent& event );
	void OnForward( wxCommandEvent& event );
	void OnGotoAnchor( wxCommandEvent& event );

	bool LoadItemContent();
	void SetPageContent( const wxString& html );

	void FocusName();
	void FocusContent();

	FinderHtmlWindow* GetHtmlWindow();

	bool Search( wxString name, wxString context, bool* matchedExactly, bool setCtlValues = FALSE );

	DECLARE_EVENT_TABLE()
	DECLARE_DYNAMIC_CLASS( HelpFinderPanel )

	bool ReturnNotFound( const wxString& name, const wxString& context );

	FinderHtmlWindow* mpHtmlWin;
protected:
	wxPMHelpFinderService* mpFinder;
	int mFileId;
	int mIndexId;
};

class AnchorSelectionDlg : public wxDialog
{
public:
	AnchorSelectionDlg();

	void OnInitDialog( wxInitDialogEvent& event );
	void OnListBox( wxCommandEvent& event );
	void OnListDblClick( wxCommandEvent& event );
	void OnText( wxCommandEvent& event );

	// notifications received from name-entry control
	void OnMoveSelection( bool up );
	void OnNameChanged();
	void OnEnd( bool end );

	int GetSelection();

	DECLARE_EVENT_TABLE()
	DECLARE_DYNAMIC_CLASS( AnchorSelectionDlg );

protected:
	void GotoClosestMatch();
	void SetSelectedEntry();

protected:
	wxListBox* mpList;
	wxTextCtrl* mpEntry;
};

#endif /* __HELPFINDER_SVC_G__ */
