/////////////////////////////////////////////////////////////////////////////
// Name:        No names yet.
// Purpose:     Contrib. demo
// Author:      Aleksandras Gluchovas
// Modified by:
// Created:     2001/03/28
// RCS-ID:      $Id: indexing_svc.cpp,v 1.24 2001/11/20 22:57:40 hansvl Exp $
// Copyright:   (c) Aleksandras Gluchovas (alex@soften.ktu.lt)
// Licence:   	GPL (version 2.0 or later)
/////////////////////////////////////////////////////////////////////////////

#ifdef __GNUG__
    #pragma implementation "indexing_svc.h"
#endif

// For compilers that support precompilation, includes "wx/wx.h".
#include "wx/wxprec.h"

#ifdef __BORLANDC__
    #pragma hdrstop
#endif

#ifndef WX_PRECOMP
    #include "wx/wx.h"
#endif

#include "indexing_svc.h"

// include source generated by DiamondBase's dsc.exe
#include "dbsrc/Idxrel.h"

#include "wx/log.h"
#include "wx/tokenzr.h"
#include "wx/dirdlg.h"
#include "wx/filedlg.h"
#include "wx/choicdlg.h"
#include "wx/xrc/xmlres.h"

/***** Implementation for class I_IndexHandler *****/

bool I_IndexHandler::DocumentMonitoringNeeded()
	{ return FALSE; }

bool I_IndexHandler::CanAcceptFile( const wxPMFileInfo& file, int indexId )
	{ return FALSE; }

void I_IndexHandler::OnDocumentsSwitched( wxPMDocument* fromDoc, wxPMDocument* toDoc, int fileId, int indexId )
{}

void I_IndexHandler::OnDocumentSaved( wxPMDocument& doc, int fileId, int indexId )
{}

void I_IndexHandler::OnDocumentChangesDiscarded( wxPMDocument& doc, int fileId, int indexId )
{}

/***** Implementation for class wxPMIndexingService *****/

static wxCommandInfo _gIndexCommands[] = {

	wxCommandInfo( wxPMIndexingService::ID_INDECIES,             "&Indecies...", "", "" ),
	wxCommandInfo( wxPMIndexingService::ID_REBUILD_ALL_INDECIES, "&Rebuild all Indecies", "", "Ctrl+F5" ),
	wxCommandInfo( wxPMIndexingService::ID_CREATE_TEST_DATA,     "&Create Test Data", "", "" ),
	wxCommandInfo( -1, "SEPARATOR" ),

	wxCommandInfo( wxPMIndexingService::ID_DUMP_INDEX,     "&Dump Index...", "", "" ),
	wxCommandInfo( wxPMIndexingService::ID_SYNC_INDECIES,  "&Synchornize Indecies", "", "" ),

	wxCommandInfo( 0, "" ) // list-terminator
};

static wxPMMergableMenuChain _gIndexChains[] = {

	wxPMMergableMenuChain( "/&Tools/", _gIndexCommands ),

	wxPMMergableMenuChain( "", NULL ) // list-terminator
};

IMPLEMENT_DYNAMIC_CLASS( wxPMIndexingService, wxPMSimpleService )

BEGIN_EVENT_TABLE( wxPMIndexingService, wxPMSimpleService )

	EVT_MENU( ID_INDECIES,				wxPMIndexingService::OnIndecies ) 
	EVT_MENU( ID_REBUILD_ALL_INDECIES,  wxPMIndexingService::OnRebuildAllIndecies ) 
	EVT_MENU( ID_CREATE_TEST_DATA,		wxPMIndexingService::OnCreateTestData ) 
	EVT_MENU( ID_DUMP_INDEX,			wxPMIndexingService::OnDumpIndex ) 
	EVT_MENU( ID_SYNC_INDECIES,			wxPMIndexingService::OnSyncIndecies ) 

END_EVENT_TABLE()

wxPMIndexingService::wxPMIndexingService()
	: mpIndexTbl( NULL ),
	  mpFileTbl( NULL ),
	  mpItemTbl( NULL ),
	  mFilesProcessed( 0 )
{}

wxPMIndexingService::~wxPMIndexingService()
{
	CleanUp();
}

bool wxPMIndexingService::Start()
{
	bool ok = theDiamondBase.loadConfig("dbconfig.txt") == 0; // where to find the db config file(s).

	if ( ok )
	{
		mpIndexTbl = new IndexTable();
		mpIndexTbl->Setup( new Index(), this );

		ok = mpIndexTbl->Begin();

		if ( ok )
			mpIndexTbl->End();
	}

	if ( !ok )
	{
		wxMessageBox( "WARNING:\n\n"
					  "Failed loading indexing-db configuration.\n"
					  "If database-files are not present in ./db/ directory,\n"
					  "use provided script to generate them,\n"
					  "(binary db files are not part of the source-code distribution)" );

		return FALSE;

	}

	mpFileTbl = new FileTable();
	mpItemTbl = new ItemTable();

	mpFileTbl->Setup( new File(), this );
	mpItemTbl->Setup( new Item(), this );

	wxPMBootSvc().GetDocManager()->AddListener( this );
	wxPMBootSvc().RegisterUniqueServiceInstance( this, "wxPMIndexingService" );

	return TRUE;
}

bool wxPMIndexingService::Stop()
{
	CleanUp();

	wxPMBootSvc().UnregisterUniqueServiceInstance( this, "wxPMIndexingService" );

	return TRUE;
}

wxPMMergableMenuChain* wxPMIndexingService::GetMergableMenuChain()
{
	return _gIndexChains;
}

void* wxPMIndexingService::GetInterface( const char* infName, const char* version )
{
	return NULL;
}

// implementation of wxPMDocumentListener interface

void wxPMIndexingService::OnDocumentSaved( wxPMDocument& doc )
{
	NotifyHandlers( DOC_SAVED, &doc, NULL );
}

void wxPMIndexingService::OnDocumentChangesDiscarded( wxPMDocument& doc )
{
	NotifyHandlers( DOC_CHANGES_DISCARED, &doc, NULL );
}

void wxPMIndexingService::OnDocumentsSwitched( wxPMDocument* fromDoc, wxPMDocument* toDoc )
{
	if ( fromDoc && fromDoc->IsModified() )

		NotifyHandlers( DOC_SWITCHED, fromDoc, toDoc );
}

void wxPMIndexingService::OnWorkplaceOpened( wxPMWorkplace& workplace )
{
	workplace.AddListener( this );
}

void wxPMIndexingService::OnDocumentOpened( wxPMDocument& doc, bool created )
{
	if ( doc.IsKindOf( CLASSINFO(wxPMProject) ) )
	{
		// upon opening of project, we check table of indexes to
		// is if there are any bound (by name) to the opened project.
		// Then we store the all encountered index-to-project bindings in a 
		// association object, which is in turn held by bootstrap service
		// (the latter svc. removes the hassle of managing associations)

		if ( !mpIndexTbl->FindIndexByProject( doc.GetTitle() ) )
			return;

		IdxHandlerToPrjBindInfo* bindInfo = 
			(IdxHandlerToPrjBindInfo*)wxPMBootSvc().GetAssociatedData( *this, doc );

		if ( !bindInfo )
		{
			bindInfo = new IdxHandlerToPrjBindInfo();
			wxPMBootSvc().AssociateData( *this, doc, bindInfo );
		}

		// go through all indexes bound to the project

		while( mpIndexTbl->Ok() && mpIndexTbl->GetBoundProject() == doc.GetTitle() )
		{
			I_IndexHandler* handler = FindHandlerByName( mpIndexTbl->GetHandlerName() );

			if ( !handler )
			{
				wxLogMessage( "Handler \"%s\" refered in index \"%s\" is not registered",
							  mpIndexTbl->GetHandlerName(), mpIndexTbl->GetName() );
				return;
			}

			if ( handler->DocumentMonitoringNeeded() )
			{
				bindInfo->mIndexIds.Add( mpIndexTbl->GetId() );
				bindInfo->mHandlers.Add( handler );
			}

			mpIndexTbl->Next();
		}

		mpIndexTbl->End();
	}
}

// implementation of wxPMWorkplaceListener interface

void wxPMIndexingService::OnFileAdded( wxPMFileInfo& file, wxPMProject* toProject )
{
	IdxHandlerToPrjBindInfo* bindInfo = 
		(IdxHandlerToPrjBindInfo*)wxPMBootSvc().GetAssociatedData( *this, *toProject );

	if ( !bindInfo )
		return;

	for( size_t i = 0; i != bindInfo->mIndexIds.Count(); ++i )
	{
		I_IndexHandler* hnd = bindInfo->mHandlers[i];
		int indexId = bindInfo->mIndexIds[i];

		if ( hnd->CanAcceptFile( file, indexId ) )
		{
			hnd->OnBeginBuild( indexId );
			BuildIndexForFile( file.GetPath(), hnd );
			hnd->OnEndBuild( indexId );
		}
	}
}

void wxPMIndexingService::OnFileRemoved( wxPMFileInfo& file, wxPMProject* fromProject  )
{
	IdxHandlerToPrjBindInfo* bindInfo = 
		(IdxHandlerToPrjBindInfo*)wxPMBootSvc().GetAssociatedData( *this, *fromProject );

	if ( !bindInfo )
		return;

	for( size_t i = 0; i != bindInfo->mIndexIds.Count(); ++i )
	{
		I_IndexHandler* hnd = bindInfo->mHandlers[i];
		int idxId = bindInfo->mIndexIds[i];

		if ( hnd->CanAcceptFile( file, idxId ) )
		
			RemoveIndexForFile( file.GetPath(), hnd );
	}
}

// command handlers

void wxPMIndexingService::OnIndecies( wxCommandEvent& event )
{
	IndeciesDlg* dlg = wxDynamicCast( 
			wxTheXmlResource->LoadWindow( "IndexingSvc_IndeciesDlg", FALSE ),
			IndeciesDlg );

	dlg->Setup( this );
	dlg->ShowModal();
	dlg->Destroy();

	wxPMBootSvc().GetMainWindow()->SetFocus();
}

void wxPMIndexingService::RebuildIndecies( wxArrayPtrVoid& indexIds )
{
	mFilesProcessed = 0;
	int idxCnt = 0;

    unsigned int i;
	for (i = 0; i != indexIds.Count(); ++i)
	{
		mpIndexTbl->FindById( (int)indexIds[i] );
		mpIndexTbl->End();

		wxString s = "Rebuilding index " + wxString( mpIndexTbl->GetName() ) + " ...";
		wxPMBootSvc().SetStatusText( s );

		if ( !RebuildCurrentIndex() )
		{
			wxPMBootSvc().SetStatusText( "Failed building index " + 
										 wxString( mpIndexTbl->GetName() ) );
			FlushAll();
			mpIndexTbl->End();
			return;
		}

		++idxCnt;
	}

	FlushAll();

	wxString s; 
	s.Printf( "Finished rebuilding all indexes! %d file(s), %d index(es) processed", 
			  mFilesProcessed, idxCnt );
	wxPMBootSvc().SetStatusText( s );
}

void wxPMIndexingService::RebuildIndex( int indexId )
{
	wxArrayPtrVoid indexIds;
	indexIds.Add( (void*)indexId );
	RebuildIndecies( indexIds );
}

void wxPMIndexingService::OnRebuildAllIndecies( wxCommandEvent& event )
{
	mpIndexTbl->Begin(); 
	mpIndexTbl->First();
	wxArrayPtrVoid indexIds;

	while( mpIndexTbl->Ok() )
	{
		indexIds.Add( (void*)mpIndexTbl->GetId() );

		mpIndexTbl->Next();
	}

	RebuildIndecies( indexIds );
}

void wxPMIndexingService::OnCreateTestData( wxCommandEvent& event )
{
	CreateTestData();
}

void wxPMIndexingService::OnDumpIndex( wxCommandEvent& event )
{
	const char* fname = "index_dump.txt";
	wxFFileOutputStream fs( fname );
	wxTextOutputStream stm( fs );

	wxStringList lst;
	mpIndexTbl->Begin();
	wxArrayInt ids;

	while( mpIndexTbl->Next(), mpIndexTbl->Ok() )
	{
		lst.Add( mpIndexTbl->GetName() );
		ids.Add( mpIndexTbl->GetId() );
	}

	mpIndexTbl->End();

	wxString* arr = new wxString[ lst.Number() ];
	for( int i = 0; i != lst.Number(); ++i )
		arr[i] = lst[i];

	wxSingleChoiceDialog dlg( wxPMBootSvc().GetMainWindow(), "Select Index",
							  "Dump Index", lst.Number(), arr );

	if ( dlg.ShowModal() == wxID_OK )
	{
		delete [] arr;

		int indexId = ids[ dlg.GetSelection() ];
		mpIndexTbl->FindById( indexId );

		stm << ">>>> Dumping index " << mpIndexTbl->GetName() << " <<<<\n\n";

		I_IndexHandler* hnd = FindHandlerByName( mpIndexTbl->GetHandlerName() );

		if ( !hnd )
		{
			mpIndexTbl->End();
			wxMessageBox( "Handler " + wxString(mpIndexTbl->GetHandlerName()) + " is not registered" );
			return;
		}

		if ( mpFileTbl->FindByIndex( indexId ) )
		{
			while( mpFileTbl->Ok() )
			{
				wxString fname = mpFileTbl->GetName();
				stm << "*** file " << fname << "\n\n";
				hnd->OnDumpIndexForFile( stm, mpFileTbl->GetId(), indexId );
				stm << "\n";
				mpFileTbl->Next();
			}

			stm << "\n\n<<<< end of \"" << mpIndexTbl->GetName() << "\" dump <<<<";
		}

		mpFileTbl->End();
		mpIndexTbl->End();

		wxPMBootSvc().SetStatusText( "Dumped index \"" + wxString(mpIndexTbl->GetName()) +
									 "\" to file \"" + wxString(fname) +"\"" );
	}
	else
		delete [] arr;

	mpIndexTbl->End();
}

void wxPMIndexingService::OnSyncIndecies( wxCommandEvent& event )
{
	wxNode* node = wxPMBootSvc().GetDocManager()->GetDocuments().First();

	while( node )
	{
		wxPMDocument* doc = (wxPMDocument*)node->GetData();

		wxPMDocument* parent = doc->GetParentDocument();

		wxPMFileInfo* file = doc->GetFileInfo();

		if ( doc->IsModified() && file && 
			 parent && parent->IsKindOf( CLASSINFO(wxPMProject) ) )
		{
			IdxHandlerToPrjBindInfo* info = 
				(IdxHandlerToPrjBindInfo*)wxPMBootSvc().GetAssociatedData( *this, *parent );

			if ( info )
			{
				int fileId = -1; // will be obtained on-demand

                unsigned int i;
				for (i = 0; i != info->mHandlers.Count(); ++i)
				{
					I_IndexHandler* hnd = info->mHandlers[i];
					int indexId = info->mIndexIds[i];

					if ( hnd->CanAcceptFile( *file, indexId ) )
					{
						if ( fileId == -1 )
							fileId = FindFileByPath( file->GetPath(), indexId );

						wxASSERT( fileId != -1 ); // DBG::
						
						hnd->OnDocumentsSwitched( doc, NULL, fileId, indexId );
					}
				}
			}
		}

		node = node->GetNext();
	}
}

// other methods

bool wxPMIndexingService::RegisterIndexHandler( I_IndexHandler* handler )
{
	if ( mHandlers.Index( handler ) != wxNOT_FOUND )
		return FALSE;

	mHandlers.Add( handler );

	wxASSERT(mHandlers.Index( handler ) != wxNOT_FOUND); // DBG::Make sure it is done!
	return TRUE;
}

bool wxPMIndexingService::UnregisterIndexHandler( I_IndexHandler* handler )
{
	if ( mHandlers.Index( handler ) == wxNOT_FOUND )
		return FALSE;

	mHandlers.Remove( handler );

	wxASSERT(mHandlers.Index( handler ) == wxNOT_FOUND); // DBG::Make sure it is done!
	return TRUE;
}

IndexHandlerListT& wxPMIndexingService::GetIndexHandlers()
{
	return mHandlers;
}

bool wxPMIndexingService::RemoveCurrentIndex()
{
	if ( !mpIndexTbl->Ok() )
		return FALSE;

	I_IndexHandler* handler = FindHandlerByName( mpIndexTbl->GetHandlerName() );

	if ( !handler )
		return FALSE;

	mpFileTbl->DeleteItemsByIndex( mpIndexTbl->GetId() );
	handler->OnClearIndexItems( mpIndexTbl->GetId() );

	bool ok = mpIndexTbl->Delete(); // remove current entry

	mpIndexTbl->End();
	return ok;
}

I_IndexHandler* wxPMIndexingService::FindHandlerByName( const char* name )
{
	for( size_t i = 0; i != mHandlers.Count(); ++i )
	
		if ( strcmp( mHandlers[i]->GetIndexHandlerName(), name ) == 0 )

			return mHandlers[i];
	
	return NULL;
}

int wxPMIndexingService::FindFileByPath( const wxString& path, int indexId )
{
	if ( mpFileTbl->FindByName( path, indexId ) )

		return mpFileTbl->GetId();
	else
		return -1;
}

bool wxPMIndexingService::RebuildCurrentIndex()
{
	I_IndexHandler* handler = FindHandlerByName( mpIndexTbl->GetHandlerName() );

	if ( !handler )
		return FALSE;

	wxString savedCwd = ::wxGetCwd();
	wxString rootPath = mpIndexTbl->GetContentRoot();

	if ( rootPath.Length() && !::wxSetWorkingDirectory( rootPath ) )
		return FALSE;

	mpFileTbl->DeleteItemsByIndex( mpIndexTbl->GetId() );

	handler->OnClearIndexItems( mpIndexTbl->GetId() );
	handler->OnBeginBuild( mpIndexTbl->GetId() );
	
	wxString boundPrj = mpIndexTbl->GetBoundProject();

	if ( boundPrj.Length() )
	{
		wxPMProject* prj = wxPMBootSvc().GetActiveProject();

		if ( prj->GetTitle() != boundPrj )
		{
			// TBD:: find the project file
            int stopCompilerWarning = 0;
			wxASSERT(stopCompilerWarning); // DBG::
		}

		PMFileInfoListT& lst = prj->GetFiles();

		for( size_t i = 0; i != lst.Count(); ++i )
		
			if ( handler->CanAcceptFile( *lst[i], mpIndexTbl->GetId() ) )
			{
				BuildIndexForFile( lst[i]->GetPath(), handler );
				::wxYield();
			}
	}
	else
	{
		wxStringTokenizer tok( mpIndexTbl->GetPaths(), ";" );

		while( tok.HasMoreTokens() )
		{
			wxString path = tok.GetNextToken();
			
			if ( path.Find( '*' ) ) // FIMXE:: for some cases wxIsWild() doesn't work
			{
				wxString po = wxPathOnly( path );
				wxString absPath = wxPMConcatPaths( rootPath, po );

				// build files for each wildcard in the list
				wxStringTokenizer tok1( path.Mid( po.Length()+1 ), "," );

				while( tok1.HasMoreTokens() )
				
					BuildIndexRecursively( absPath, handler, tok1.GetNextToken() );
			}
			else
				BuildIndexForFile( wxPMConcatPaths( rootPath, path ), handler );

			::wxYield();
		}
	}

	::wxSetWorkingDirectory( savedCwd );

	handler->OnEndBuild( mpIndexTbl->GetId() );

	return TRUE;
}

void wxPMIndexingService::TestIt()
{
	Index entry;

	entry.begin( Index::dbIdx_by_name );

	for ( entry.first(); entry.ok(); entry.next() )
	{
	}

	entry.end();
}

void wxPMIndexingService::CreateTestData()
{
	//DropAllData();

#if 1
	mpIndexTbl->SetName( "Sample" );
	mpIndexTbl->SetLanguage( "C++" );
	mpIndexTbl->SetHandlerName( "C++-Class-Info-Handler" );
	mpIndexTbl->SetBoundProject( "" );
	mpIndexTbl->SetContentRoot( "D:\\sample\\" );
	mpIndexTbl->SetPaths( "*.h;*.cpp" );
#endif

	mpIndexTbl->Add();
}

void wxPMIndexingService::DropAllData()
{
	mpIndexTbl->RemoveAll();
	mpFileTbl->RemoveAll();
	mpItemTbl->RemoveAll();
}

IndexTable& wxPMIndexingService::GetIndexTable()
{
	return *mpIndexTbl;
}

FileTable& wxPMIndexingService::GetFileTable()
{
	return *mpFileTbl;
}

ItemTable& wxPMIndexingService::GetItemTable()
{
	return *mpItemTbl;
}

void wxPMIndexingService::FlushAll()
{
	mpIndexTbl->Flush();
	mpFileTbl->Flush();
	mpItemTbl->Flush();
}

/*** protected methods ***/

void wxPMIndexingService::BuildIndexForFile( const wxString& file, I_IndexHandler* handler )
{
	wxString name = file.Mid( wxPathOnly( file ).Length()+1 );
	wxPMBootSvc().SetStatusText( "Indexing file " + name + " ..." );
	++mFilesProcessed;

	// FIXME:: is this the right thing?
	wxString fsName = file;
	fsName.Replace( "\\", "/" );

	mpFileTbl->Begin();
	mpFileTbl->SetName( fsName );
	mpFileTbl->SetIndexId( mpIndexTbl->GetId() );
	mpFileTbl->SetLastIndexingDate( wxDateTime::Now() );
	mpFileTbl->Add();
	mpFileTbl->End();

#if 1
	handler->OnIndexFile( wxPMFileInfo( file ), 
						  mpFileTbl->GetId(), mpIndexTbl->GetId() );
#else
	wxLogMessage( "DBG:: Building index for file " + file );
#endif
}

void wxPMIndexingService::RemoveIndexForFile( const wxString& file, I_IndexHandler* handler )
{
	if ( !mpFileTbl->FindByName( file, mpIndexTbl->GetId() ) )
		return;

	mpItemTbl->DeleteItemsByFile( mpFileTbl->GetId(), mpFileTbl->GetIndexId() );

	mpFileTbl->Delete();
	mpFileTbl->End();
}

void wxPMIndexingService::GetDirEntries( wxStringList& toLst, const wxString& mask, int flags )
{
	wxString path = ::wxFindFirstFile( mask, flags );

	toLst.Clear();

	while ( path.Length() )
	{
		toLst.Add( path );
		path = ::wxFindNextFile();
	}
}

void wxPMIndexingService::BuildIndexRecursively( const wxString& dir, I_IndexHandler* handler, 
												 const wxString& fileMask )
{
	if ( !::wxSetWorkingDirectory( dir ) )
		return;

	wxStringList lst; 
	GetDirEntries( lst, "*", wxDIR );
	
	size_t i = 0;
	for ( ; i != lst.GetCount(); ++i )
	{
		wxString path = lst[i];

		if ( path != "." && path != ".." )

			BuildIndexRecursively( wxPMConcatPaths( dir, path ), handler, fileMask );
	}

	// index files matching the mask in current dir

	::wxSetWorkingDirectory( dir );

	GetDirEntries( lst, fileMask, wxFILE );

	for( i = 0; i != lst.GetCount(); ++i )
	{
		BuildIndexForFile( wxPMConcatPaths( dir, lst[i] ), handler );
		::wxYield();
	}
}

void wxPMIndexingService::CleanUp()
{
    wxDELETE(mpIndexTbl);
	wxDELETE(mpFileTbl);
	wxDELETE(mpItemTbl);
}

void wxPMIndexingService::NotifyHandlers( int notifyType, wxPMDocument* doc1, wxPMDocument* doc2 )
{
	if ( !doc1->CanHaveFile() )
		return;

	wxPMFileInfo* file = doc1->GetFileInfo();

	if ( !file )
		return;

	wxPMDocument* prj = doc1->GetParentDocument();

	if ( !prj || !prj->IsKindOf( CLASSINFO(wxPMProject) ) )
		return;

	IdxHandlerToPrjBindInfo* bindInfo = 
		(IdxHandlerToPrjBindInfo*)wxPMBootSvc().GetAssociatedData( *this, *prj );

	if ( !bindInfo )
		return;

	for( size_t i = 0; i != bindInfo->mIndexIds.Count(); ++i )
	{
		I_IndexHandler* hnd = bindInfo->mHandlers[i];
		int indexId = bindInfo->mIndexIds[i];
		int fileId = FindFileByPath( wxPMToUnixPath( file->GetPath() ), indexId );

		if ( hnd->CanAcceptFile( *file, indexId ) )
		{
			switch( notifyType )
			{
				case DOC_SWITCHED : 
					hnd->OnDocumentsSwitched( doc1, doc2, fileId, indexId ); break;

				case DOC_SAVED    : 
					hnd->OnDocumentSaved( *doc1, fileId, indexId ); break;

				case DOC_CHANGES_DISCARED : 
					hnd->OnDocumentChangesDiscarded( *doc1, fileId, indexId ); break;

				default: break;
			}
		}
	}
}

/***** Implementation for class IndeciesDlg *****/

IMPLEMENT_DYNAMIC_CLASS( IndeciesDlg, wxDialog )

BEGIN_EVENT_TABLE( IndeciesDlg, wxDialog )

	EVT_BUTTON( XMLID("new"),			IndeciesDlg::OnNew )
	EVT_BUTTON( XMLID("edit"),			IndeciesDlg::OnEdit )
	EVT_BUTTON( XMLID("rebuild"),		IndeciesDlg::OnRebuild )
	EVT_BUTTON( XMLID("rebuildAll"),	IndeciesDlg::OnRebuildAll )
	EVT_BUTTON( XMLID("remove"),		IndeciesDlg::OnRemove )

	EVT_BUTTON( XMLID("addSingle"),		IndeciesDlg::OnAddSingle )
	EVT_BUTTON( XMLID("addAll"),		IndeciesDlg::OnAddAll )
	EVT_BUTTON( XMLID("removeSingle"),	IndeciesDlg::OnRemoveSingle )
	EVT_BUTTON( XMLID("removeAll"),		IndeciesDlg::OnRemoveAll )

	EVT_BUTTON( XMLID("up"),   IndeciesDlg::OnUp )
	EVT_BUTTON( XMLID("down"), IndeciesDlg::OnDown )

	EVT_INIT_DIALOG( IndeciesDlg::OnInitDialog )

END_EVENT_TABLE()

IndeciesDlg::IndeciesDlg()
{
}

void IndeciesDlg::Setup( wxPMIndexingService* svc )
{
	mpSvc = svc;

	mpLstAll = (wxListCtrl*)wxXmlFindWindow( "lstAll", this );
	mpLstPrj = (wxListCtrl*)wxXmlFindWindow( "lstProject", this );

	mpLstAll->InsertColumn(0, "Index", wxLIST_FORMAT_LEFT, 80 );
	mpLstAll->InsertColumn(1, "Handler", wxLIST_FORMAT_LEFT, 150 );
	mpLstAll->InsertColumn(2, "Bound-project", wxLIST_FORMAT_LEFT, 70 );

	mpLstPrj->InsertColumn(0, "Index", wxLIST_FORMAT_LEFT, 200 );

	ReloadAllIndexList();
	ReloadProjectIndexList();
}

void IndeciesDlg::OnNew( wxCommandEvent& event )
{
	IndexPropsDlg* dlg = wxDynamicCast( 
			wxTheXmlResource->LoadWindow( "IndexingSvc_IndexPropsDlg", FALSE ),
			IndexPropsDlg );

	IndexTable& it = mpSvc->GetIndexTable();

	it.SetName( "unnamed" );
	it.SetHandlerName( "" );
	it.SetContentRoot( "" );
	it.SetPaths( "" );

	dlg->Setup( mpSvc );

	if ( dlg->ShowModal() == wxID_OK )
	{
		it.SetId( -1 );

		if ( !it.Add() )
		{
			wxMessageBox( wxString("Failed adding index ") +  it.GetName() );
		}
		else
		{
			it.End();
			mpSvc->RebuildIndex( it.GetId() );
			ReloadAllIndexList();
		}

		// TBD::
	}

	it.End();
	dlg->Destroy(); // FIXME:: why?
	this->SetFocus();
}

void IndeciesDlg::OnEdit( wxCommandEvent& event )
{
	IndexPropsDlg* dlg = wxDynamicCast( 
			wxTheXmlResource->LoadWindow( "IndexingSvc_IndexPropsDlg", FALSE ),
			IndexPropsDlg );

	dlg->Setup( mpSvc );

	IndexTable& it = mpSvc->GetIndexTable(); 

	if ( !it.FindById( GetListCtrlSelection( mpLstAll ) ) )
		wxMessageBox( "Cannot find selected item" );

	if ( dlg->ShowModal() == wxID_OK )
	
		if ( !it.Write() )
			wxMessageBox( wxString("Failed updating index ") +  it.GetName() );
		else
		{
			it.End();
			mpSvc->RebuildIndex( it.GetId() );
		}

	it.End();

	dlg->Destroy(); // FIXME:: why?
	this->SetFocus();
}

void IndeciesDlg::OnRebuild( wxCommandEvent& event )
{
	if ( !CheckSelection() )
		return;

	wxArrayPtrVoid indexIds;
	indexIds.Add( (void*)GetListCtrlSelection( mpLstAll ) );

	mpSvc->RebuildIndecies( indexIds );
}

void IndeciesDlg::OnRebuildAll( wxCommandEvent& event )
{
	mpSvc->OnRebuildAllIndecies( event );
}

void IndeciesDlg::OnRemove( wxCommandEvent& event )
{
	if ( !CheckSelection() )
		return;

	IndexTable& it = mpSvc->GetIndexTable(); 

	if ( !it.FindById( GetListCtrlSelection( mpLstAll ) ) ||
		 !mpSvc->RemoveCurrentIndex() 
	   )
		wxMessageBox( "Failed removing index" );
	else
	{
		it.End();
		ReloadAllIndexList();
	}

	it.End();
}

void IndeciesDlg::OnAddSingle( wxCommandEvent& event )
{
	if ( !CheckSelection() )
		return;

	wxStringList& lst = wxPMBootSvc().GetActiveProject()->GetUsedIndecies();

	wxString index = mpLstAll->GetItemText( GetListCtrlSelection( mpLstAll ) );
	lst.Add( index );

	wxPMBootSvc().GetActiveProject()->Modify(TRUE);

	ReloadProjectIndexList();
}

void IndeciesDlg::OnAddAll( wxCommandEvent& event )
{
	int cnt = mpLstAll->GetItemCount();

	wxStringList& lst = wxPMBootSvc().GetActiveProject()->GetUsedIndecies();
	lst.Clear();

	for( int i = 0; i != cnt; ++i )

		lst.Add( mpLstAll->GetItemText( i ) );

	wxPMBootSvc().GetActiveProject()->Modify(TRUE);

	ReloadProjectIndexList();
}

void IndeciesDlg::OnRemoveSingle( wxCommandEvent& event )
{
	if ( GetListCtrlSelection( mpLstPrj ) == -1 )
		return;

	wxStringList& lst = wxPMBootSvc().GetActiveProject()->GetUsedIndecies();

	wxString index = mpLstPrj->GetItemText( GetListCtrlSelection( mpLstPrj ) );
	lst.Delete( index );

	wxPMBootSvc().GetActiveProject()->Modify(TRUE);

	ReloadProjectIndexList();
}

void IndeciesDlg::OnRemoveAll( wxCommandEvent& event )
{
	wxPMBootSvc().GetActiveProject()->GetUsedIndecies().Clear();

	wxPMBootSvc().GetActiveProject()->Modify(TRUE);

	ReloadProjectIndexList();
}

void IndeciesDlg::OnUp( wxCommandEvent& event )
{
	// aleksandras@cvs.wxWorkshop.sourceforge.net:/cvsroot/wxworkshop

	wxMessageBox( "TBD:: OnUp()" );
}

void IndeciesDlg::OnDown( wxCommandEvent& event )
{
	wxMessageBox( "TBD:: OnDown()" );
}

void IndeciesDlg::OnInitDialog( wxInitDialogEvent& event )
{
}

long IndeciesDlg::GetListCtrlSelection( wxListCtrl* lst )
{
	wxListItem item;

	for( long i = lst->GetTopItem(); i != lst->GetItemCount(); ++i )
	{
		item.m_itemId = i;
		item.m_col = 0;
		item.m_mask = wxLIST_MASK_STATE;
		lst->GetItem( item );

		if ( item.m_state & wxLIST_STATE_SELECTED )
			return i;
	}

	return -1;
}

void IndeciesDlg::SetListCtrlSelection( wxListCtrl* lst, long sel )
{
	wxListItem item;
	item.m_itemId = sel;
	item.m_col = 0;
	item.m_mask = wxLIST_MASK_STATE;
	item.m_state = wxLIST_STATE_SELECTED;

	lst->SetItem( item );
}

/* protected methods */

void IndeciesDlg::ReloadAllIndexList()
{
	mpLstAll->DeleteAllItems();
	IndexTable& it = mpSvc->GetIndexTable(); 

	it.Begin();
	it.First(); int i = 0;

	while( it.Ok() )
	{
		wxListItem item;
		item.m_itemId = i++;
		item.m_mask = wxLIST_MASK_TEXT;

		item.m_col = 0;
		item.m_text = it.GetName();
		mpLstAll->InsertItem( item );

		item.m_col = 1;
		item.m_text = it.GetHandlerName();
		mpLstAll->SetItem( item );

		item.m_col = 2;
		item.m_text = it.GetBoundProject();
		mpLstAll->SetItem( item );

		it.Next();
	}

	it.End();
}

void IndeciesDlg::ReloadProjectIndexList()
{
	mpLstPrj->DeleteAllItems();

	if (wxPMBootSvc().GetActiveProject())
	{
		wxStringList& lst = wxPMBootSvc().GetActiveProject()->GetUsedIndecies();
		
		for( size_t i = 0; i != lst.GetCount(); ++i )
			
			mpLstPrj->InsertItem( i, lst[i] );
	}
}

bool IndeciesDlg::CheckSelection()
{
	if ( GetListCtrlSelection( mpLstAll ) == -1 )
	{
		wxMessageBox( "Select item from all-indexes list first" );
		return FALSE;
	}
	else
		return TRUE;
}

/***** Implementation for class IndexPropsDlg *****/

IMPLEMENT_DYNAMIC_CLASS( IndexPropsDlg, wxDialog )

BEGIN_EVENT_TABLE( IndexPropsDlg, wxDialog )

	EVT_BUTTON( XMLID("del"), IndexPropsDlg::OnDel )
	EVT_BUTTON( XMLID("add"), IndexPropsDlg::OnAdd )
	EVT_BUTTON( XMLID("browse"), IndexPropsDlg::OnBrowse )
	EVT_BUTTON( wxID_OK, IndexPropsDlg::OnOk )
	EVT_RADIOBUTTON( XMLID("toProject"), IndexPropsDlg::OnBoundToPrj )
	EVT_RADIOBUTTON( XMLID("toFiles"),   IndexPropsDlg::OnBoundToFiles )

	EVT_INIT_DIALOG( IndexPropsDlg::OnInitDialog )

END_EVENT_TABLE()

IndexPropsDlg::IndexPropsDlg()
{}

void IndexPropsDlg::Setup( wxPMIndexingService* svc )
{
	mpSvc = svc;

	mpName		   = (wxTextCtrl*)wxXmlFindWindow( "name", this );
	mpHandler	   = (wxChoice*)  wxXmlFindWindow( "handler", this );
	mpBoundProject = (wxTextCtrl*)wxXmlFindWindow( "boundProject", this );
	mpFiles		   = (wxListCtrl*)wxXmlFindWindow( "fileList", this );

	mpFiles->InsertColumn( 0, "Name", wxLIST_FORMAT_LEFT, 400 );

	IndexTable& tbl = mpSvc->GetIndexTable();

	mpName->SetValue( tbl.GetName() );

	IndexHandlerListT& lst = mpSvc->GetIndexHandlers();
	
    unsigned int i;
	for (i = 0 ; i != lst.Count(); ++i)
	{
		const char* name = lst[i]->GetIndexHandlerName();
		mpHandler->Append( lst[i]->GetIndexHandlerDescription() );

		if ( strcmp( name, tbl.GetHandlerName() ) == 0 )

			mpHandler->SetSelection(i);
	}

	wxString bound = tbl.GetBoundProject();
	mpBoundProject->SetValue( bound );

	EnableFilesList( !bound.Length() );
	((wxRadioButton*)wxXmlFindWindow( "toFiles", this ))->SetValue( !bound.Length() );
	((wxRadioButton*)wxXmlFindWindow( "toProject", this ))->SetValue( bound.Length() != 0 );

	wxStringTokenizer tok( tbl.GetPaths(), ";", wxTOKEN_STRTOK );
	i = 0;

	while ( tok.HasMoreTokens() )
	{
		wxString path = tok.GetNextToken();

		mpFiles->InsertItem( i++, path );
	}
}

void IndexPropsDlg::OnBrowse( wxCommandEvent& event )
{
	wxDirDialog dlg( this, "Select content root directory", wxGetCwd() );

	if ( dlg.ShowModal() == wxID_OK )
	{
	}
}

void IndexPropsDlg::OnOk( wxCommandEvent& event )
{
	if ( !mpHandler->GetStringSelection().Length() )
	{
		wxMessageBox( "No index handler selected" );
		return;
	}

	if ( !mpName->GetValue().Length() )
	{
		wxMessageBox( "Index name cannot be empty" );
		return;
	}

	wxRadioButton* toProject = (wxRadioButton*)wxXmlFindWindow( "toProject", this );


	IndexTable& tbl = mpSvc->GetIndexTable();

	tbl.SetName( mpName->GetValue() );

	if ( toProject->GetValue() ) tbl.SetBoundProject( mpBoundProject->GetValue() );
							else tbl.SetBoundProject( "" );

	wxString name = mpSvc->GetIndexHandlers()[ mpHandler->GetSelection() ]->GetIndexHandlerName();
	tbl.SetHandlerName( name );

	wxString paths;

	if ( !toProject->GetValue() )
    {
        int i;
		for (i = 0; i != mpFiles->GetItemCount(); ++i)
		{
			if ( i != 0 )
				paths += ';';

			paths += mpFiles->GetItemText(i);
		}
    }

	tbl.SetPaths( paths );

	EndModal( wxID_OK );
}

void IndexPropsDlg::OnDel( wxCommandEvent& event )
{
	if ( GetFileSelection() == -1 )
	{
		wxMessageBox( "Select path first" );
		return;
	}

	int sel = GetFileSelection();
	mpFiles->DeleteItem( GetFileSelection() );

	if ( sel == mpFiles->GetItemCount() )
		--sel;

	if ( sel >= 0 )
		SetFileSelection( sel );
}

void IndexPropsDlg::OnAdd( wxCommandEvent& event )
{
	mpFiles->SetFocus();

	AddFilesDlg* dlg = wxDynamicCast( 
			wxTheXmlResource->LoadWindow( "IndexingSvc_AddFilesDlg", FALSE ),
			AddFilesDlg );

	dlg->Setup( mpSvc );

	if ( dlg->ShowModal() == wxID_OK )
	{
		if ( !dlg->GetPath().Length() )
		
			wxMessageBox( "Cannot add empty path" );
		else
		{
			this->SetFocus();
			mpFiles->InsertItem( mpFiles->GetItemCount(), dlg->GetPath() );
			SetFileSelection( mpFiles->GetItemCount() - 1 );
		}
	}

	dlg->Destroy(); // FIXME:: why?
}

void IndexPropsDlg::OnBoundToPrj( wxCommandEvent& event )
{
	EnableFilesList( FALSE );
	event.Skip();
}

void IndexPropsDlg::OnBoundToFiles( wxCommandEvent& event )
{
	EnableFilesList( TRUE );
	event.Skip();
}

void IndexPropsDlg::OnInitDialog( wxInitDialogEvent& event )
{
}

/* protected methods */

long IndexPropsDlg::GetFileSelection()
{
	return IndeciesDlg::GetListCtrlSelection( mpFiles );
}

void IndexPropsDlg::SetFileSelection( long sel )
{
	IndeciesDlg::SetListCtrlSelection( mpFiles, sel );
}

void IndexPropsDlg::EnableFilesList( bool enable )
{
	wxWindow* lst = wxXmlFindWindow( "fileList", this );
	wxWindow* add = wxXmlFindWindow( "add", this );
	wxWindow* del = wxXmlFindWindow( "del", this );
	wxWindow* prj = wxXmlFindWindow( "boundProject", this );

	lst->Enable( enable );
	add->Enable( enable );
	del->Enable( enable );
	prj->Enable( !enable );
}

/***** Implementation for class AddFilesDlg *****/

IMPLEMENT_DYNAMIC_CLASS( AddFilesDlg, wxDialog )

BEGIN_EVENT_TABLE( AddFilesDlg, wxDialog )

	EVT_BUTTON( XMLID("select"), AddFilesDlg::OnSelect )

	EVT_INIT_DIALOG( AddFilesDlg::OnInitDialog )

END_EVENT_TABLE()

AddFilesDlg::AddFilesDlg()
{}

void AddFilesDlg::Setup( wxPMIndexingService* svc )
{
	mpSvc = svc;
}

void AddFilesDlg::OnSelect( wxCommandEvent& event )
{
	if ( IsDirectory() )
	{
		wxDirDialog dlg( this, "Choose directory", wxGetCwd() );

		if ( dlg.ShowModal() == wxID_OK )
		
			mpPath->SetValue( dlg.GetPath() + wxFILE_SEP_PATH );
	}
	else
	{
		wxFileDialog dlg( this, "Choose file", wxGetCwd() );

		if ( dlg.ShowModal() == wxID_OK )

			mpPath->SetValue( dlg.GetPath() );
	}

	mpPath->SetInsertionPoint( mpPath->GetValue().Length() );
	mpPath->SetFocus();
}

void AddFilesDlg::OnInitDialog( wxInitDialogEvent& event )
{
	mpIsDir = (wxRadioButton*)wxXmlFindWindow( "is_dir", this );
	mpPath = (wxTextCtrl*)wxXmlFindWindow( "path", this );
	mpMask = (wxComboBox*)wxXmlFindWindow( "mask", this );

	mpMask->Append( "*.h" );
	mpMask->Append( "*.cpp" );
	mpMask->Append( "*.cpp,*.h" );
	mpMask->Append( "*.h,*.hpp,*.hxx" );
	mpMask->Append( "*.cpp,*.C,*.c,*.cxx" );
	mpMask->Append( "*.cpp,*.h,*.hpp,*.cxx,*.C,*.c" );
	mpMask->Append( "*.py" );
	mpMask->Append( "*.pl" );
	mpMask->Append( "*.html,*.htm" );
	mpMask->Append( "*.txt" );
}

wxString AddFilesDlg::GetPath()
{
	if ( IsDirectory() )
	
		return mpPath->GetValue() + mpMask->GetValue();
	else
		return mpPath->GetValue();
}

bool AddFilesDlg::IsDirectory()
{
	return mpIsDir->GetValue();
}

/***** Implementation for class DbTableBase *****/

DbTableBase::DbTableBase()
	: mpRel( NULL ),
	  mpSvc( NULL )
{}

DbTableBase::~DbTableBase()
{
	wxDELETE(mpRel);
}

bool DbTableBase::Begin( int indexId )
{
	if ( indexId == -1 )
		return mpRel->begin( GetPrimaryIndexId() ) == db_ok;
	else
		return mpRel->begin( (const int32)indexId ) == db_ok;
}

bool DbTableBase::End()
{
	return mpRel->end() == db_ok;
}

bool DbTableBase::First()
{
	return mpRel->first() == db_ok;
}

bool DbTableBase::Next()
{
	return mpRel->next() == db_ok;
}

bool DbTableBase::Ok()
{
	return mpRel->ok();
}

bool DbTableBase::Add()
{
	return mpRel->add() == db_ok;
}

bool DbTableBase::Delete()
{
	return mpRel->del() == db_ok;
}

void DbTableBase::Setup( diaRel* rel, wxPMIndexingService* svc )
{
	mpRel = rel;
	mpSvc = svc;
}

bool DbTableBase::Extract()
{
	return mpRel->extract() == db_ok;
}

bool DbTableBase::Write()
{
	return mpRel->write() == db_ok;
}

bool DbTableBase::RemoveAll()
{
	mpRel->begin();
	bool failed = FALSE;

	for ( mpRel->first(); mpRel->ok(); mpRel->next() )
    {
		failed |= (mpRel->del() != db_ok);
    }

	mpRel->end();
	mpRel->flush();

	return failed == FALSE;
}

bool DbTableBase::Flush()
{
	return mpRel->flush() == db_ok;
}

bool DbTableBase::Seek()
{
	return mpRel->seek() == db_ok;
}

bool DbTableBase::Find()
{
	return mpRel->find() == db_ok;
}

bool DbTableBase::FindById( int id )
{
	SetId( id );
	mpRel->begin( GetPrimaryIndexId() );

	bool ok = mpRel->find() == db_ok;

	return ok;
}

bool DbTableBase::DeleteById( int id )
{
	if ( !FindById( id ) )
	{
		mpRel->end();
		return FALSE;
	}

	bool ok = mpRel->del() == db_ok;

	mpRel->end();

	return ok;
}

/***** Implementation for class IndexTable *****/

const char* IndexTable::GetName()
	
	{ return ((Index*)mpRel)->name; }

const char* IndexTable::GetLanguage()

	{ return ((Index*)mpRel)->language; }

const char* IndexTable::GetHandlerName()
	
	{ return ((Index*)mpRel)->handlerName; }

const char* IndexTable::GetBoundProject()

	{ return ((Index*)mpRel)->boundProject;}

const char* IndexTable::GetContentRoot()
	
	{ return ((Index*)mpRel)->contentRoot; }

const char* IndexTable::GetPaths()

	{ return ((Index*)mpRel)->paths; }

const char* IndexTable::GetDataStr()
	
	{ return (const char*)((Index*)mpRel)->dataStr; }

int IndexTable::GetId()
	
	{ return ((Index*)mpRel)->id; }

void IndexTable::SetName( const wxString& name )

	{ strcpy( ((Index*)mpRel)->name, name.c_str() ); }

void IndexTable::SetLanguage( const wxString& lang )

	{ ((Index*)mpRel)->language = lang.c_str(); }

void IndexTable::SetHandlerName( const wxString& handler )

	{ ((Index*)mpRel)->handlerName = handler.c_str(); }

void IndexTable::SetBoundProject( const wxString& project )

	{ strcpy( ((Index*)mpRel)->boundProject, project.c_str() ); }

void IndexTable::SetContentRoot( const wxString& root )

	{ ((Index*)mpRel)->contentRoot = root.c_str(); }

void IndexTable::SetDataStr( const wxString& dataStr )

	{ ((Item*)mpRel)->dataStr = dataStr.c_str(); }

void IndexTable::SetPaths( const wxString& paths )

	{ ((Index*)mpRel)->paths = paths.c_str(); }

void IndexTable::SetId( int id )

	{ ((Index*)mpRel)->id = id; }

int IndexTable::GetPrimaryIndexId()

	{ return Index::dbIdx_by_id; }

bool IndexTable::FindIndexByProject( const wxString& boundPrjName )
{
	SetId( -1 );
	SetBoundProject( boundPrjName );

	mpRel->begin( Index::dbIdx_by_project );

	mpRel->seek();
	mpRel->next();

	return mpRel->ok() && GetBoundProject() == boundPrjName;
}

bool IndexTable::FindById( int indexId )
{
	SetId( indexId );
	mpRel->begin( Index::dbIdx_by_id );

	mpRel->seek();
	mpRel->next();

	bool ok = mpRel->ok() && GetId() == indexId;

	return ok;
}

/***** Implementation for class FileTable *****/

const char* FileTable::GetName()

	{ return ((File*)mpRel)->name; }

wxDateTime FileTable::GetLastIndexingDate()
{ 
	// FIXME:: TBD:: ...
	return wxDateTime::Now();
}

int FileTable::GetId()

	{ return ((File*)mpRel)->id; }

int FileTable::GetIndexId()

	{ return ((File*)mpRel)->fk_index; }

void FileTable::SetName( const wxString& name )

	{ strcpy( ((File*)mpRel)->name, name.c_str() ); }

void FileTable::SetLastIndexingDate( const wxDateTime& date )
{ 
	// FIXME:: TBD:: ... 
}

void FileTable::SetIndexId( int indexId )

	{ ((File*)mpRel)->fk_index = indexId; }

void FileTable::DeleteItemsByIndex( int indexId )
{
	SetId( -1 );
	SetIndexId( indexId );
	mpRel->begin( File::dbIdx_by_index );

	mpRel->seek();
	
	for( mpRel->next(); mpRel->ok() && ((File*)mpRel)->fk_index == indexId; 
		 mpRel->next() )
	
		mpRel->del();
	
	mpRel->end();
}

bool FileTable::FindByName( const wxString& name, int indexId )
{
	SetId( -1 );
	SetIndexId( indexId );
	SetName( name );
	mpRel->begin( File::dbIdx_by_name );

	mpRel->seek();
	mpRel->next();
	bool ok = FALSE;

	if ( mpRel->ok() && 
		 ((File*)mpRel)->fk_index == indexId &&
		 GetName() == name )
	
		 ok = TRUE;

	mpRel->end();
	return ok;
}

bool FileTable::FindByIndex( int indexId )
{
	SetId( -1 );
	SetIndexId( indexId );
	mpRel->begin( File::dbIdx_by_index );

	mpRel->seek();
	mpRel->next();

	if ( mpRel->ok() && 
		 ((File*)mpRel)->fk_index == indexId )

		return TRUE;

	mpRel->end();
	return FALSE;
}

void FileTable::SetId( int id )
{
	((File*)mpRel)->id = id;
}

int FileTable::GetPrimaryIndexId()
{
	return File::dbIdx_by_id;
}

/***** Implementation for class ItemTable *****/

bool ItemTable::Begin( int indexId )
{
	int realIdxId = -1;
	switch( indexId )
	{
		case BY_NAME  : realIdxId = Item::dbIdx_by_name; break;
		case BY_INDEX : realIdxId = Item::dbIdx_by_index; break;
		case BY_FILE  : realIdxId = Item::dbIdx_by_file; break;
		case BY_TYPE  : realIdxId = Item::dbIdx_by_type; break;
		default: return mpRel->begin() == db_ok;
	}

	return mpRel->begin( realIdxId ) == db_ok;
}

const char* ItemTable::GetName()

	{ return ((Item*)mpRel)->name; }

const char* ItemTable::GetContext()

	{ return ((Item*)mpRel)->context; }

wxString ItemTable::GetDataStr()

	{ return wxString( (const char*)((Item*)mpRel)->dataStr ); }

int ItemTable::GetLineNo()

	{ return ((Item*)mpRel)->lineNo; }

int ItemTable::GetLineCount()

	{ return ((Item*)mpRel)->lineCount; }

int ItemTable::GetType()

	{ return ((Item*)mpRel)->type; }

int ItemTable::GetId()

	{ return ((Item*)mpRel)->id; }

int ItemTable::GetFileId()

	{ return ((Item*)mpRel)->fk_file; }

int ItemTable::GetIndexId()

	{ return ((Item*)mpRel)->fk_index; }

void ItemTable::SetName( const wxString& name )

	{ strcpy( ((Item*)mpRel)->name, name.c_str() ); }

void ItemTable::SetContext( const wxString& context )

	{ strcpy( ((Item*)mpRel)->context, context.c_str() ); }

void ItemTable::SetDataStr( const wxString& dataStr )

	{ ((Item*)mpRel)->dataStr = dataStr.c_str(); }

void ItemTable::SetLineNo( int lineNo )

	{ ((Item*)mpRel)->lineNo = lineNo; }

void ItemTable::SetLineCount( int count )

	{ ((Item*)mpRel)->lineCount = count; }

void ItemTable::SetType( int type )

	{ ((Item*)mpRel)->type = type; }

void ItemTable::SetFileId( int fileId )

	{ ((Item*)mpRel)->fk_file = fileId; }

void ItemTable::SetIndexId( int indexId )

	{ ((Item*)mpRel)->fk_index = indexId; }

void ItemTable::DeleteItemsByType( int _type, int indexId )
{
	SetId( -1 );
	SetIndexId( indexId );
	SetType( _type );
	mpRel->begin( Item::dbIdx_by_type );

	mpRel->seek();

	for ( mpRel->next(); mpRel->ok() &&
						GetType() == _type &&
						GetIndexId() == indexId ; mpRel->next() )
    {
		mpRel->del();
    }

	mpRel->end();
}

void ItemTable::SetId( int id )
{
	((Item*)mpRel)->id = id;
}

int ItemTable::GetPrimaryIndexId()
{
	return -1; // "Item" relation has no primary index
}

bool ItemTable::FindByName( const wxString& _name, const wxString& _context, 
							int _type, bool matchSubstring )
{
	SetId( -1 );
	SetName( _name );
	SetContext( _context );
	SetType( _type );

	mpRel->begin( Item::dbIdx_by_name );
	//mpRel->first();

	mpRel->seek();
	mpRel->next();

	while ( mpRel->ok() )
	{
		wxString n = GetName();
		wxString c = GetContext();
		int t = GetType();
		int id = GetId();

		if ( matchSubstring ||
			 ( GetName() == _name &&
			   GetContext() == _context &&
			   GetType() == _type )
		   )
		{
			mpRel->end();
			return TRUE;
		}

		mpRel->next();
	}
	
	mpRel->end();
	
	return FALSE;
}

bool ItemTable::FindByType( int _type )
{
	SetId( -1 );
	SetType( _type );

	mpRel->begin( Item::dbIdx_by_type );

	mpRel->seek();
	mpRel->next();

	return  mpRel->ok() && GetType() == _type;
}

bool ItemTable::FindByType( int _type, int indexId )
{
	SetId( -1 );
	SetIndexId( indexId );
	SetType( _type );

	mpRel->begin( Item::dbIdx_by_type );

	mpRel->seek();
	mpRel->next();

	return  mpRel->ok() && 
		    GetType() == _type &&
			GetIndexId() == indexId;
}

bool ItemTable::FindByFileAndIndex( int fileId, int indexId )
{
	SetId( -1 );
	SetFileId( fileId );
	SetIndexId( indexId );

	mpRel->begin( Item::dbIdx_by_index );

	mpRel->seek();
	mpRel->next();

	return  mpRel->ok() && 
			GetIndexId() == indexId && 
			GetFileId()  == fileId;
}

void ItemTable::DeleteItemsByFile( int fileId, int indexId )
{
	SetId( -1 );
	SetFileId( fileId );
	SetIndexId( indexId );

	mpRel->begin( Item::dbIdx_by_file );

	mpRel->seek();

	for( mpRel->next(); mpRel->ok() &&
						GetFileId() == fileId &&
						GetIndexId() == indexId ; mpRel->next() )

		mpRel->del();

	mpRel->end();
}

bool ItemTable::DeleteByIdAndType( int type, int itemId )
{
	SetId( itemId );
	SetType( type );

	mpRel->begin( Item::dbIdx_by_type );
	mpRel->seek();
	mpRel->next();

	bool ok = FALSE;

	if ( mpRel->ok() && 
		 GetId() == itemId &&
		 GetType() == type )
	
		ok = mpRel->del() == db_ok;

	mpRel->end();
	return ok;
}
