/////////////////////////////////////////////////////////////////////////////
// Name:        No names yet.
// Purpose:     Contrib. demo
// Author:      Aleksandras Gluchovas
// Modified by:
// Created:     2001/03/28
// RCS-ID:      $Id: indexing_svc.h,v 1.8 2001/11/15 20:07:12 georgetasker Exp $
// Copyright:   (c) Aleksandras Gluchovas (alex@soften.ktu.lt)
// Licence:   	GPL (version 2.0 or later)
/////////////////////////////////////////////////////////////////////////////

#ifndef __INDEXING_SVC_G__
#define __INDEXING_SVC_G__

#ifdef __GNUG__
    #pragma interface "indexing_svc.h"
#endif

#include "wx/pmf/pm.h"
#include "wx/list.h"
#include "wx/dialog.h"			
#include "wx/dynarray.h"
#include "wx/datetime.h"
#include "wx/listctrl.h"
#include "wx/txtstrm.h"

class wxPMIndexingService;
class DbTableBase;
class IndexTable;
class FileTable;
class ItemTable;

class I_IndexHandler
{
public:
	virtual const char* GetIndexHandlerName() = 0;
	virtual const char* GetIndexHandlerDescription() = 0;

	virtual bool OnClearIndexItems( int indexId ) = 0;
	virtual bool OnBeginBuild( int indexId ) = 0;
	virtual bool OnIndexFile( const wxPMFileInfo& file, int fileId, int indexId ) = 0;
	virtual void OnDumpIndexForFile( wxTextOutputStream& stm, int fileId, int indexId ) = 0;
	virtual bool OnEndBuild( int indexId ) = 0;

	virtual bool DocumentMonitoringNeeded(); // default imp. returns FALSE
	virtual bool CanAcceptFile( const wxPMFileInfo& file, int indexId );
	virtual void OnDocumentsSwitched( wxPMDocument* fromDoc, wxPMDocument* toDoc, int fileId, int indexId );
	virtual void OnDocumentSaved( wxPMDocument& doc, int fileId, int indexId );
	virtual void OnDocumentChangesDiscarded( wxPMDocument& doc, int fileId, int indexId );
};

WX_DEFINE_ARRAY( I_IndexHandler*, IndexHandlerListT );

// used to store association between opened project 
// and indexes bound to it
class IdxHandlerToPrjBindInfo : public wxObject
{
public:
	wxArrayInt mIndexIds;
	IndexHandlerListT mHandlers;
};

class wxPMIndexingService : public wxPMSimpleService,
						    public wxPMDocumentListener,
							public wxPMWorkplaceListener
{
public:
	wxPMIndexingService();
	~wxPMIndexingService();

	enum COMMAND_IDS
	{
		ID_INDECIES,
		ID_REBUILD_ALL_INDECIES,
		ID_CREATE_TEST_DATA,
		ID_DUMP_INDEX,
		ID_SYNC_INDECIES
	};

	// overriden methods of wxPMService

	virtual bool Start();
	virtual bool Stop();

	virtual wxPMMergableMenuChain* GetMergableMenuChain();

	IndexTable& GetIndexTable();
	FileTable& GetFileTable();
	ItemTable& GetItemTable();

	void FlushAll();

	virtual void* GetInterface( const char* infName, const char* version = NULL );

	// implementation of wxPMDocumentListener interface

	virtual void OnDocumentSaved( wxPMDocument& doc );
	virtual void OnDocumentChangesDiscarded( wxPMDocument& doc );
	virtual void OnDocumentsSwitched( wxPMDocument* fromDoc, wxPMDocument* toDoc );

	virtual void OnDocumentOpened( wxPMDocument& doc, bool created );
	virtual void OnWorkplaceOpened( wxPMWorkplace& workplace );

	// implementation of wxPMWorkplaceListener interface

	virtual void OnFileAdded( wxPMFileInfo& file, wxPMProject* toProject );
	virtual void OnFileRemoved( wxPMFileInfo& file, wxPMProject* fromProject  );

	// command handlers

	void OnIndecies( wxCommandEvent& event );
	void OnRebuildAllIndecies( wxCommandEvent& event );
	void RebuildIndecies( wxArrayPtrVoid& indexIds );
	void RebuildIndex( int indexId );
	void OnCreateTestData( wxCommandEvent& event );
	void OnDumpIndex( wxCommandEvent& event );
	void OnSyncIndecies( wxCommandEvent& event );

	// other methods

	virtual bool RegisterIndexHandler( I_IndexHandler* handler );
	virtual bool UnregisterIndexHandler( I_IndexHandler* handler );
	IndexHandlerListT& GetIndexHandlers();

	bool RebuildCurrentIndex();
	bool RemoveCurrentIndex();
	I_IndexHandler* FindHandlerByName( const char* name );

	// returns id or -1 if not found
	int FindFileByPath( const wxString& path, int indexId );

	void TestIt();
	void DropAllData();
	void CreateTestData();

	DECLARE_EVENT_TABLE()
	DECLARE_DYNAMIC_CLASS( wxPMIndexingService )

protected:
	void CleanUp();
	void BuildIndexForFile( const wxString& file, I_IndexHandler* handler );
	void RemoveIndexForFile( const wxString& file, I_IndexHandler* handler );
	void BuildIndexRecursively( const wxString& dir, I_IndexHandler* handler, 
								const wxString& fileMask );

	void GetDirEntries( wxStringList& toLst, const wxString& mask, int flags );
	void NotifyHandlers( int notifyType, wxPMDocument* doc1, wxPMDocument* doc2 );

	enum NOTIFY_TYPES
	{
		DOC_SWITCHED,
		DOC_SAVED,
		DOC_CHANGES_DISCARED
	};

protected:
	IndexTable* mpIndexTbl;
	FileTable*  mpFileTbl;
	ItemTable*  mpItemTbl;
	int	mFilesProcessed;

	IndexHandlerListT mHandlers;
};

class IndeciesDlg : public wxDialog
{
public:
	IndeciesDlg();

	void Setup( wxPMIndexingService* svc );

	void OnNew( wxCommandEvent& event );
	void OnEdit( wxCommandEvent& event );
	void OnRebuild( wxCommandEvent& event );
	void OnRebuildAll( wxCommandEvent& event );
	void OnRemove( wxCommandEvent& event );

	void OnAddSingle( wxCommandEvent& event );
	void OnAddAll( wxCommandEvent& event );
	void OnRemoveSingle( wxCommandEvent& event );
	void OnRemoveAll( wxCommandEvent& event );

	void OnUp( wxCommandEvent& event );
	void OnDown( wxCommandEvent& event );

	void OnInitDialog( wxInitDialogEvent& event );

	static long GetListCtrlSelection( wxListCtrl* lst );
	static void SetListCtrlSelection( wxListCtrl* lst, long sel );

	DECLARE_EVENT_TABLE()
	DECLARE_DYNAMIC_CLASS( IndeciesDlg )
	
protected:
	void ReloadAllIndexList();
	void ReloadProjectIndexList();
	bool CheckSelection();

protected:
	wxPMIndexingService* mpSvc;
	wxListCtrl* mpLstAll;
	wxListCtrl* mpLstPrj;
};

class IndexPropsDlg : public wxDialog
{
public:
	IndexPropsDlg();

	void Setup( wxPMIndexingService* svc );

	void OnBrowse( wxCommandEvent& event );
	void OnDel( wxCommandEvent& event );
	void OnAdd( wxCommandEvent& event );
	void OnOk( wxCommandEvent& event );
	void OnBoundToPrj( wxCommandEvent& event );
	void OnBoundToFiles( wxCommandEvent& event );

	void OnInitDialog( wxInitDialogEvent& event );

	DECLARE_EVENT_TABLE()
	DECLARE_DYNAMIC_CLASS( IndexPropsDlg )

protected:
	void EnableFilesList( bool enable );
	long GetFileSelection();
	void SetFileSelection( long sel );
	
protected:
	wxPMIndexingService* mpSvc;

	wxTextCtrl* mpName;
	wxChoice*   mpHandler;
	wxTextCtrl* mpBoundProject;
	wxTextCtrl* mpRelativeDir;
	wxListCtrl* mpFiles;
};

class AddFilesDlg : public wxDialog
{
public:
	AddFilesDlg();

	void Setup( wxPMIndexingService* svc );

	void OnSelect( wxCommandEvent& event );
	void OnInitDialog( wxInitDialogEvent& event );

	wxString GetPath();
	bool IsDirectory();

	DECLARE_EVENT_TABLE()
	DECLARE_DYNAMIC_CLASS( AddFilesDlg )
	
protected:
	wxPMIndexingService* mpSvc;

	wxRadioButton* mpIsDir;
	wxTextCtrl* mpPath;
	wxComboBox* mpMask;
};

/****************************************
 *              Table classes           *
 ****************************************/

class diaRel;

class DbTableBase : public wxObject
{
public:
	DbTableBase();
	virtual ~DbTableBase();

	virtual bool Begin( int indexId = -1 );
	virtual bool End();
	virtual bool First();
	virtual bool Next();
	virtual bool Ok();
	virtual bool Add();
	virtual bool Delete();
	virtual bool Seek();

	virtual bool Extract();
	virtual bool Write();

	virtual bool RemoveAll();
	virtual bool Flush();

	virtual void Setup( diaRel* rel, wxPMIndexingService* svc );

	virtual void SetId( int id ) = 0;
	virtual int GetPrimaryIndexId() = 0;

	virtual bool Find();
	virtual bool FindById( int id );
	virtual bool DeleteById( int id );

protected:
	diaRel* mpRel;
	wxPMIndexingService* mpSvc;
};

class IndexTable : public DbTableBase
{
public:
	const char* GetName();
	const char* GetLanguage();
	const char* GetHandlerName();
	const char* GetBoundProject();
	const char* GetContentRoot();
	const char* GetPaths();
	const char* GetDataStr();
	int GetId();

	void SetName( const wxString& name );
	void SetLanguage( const wxString& lang );
	void SetHandlerName( const wxString& handler );
	void SetBoundProject( const wxString& project );
	void SetContentRoot( const wxString& root );
	void SetDataStr( const wxString& dataStr );
	void SetPaths( const wxString& paths );

	virtual void SetId( int id );
	virtual int GetPrimaryIndexId();

	// utilities

	bool FindIndexByProject( const wxString& boundPrjName );
	bool FindById( int indexId );
};

class FileTable : public DbTableBase
{
public:
	const char* GetName();
	wxDateTime GetLastIndexingDate();
	int GetId();
	int GetIndexId();

	void SetName( const wxString& name );
	void SetLastIndexingDate( const wxDateTime& date );
	void SetIndexId( int indexId );

	virtual void SetId( int id );
	virtual int GetPrimaryIndexId();

	// utilities

	void DeleteItemsByIndex( int indexId );
	bool FindByName( const wxString& name, int indexId );
	bool FindByIndex( int indexId );
};

class ItemTable : public DbTableBase
{
public:
	enum INDEXES
	{
		BY_NAME,  // by_name on		context,  name,     type,  id
		BY_INDEX, // by_index on	fk_index, fk_file,  id
		BY_FILE,  // by_type on		type,     fk_index, id
		BY_TYPE   // by_file on		fk_file,  fk_index, id
	};

	virtual bool Begin( int indexId = -1 );

	const char* GetName();
	const char* GetContext();
	wxString GetDataStr();
	int GetLineNo();
	int GetLineCount();
	int GetType();

	int GetId();
	int GetFileId();
	int GetIndexId();

	void SetName( const wxString& name );
	void SetContext( const wxString& context );
	void SetDataStr( const wxString& dataStr );
	void SetLineNo( int lineNo );
	void SetLineCount( int count );
	void SetType( int type );
	void SetFileId( int fileId );
	void SetIndexId( int indexId );
	
	virtual void SetId( int id );
	virtual int GetPrimaryIndexId();

	// utilities

	bool FindByName( const wxString& _name, const wxString& _context, int _type, 
					 bool matchSubstring = TRUE );

	bool FindByType( int _type );
	bool FindByType( int _type, int indexId );
	bool FindByFileAndIndex( int fileId, int indexId );


	bool DeleteByIdAndType( int type, int itemId );
	void DeleteItemsByType( int type, int indexId );
	void DeleteItemsByFile( int fileId, int indexId );
};

#endif /* __INDEXING_SVC_G__ */

