///////////////////////////////////////////////////////////////////////////////
// Name:        espinbut.h
// Purpose:     Implementation of an enhanced version of wxSpinButton
// Author:      Hans Van Leemputten
// Modified by:
// Created:     21.01.2001
// RCS-ID:      $Id: espinbut.h,v 1.1 2001/09/07 15:10:43 hansvl Exp $
// Copyright:   Hans Van Leemputten & wxWorkshop Team (c) 2001
// Licence:     wxWindows licence
///////////////////////////////////////////////////////////////////////////////

#ifndef _WX_ESPINBUT_H_
#define _WX_ESPINBUT_H_

#ifdef __GNUG__
    #pragma interface "espinbut.h"
#endif

#if !wxUSE_SPINBTN
    #error You need to include wxUSE_SPINBTN in your wxWindows build!
#endif

#include "wx/spinbutt.h"

class wxBButtonHelper;

// ----------------------------------------------------------------------------
// wxEnhSpinButton
// ----------------------------------------------------------------------------

// TODO:: Fix size problems...

#define wxENHANCED_SPIN_BUTTON_NAME _T("wxEnhSpinButton")

// Extra settings...
#define wxSP_CENTER_BUTTON       0x4000
#define wxSP_MINMAX_BUTTON       0x8000
#define wxSP_DISABLE_BUTTON      0x00010000

// Some appearance defines.
#define wxWIDTH_MINMAX_LINE 1;

class WXDLLEXPORT wxEnhSpinButton : public wxSpinButtonBase
{
public:
    // Constructors
    wxEnhSpinButton();
    wxEnhSpinButton(wxWindow *parent,
                    wxWindowID id = -1,
                    const wxPoint& pos = wxDefaultPosition,
                    const wxSize& size = wxDefaultSize,
                    long style = wxSP_HORIZONTAL | wxSP_ARROW_KEYS | wxSP_MINMAX_BUTTON,
                    const wxString& name = wxENHANCED_SPIN_BUTTON_NAME);

    // Destructor
    ~wxEnhSpinButton();

    bool Create(wxWindow *parent,
                wxWindowID id = -1,
                const wxPoint& pos = wxDefaultPosition,
                const wxSize& size = wxDefaultSize,
                long style = wxSP_HORIZONTAL | wxSP_ARROW_KEYS | wxSP_MINMAX_BUTTON,
                const wxString& name = wxENHANCED_SPIN_BUTTON_NAME);

    // implement wxSpinButtonBase methods
    virtual void SetRange(int minVal, int maxVal);
    virtual int  GetValue() const;
    virtual void SetValue(int val);

    // Enable/Disable all buttons...
    virtual bool Enable( bool enable = TRUE );

protected:

    // the common part of all ctors
    void Init();

    // normalize the value to fit into min..max range
    int NormalizeValue(int value) const;

    // change the value by +1/-1 and send the event, return TRUE if value was
    // changed
    bool ChangeValue(int inc);

    // Change enable/disable state of the buttons... if needed.
    void ChangeEnable();

    // Button Events...
    void OnButton(wxCommandEvent &event);

    // Sizing the widget
    virtual wxSize DoGetBestSize() const;
    void SetBestSize(const wxSize& size);

    // Fit the Buttons in the size set...
    void Fit();

    // the current controls value
    int m_value;

    // The number of buttons on the control.
    int m_nButtons;

    // Style info.
    bool m_bMinMaxBnt;  // Show Min and Max Button?
    bool m_bCenterBnt;  // Show Center Button?
    bool m_bArrowBnt;   // Show Normal Arrow Button?
    bool m_bDisableBnt; // Disable buttons if max range is reached for that button

private:

    // List of the 5 posible buttons (check m_nButtons to know the counter) 
    wxBButtonHelper *m_pButton[5];

    // Call to construct all needed buttons... in the correct style
    void CreatButtons();
    // Call to Destruct all buttons in the button list.
    void DestroyButtons();

    DECLARE_DYNAMIC_CLASS(wxEnhSpinButton)
    DECLARE_EVENT_TABLE()
};

#endif // _WX_ESPINBUT_H_

