/////////////////////////////////////////////////////////////////////////////
// Name:        pagectrl.h
// Purpose:     wxPagingCtrl class
// Author:      Hans Van Leemputten
// Modified by:
// Created:     27/May/2001
// RCS-ID:      $Id: pagectrl.h,v 1.8 2001/09/07 15:12:46 hansvl Exp $
// Copyright:   (c) Hans Van Leemputten
// Licence:     wxWindows license
/////////////////////////////////////////////////////////////////////////////

#ifndef _WX_PAGING_CTRL_H_
#define _WX_PAGING_CTRL_H_

#ifdef __GNUG__
    #pragma interface "pagectrl.h"
#endif

// ----------------------------------------------------------------------------
// headers and constants
// ----------------------------------------------------------------------------

#include "flipperb.h"

#include "wx/spinbutt.h"
class wxEnhSpinButton;

#define RESIZE_STICK_SIZE   6   // Size of the tab zone resizer.

// ----------------------------------------------------------------------------
// definition
// ----------------------------------------------------------------------------
class wxPagingCtrl : public wxFlipperCtrlBase
{
    DECLARE_DYNAMIC_CLASS(wxPagingCtrl)
    DECLARE_EVENT_TABLE()

public:
    // Constructor
    wxPagingCtrl();
    wxPagingCtrl( wxWindow *parent,
                  wxWindowID id = -1,
                  const wxPoint& pos = wxDefaultPosition,
                  const wxSize& size = wxDefaultSize,
                  long  style = 0,
                  const wxString& name = wxT("wxPagingCtrl"));

    // Destructor
    virtual ~wxPagingCtrl();

    // Create and init.
    bool Create( wxWindow *parent, wxWindowID id = -1,
                 const wxPoint& pos = wxDefaultPosition,
                 const wxSize& size = wxDefaultSize,
                 long  style = 0,
                 const wxString& name = wxT("wxPagingCtrl"));

private:
    // Basic init for this class.
    void DoInitPagingCtrl();

protected:
    // Draw a Shaded rectangle
    void DrawShadedRect(int x, int y, int width, int height,
        wxPen& upperPen, wxPen& lowerPen, wxDC& dc);

    // Draw 1 paper bar in dc
    void DrawPaperBar(twTabInfo& tab, int x, int y,
         wxBrush& brush, wxPen& pen, wxDC& dc);

    // Get the coordinates where it is allowed to draw the title bar (Tab)
    virtual void GetTitleCoord(int *x1, int *y1, int *x2, int *y2);
    // Get the coordinates where it is allowed to draw the body in
    virtual void GetBodyCoord(int *x1, int *y1, int *x2, int *y2);
    // Get the coordinates where it is allowed to draw the boarder in
    virtual void GetCtrlCoord(int *x1, int *y1, int *x2, int *y2);
    // Get the total space needed by all tab's
    int GetWholeTabRowLen(bool addTabSizer = FALSE);

public:
    // Should be invoked to restore the window without any changes
    // Like on paint...
    virtual void DrawDecorations(wxDC& dc);

    // Should be invoked to redisplay window with changed properties
    virtual void RecalcLayout(bool andRepaint = TRUE);
    // Should be invoked to redisplay TabScroll with changed properties 
    // Is done on RecalcLayout...
    virtual void RecalcTabScroll();
    // Should be invoked to redisplay Horiz and Vert Scrollbars with changed properties 
    // Is done on RecalcLayout...
    virtual void RecalcHorizVertScroll();

    // Is called when a Tab is added, Recalc's the values of m_pTabScroll
    virtual void OnTabAdded(twTabInfo* pInfo);

    // Set tabNo as active one, and re init the scrollbars if new active tab.
    virtual void SetActiveTab(int tabNo);

    // Set the tab scroll rate, 0 = only reset the tab scroller... 0 is not stored in m_TabScrollRate!
    void SetTabScrollRate(int nRate = 0);
    // Get the tab scroll rate.
    int GetTabScrollRate();

    // Enables/Disables scroll bar handeling of the client control 
    void SetScrollBars(int tabID, bool bHorizScroll, bool bVertScroll); 
    // Get the client object to refresh the scrollbars.
    void ScrollRefresh();

    // This function is to have interaction with a 
    // wxPagingCtrl aware object.
    bool ResultLastScrollAction();
    void SetNextCaller(wxWindow *caller);

    // Virtual function need by the embeded control to 
    // handel scrolling
    
    // Sets the scrollbar properties of a built-in or external scrollbar.
    virtual void SetScrollbar(int orientation, int position, int thumbSize, int range, bool refresh = FALSE);
    // Sets the position of one of the built-in or external scrollbars
    virtual void SetScrollPos(int orientation, int pos, bool refresh = FALSE);
    
    // Returns the built-in or external scrollbar thumb size
    virtual int GetScrollThumb(int orientation);
    // Returns the built-in or external scrollbar position.
    virtual int GetScrollPos(int orientation);
    // Returns the built-in or external scrollbar range.
    virtual int GetScrollRange(int orientation);

    // Mouse Event handlers
    virtual void OnLButtonDown(wxMouseEvent& event);
    virtual void OnLButtonUp(wxMouseEvent& event);
    virtual void OnRButtonDown(wxMouseEvent& event);
    virtual void OnMouseMove(wxMouseEvent& event);
    virtual void OnScroll(wxScrollEvent& event);
    virtual void OnTabScroll(wxSpinEvent& event);

protected:
    // All pen's used in this control
    wxPen  m_WhitePen;          // default: RGB(255,255,255)
    wxPen  m_GrayPen;           // default: RGB(192,192,192)
    wxPen  m_DarkPen;           // default: RGB(128,128,128)
    wxPen  m_BlackPen;          // default: RGB( 0,  0,  0)
    wxPen  m_TabPen;            // default: wxSYS_COLOUR_BTNFACE
    wxPen  m_ActiveTabPen;      // default: wxSYS_COLOUR_WINDOW

    // Tab Brush used to erase the background
    wxBrush m_TabBrush;         // default: color of the m_TabPen.
    wxBrush m_ActiveTabBrush;   // default: color of the m_ActiveTabPen.

    // Size/layout related variables (These don't have a Set/Get function).
    int    m_ImageTextGap;      // The Gap between text and Image
                                // Default: 2
    int    m_TabTrianGap;       // The gap that cause the to be a trapezium.
                                // Default: 4
    int    m_TitleHorizGap;     // This is the gap between the tab border and text/image .
                                // If the the border is on X than the text/Image starts at X + m_TitleHorizGap.
                                // Default: 10
    int    m_TitleVertGap;      // This is the gap between the tab border and text/image .
                                // If the the border is on Y than the text/Image starts at Y + m_TitleVertGap.
                                // Default: 2
    int    m_CurentRowOfs;      // This var. contains the current row offset of the title bar.
                                // Default: 0
    int    m_TitleSpace;        // Space that is allowed to be taken by the title bar.
                                // Default: Width of the base class. => Old default 300

    // Scrollbars that will be used in this Control.
    wxEnhSpinButton* m_pTabScroll; // This Scrollbar is used to browse trough the tab 
                                // when there is not enough space to display all tabs
    wxScrollBar* m_pHorizScroll;// Horizontal scrollbar, displayed when active control is class aware.
    wxScrollBar* m_pVertScroll; // Vertical scrollbar, displayed when active control is class aware.

private:
    // Size/layout related variables (These do have a Set/Get function).
    int    m_TabScrollRate;     // The rate that will be used to scroll with the tab scroller.
                                // Default: 20
    bool   m_bIsDragged;        // Is the resize stick dragged or not?
                                // Default: FALSE
    bool   m_bCursorChanged;    // Did we change the cursor or not?
                                // Default: FALSE
    int    m_DragTitleSpace;    // Is used to save the m_TitleSpace when starting dragging.
                                // Default: None. Only has a correct value when m_bIsDragged = TRUE
    int    m_DagOrigin;         // Is used to save the start (x) location, where dragging started.
                                // Default: None. Only has a correct value when m_bIsDragged = TRUE
    int    m_DragMinTitleSpace; // Will contain the minimal size of m_TitleSpace.
                                // Default: None. Only has a correct value when m_bIsDragged = TRUE
    
    // Cursor's used by this class.
    wxCursor m_ResizeCursor;
    wxCursor m_NormalCursor;

    // Scroll redirection variables.
    bool   m_bScrollResult;     // Temp var. that contains the latest scroll action.
                                // Default: FALSE
    wxWindow* m_pCaller;        // Temp var. that contains the calling scroller.
                                // Default: NULL
    bool   m_bScrollRefresh;    // This is used by the refresh function to restrict SetScrollbar function calls
                                // Default: FALSE

    // Control state
    bool   m_bInitDone;         // State of creation.

};

#endif /* _WX_PAGING_CTRL_H_ */

