/////////////////////////////////////////////////////////////////////////////
// Name:        scrollre.h
// Purpose:     Macro implementation
// Author:      Hans Van Leemputten
// Modified by:
// Created:     26/Apr/2001
// RCS-ID:      $Id: scrollre.h,v 1.5 2001/08/20 14:45:46 georgetasker Exp $
// Copyright:   (c) Hans Van Leemputten
// Licence:     wxWindows license
/////////////////////////////////////////////////////////////////////////////

#ifndef _WX_WRAPER_MACRO_SCROLLRE_H_
#define _WX_WRAPER_MACRO_SCROLLRE_H_

// ----------------------------------------------------------------------------
// headers and constants
// ----------------------------------------------------------------------------

// Not needed, there is no corresponding .cpp file.
//#ifdef __GNUG__
//    #pragma implementation "scrollre.h"
//#endif

// Include's for known classes...
#include "pagectrl.h"   // for wxPagingCtrl

// ----------------------------------------------------------------------------
// ScrollBar redirector
// ----------------------------------------------------------------------------

/* ============================================================================ */ 
/* Macro Definition                                                             */ 
/* ============================================================================ */ 

// Use this WX_DEFINE_SCROLL_REDIRECT and WX_IMPLEMENT_SCROLL_REDIRECT macro
// in your class to make it able to redirect the native scroll bars to a
// other class. These Marcos make your class suitable for use with wxPagingCtrl.
// Note: Don't forget to call "InitScrollRedirect" in all your constructors!

/* 
 * Macro: to define new classes that can be used to redirect 
 *        the scrollbars (Actions, etc).
 */

#define  WX_DEFINE_SCROLL_REDIRECT()                                \
protected:                                                          \
    /* Init function to be called in all constructors!!!    */      \
    void InitScrollRedirect(wxWindow *parent = NULL);               \
                                                                    \
public:                                                             \
    /* Interface to alow you to set the control that        */      \
    /* is controlling the ScrollBars from newclass.         */      \
    void SetControlWindow( wxWindow *controller );                  \
    wxWindow *GetControlWindow() { return m_ScrollController; }     \
                                                                    \
    /* Adjust the scrollbars                                */      \
    /* virtual void AdjustScrollbars(void);                 */      \
    /*** Don't know if I need this one???!!!                */      \
                                                                    \
    /* Sets the scrollbar properties of a built-in          */      \
    /* or external scrollbar.                               */      \
    virtual void SetScrollbar(int orientation,                      \
                              int position,                         \
                              int thumbSize,                        \
                              int range,                            \
                              bool refresh = TRUE);                 \
                                                                    \
    /* Sets the position of one of the built-in             */      \
    /* or external scrollbars                               */      \
    virtual void SetScrollPos(int orientation,                      \
                              int pos,                              \
                              bool refresh = TRUE);                 \
                                                                    \
                                                                    \
    /* Returns the built-in or external scrollbar thumb size*/      \
    virtual int GetScrollThumb(int orientation);                    \
                                                                    \
    /* Returns the built-in or external scrollbar position. */      \
    virtual int GetScrollPos(int orientation);                      \
                                                                    \
    /* Returns the built-in or external scrollbar range.    */      \
    virtual int GetScrollRange(int orientation);                    \
                                                                    \
protected:                                                          \
    /* Function to check if the parent/Controller is        */      \
    /* known to this class?                                 */      \
    virtual bool KnownController(wxWindow *parent);                 \
                                                                    \
    /* Get info from the controller                         */      \
    /* Did the call last redirected succeeded?              */      \
    virtual bool CallResult();                                      \
                                                                    \
    /* Let the Controller how is calling next...            */      \
    void SetCaller();                                               \
                                                                    \
protected:                                                          \
    wxWindow             *m_ScrollController;                       
  
/* ============================================================================ */ 
/* Macro implementation                                                         */ 
/* ============================================================================ */ 
                                                                    
#define  WX_IMPLEMENT_SCROLL_REDIRECT(className, baseClassName)     \
/* ---------------------------------------------------------------------------- */ \
/* Init member, Needs to be called in all constructors!                         */ \
/* ---------------------------------------------------------------------------- */ \
void className::InitScrollRedirect(wxWindow *parent)                \
{                                                                   \
    if (parent && KnownController(parent))                          \
        m_ScrollController = parent;                                \
    else                                                            \
        m_ScrollController = this;                                  \
}                                                                   \
/* ---------------------------------------------------------------------------- */ \
/* Controller window handling                                                   */ \
/* ---------------------------------------------------------------------------- */ \
void className::SetControlWindow( wxWindow *target )                \
{                                                                   \
    wxASSERT_MSG( target, wxT("Controller window must not be NULL") ); \
    m_ScrollController = target;                                    \
                                                                    \
    /* TODO:: ADD code here to force a update               */      \
    /*        get a correct scrollbar view.                 */      \
}                                                                   \
                                                                    \
/* ---------------------------------------------------------------------------- */ \
/* New scrolling implementation, Redirect it all :-)                            */ \
/* ---------------------------------------------------------------------------- */ \
void className::SetScrollbar(int orientation, int position, int thumbSize, int range, bool refresh) \
{                                                                   \
    SetCaller();                                                    \
                                                                    \
    if (m_ScrollController == this)                                 \
        baseClassName::SetScrollbar(orientation, position, thumbSize, range, refresh); \
    else                                                            \
    {                                                               \
        m_ScrollController->SetScrollbar(orientation, position, thumbSize, range, refresh); \
                                                                    \
        if (CallResult() == FALSE)                                  \
            baseClassName::SetScrollbar(orientation, position, thumbSize, range, refresh); \
    }                                                               \
}                                                                   \
/* ---------------------------------------------------------------------------- */ \
void className::SetScrollPos(int orientation, int pos, bool refresh) \
{                                                                   \
    SetCaller();                                                    \
                                                                    \
    if (m_ScrollController == this)                                 \
        baseClassName::SetScrollPos(orientation, pos, refresh);     \
    else                                                            \
    {                                                               \
        m_ScrollController->SetScrollPos(orientation, pos, refresh); \
                                                                    \
        if (CallResult() == FALSE)                                  \
            baseClassName::SetScrollPos(orientation, pos, refresh); \
    }                                                               \
}                                                                   \
/* ---------------------------------------------------------------------------- */ \
int className::GetScrollThumb(int orientation)                      \
{                                                                   \
    int result;                                                     \
                                                                    \
    SetCaller();                                                    \
                                                                    \
    if (m_ScrollController == this)                                 \
        result = baseClassName::GetScrollThumb(orientation);        \
    else                                                            \
    {                                                               \
        result = m_ScrollController->GetScrollThumb(orientation);   \
                                                                    \
        if (CallResult() == FALSE)                                  \
            result = baseClassName::GetScrollThumb(orientation);    \
    }                                                               \
                                                                    \
    return result;                                                  \
}                                                                   \
/* ---------------------------------------------------------------------------- */ \
int className::GetScrollPos(int orientation)                        \
{                                                                   \
    int result;                                                     \
                                                                    \
    SetCaller();                                                    \
                                                                    \
    if (m_ScrollController == this)                                 \
        result = baseClassName::GetScrollPos(orientation);          \
    else                                                            \
    {                                                               \
        result = m_ScrollController->GetScrollPos(orientation);     \
                                                                    \
        if (CallResult() == FALSE)                                  \
            result = baseClassName::GetScrollPos(orientation);      \
    }                                                               \
                                                                    \
    return result;                                                  \
}                                                                   \
/* ---------------------------------------------------------------------------- */ \
int className::GetScrollRange(int orientation)                      \
{                                                                   \
    int result;                                                     \
                                                                    \
    SetCaller();                                                    \
                                                                    \
    if (m_ScrollController == this)                                 \
        result = baseClassName::GetScrollRange(orientation);        \
    else                                                            \
    {                                                               \
        result = m_ScrollController->GetScrollRange(orientation);   \
                                                                    \
        if (CallResult() == FALSE)                                  \
            result = baseClassName::GetScrollRange(orientation);    \
    }                                                               \
                                                                    \
    return result;                                                  \
}                                                                   \
                                                                    \
/* ---------------------------------------------------------------------------- */ \
/* Implement controller handeling.                                              */ \
/* ---------------------------------------------------------------------------- */ \
bool className::KnownController(wxWindow *parent)                   \
{                                                                   \
    wxASSERT(parent); /* DBG:: */                                   \
                                                                    \
    return parent->IsKindOf(CLASSINFO(wxPagingCtrl));               \
                                                                    \
    /* TODO Extend the return to newly supported control classes. */ \
}                                                                   \
/* ---------------------------------------------------------------------------- */ \
bool className::CallResult()                                        \
{                                                                   \
    if (KnownController(m_ScrollController))                        \
    {                                                               \
        wxPagingCtrl *tmp_wxPagingCtrl = wxDynamicCast(m_ScrollController, wxPagingCtrl); \
        if (tmp_wxPagingCtrl)                                       \
            return tmp_wxPagingCtrl->ResultLastScrollAction();      \
                                                                    \
        /* TODO Add new supported control classes here! */          \
                                                                    \
        wxASSERT_MSG(NULL, wxT("The class is missing a supported if()!!!")); \
        return FALSE;                                               \
    }                                                               \
    else                                                            \
        return TRUE; /* There is no way to check if it was ok. */  \
}                                                                   \
/* ---------------------------------------------------------------------------- */ \
void className::SetCaller()                                         \
{                                                                   \
    if (KnownController(m_ScrollController))                        \
    {                                                               \
        wxPagingCtrl *tmp_wxPagingCtrl = wxDynamicCast(m_ScrollController, wxPagingCtrl); \
        if (tmp_wxPagingCtrl)                                       \
        {                                                           \
            tmp_wxPagingCtrl->SetNextCaller(this);                  \
            return;                                                 \
        }                                                           \
                                                                    \
        /* TODO Add new supported control classes here! */          \
    }                                                               \
}                                                                   

#endif /* _WX_WRAPER_MACRO_SCROLLRE_H_ */

