/////////////////////////////////////////////////////////////////////////////
// Name:        tabbctrl.h
// Purpose:     wxTabbedCtrl class
// Author:      Aleksandras Gluchovas & Hans Van Leemputten
// Modified by:
// Created:     07/May/2001
// RCS-ID:      $Id: tabbctrl.h,v 1.3 2001/06/24 09:41:23 hansvl Exp $
// Copyright:   (c) Hans Van Leemputten & Aleksandras Gluchovas
// Licence:     wxWindows license
/////////////////////////////////////////////////////////////////////////////

#ifndef _WX_TABBED_CTRL_H_
#define _WX_TABBED_CTRL_H_

#ifdef __GNUG__
    #pragma interface "tabbctrl.h"
#endif

// ----------------------------------------------------------------------------
// headers and constants
// ----------------------------------------------------------------------------
#include "flipperb.h"

// Default height of a tab in wxTITLE_BORDER_ONLY mode.
#define BORDER_ONLY_HEIGHT 16   

// ----------------------------------------------------------------------------
// wxFlipperCtrlBase declaration
// ----------------------------------------------------------------------------
class wxTabbedCtrl : public wxFlipperCtrlBase
{
    DECLARE_DYNAMIC_CLASS(wxTabbedCtrl)
    DECLARE_EVENT_TABLE()

public:
    // Constructor
    wxTabbedCtrl();
    wxTabbedCtrl( wxWindow *parent,
                  wxWindowID id = -1,
                  const wxPoint& pos = wxDefaultPosition,
                  const wxSize& size = wxDefaultSize,
                  long  style = 0,
                  const wxString& name = wxT("wxTabbedCtrl"));

    // Destructor
    virtual ~wxTabbedCtrl();

    // Create and init.
    bool Create( wxWindow *parent, wxWindowID id = -1,
                 const wxPoint& pos = wxDefaultPosition,
                 const wxSize& size = wxDefaultSize,
                 long  style = 0,
                 const wxString& name = wxT("wxTabbedCtrl"));

private:
    // Basic init for this class.
    void DoInitTabbedCtrl();

protected:
    // Draw a Shaded rectangle
    void DrawShadedRect(int x, int y, int width, int height,
        wxPen& upperPen, wxPen& lowerPen, wxDC& dc);

    // Overwrite the easy callable SizeBodyContent 
    // to change the way the value is calculated.
    virtual void EasySizeBodyContent(bool repaint);

    // Get the coordinates where it is allowed to draw the title bar (Tab)
    virtual void GetTitleCoord(int *x1, int *y1, int *x2, int *y2);

public:
    // Should be invoked to restore the window without any changes
    // Like on paint...
    virtual void DrawDecorations(wxDC& dc);

    // Should be invoked to redisplay window with changed properties
    virtual void RecalcLayout(bool andRepaint = TRUE);

    // Mouse Event handlers
    virtual void OnLButtonDown(wxMouseEvent& event);
    virtual void OnRButtonDown(wxMouseEvent& event);

protected:
    // All pen's used in this control
    wxPen  m_WhitePen;          // default: RGB(255,255,255)
    wxPen  m_GrayPen;           // default: RGB(192,192,192)
    wxPen  m_DarkPen;           // default: RGB(128,128,128)
    wxPen  m_BlackPen;          // default: RGB( 0,  0,  0)
    wxPen  m_TabPen;            // default: wxSYS_COLOUR_BTNFACE

    // Tab Brush used to erase the background
    wxBrush m_TabBrush;         // default: color of the m_TabPen.

    // Size/layout related variables (These don't have a Set/Get function).
    int    m_ImageTextGap;      // The Gap between text and Image
                                // Default: 2
    int    m_FirstTitleGap;     // This is the Gap in front the first tab.
                                // Default: 11
    int    m_TitleHorizGap;     // This is the gap between the tab border and text/image .
                                // If the the border is on X than the text/Image starts at X + m_TitleHorizGap.
                                // Default: 4
    int    m_TitleVertGap;      // This is the gap between the tab border and text/image .
                                // If the the border is on Y than the text/Image starts at Y + m_TitleVertGap.
                                // Default: 3

private:
    // Control state
    bool   m_bInitDone;         // State of creation.

};

#endif /* _WX_TABBED_CTRL_H_ */

