/*
 *
 * Program: base code from scroll sample in wxWindows
 *
 */

// For compilers that support precompilation, includes "wx/wx.h".
#include "wx/wxprec.h"

#ifdef __BORLANDC__
#pragma hdrstop
#endif

#ifndef WX_PRECOMP
#include "wx/wx.h"
#endif

#include "wx/image.h"
#include "wx/listctrl.h"
#include "wx/sizer.h"
#include "wx/log.h"

#include "wx/wspp/flipperb.h"
#include "wx/wspp/pagectrl.h"
#include "wx/wspp/tabbctrl.h"

#include "wx/wspp/scrollre.h"
#include "wx/wspp/rescrolw.h"

#include "smile.xpm"

// derived classes
class MyFrame;
class MyApp;

// ****************************************************************************
// ** Test Redirection Macro. Create wxReTextCtrl
// ****************************************************************************
class wxReTextCtrl: public wxTextCtrl
{
public:
	wxReTextCtrl();

	wxReTextCtrl(wxWindow *parent, wxWindowID id,
               const wxString& value = wxEmptyString,
               const wxPoint& pos = wxDefaultPosition,
               const wxSize& size = wxDefaultSize,
               long style = 0,
               const wxValidator& validator = wxDefaultValidator,
               const wxString& name = wxTextCtrlNameStr)
	{
        Create(parent, id, value, pos, size, style, validator, name);
    }

    bool Create(wxWindow *parent, wxWindowID id,
                const wxString& value = wxEmptyString,
                const wxPoint& pos = wxDefaultPosition,
                const wxSize& size = wxDefaultSize,
                long style = 0,
                const wxValidator& validator = wxDefaultValidator,
                const wxString& name = wxTextCtrlNameStr);

	~wxReTextCtrl();

    
	WX_DEFINE_SCROLL_REDIRECT()

	DECLARE_ABSTRACT_CLASS(wxReTextCtrl)

	
};

// ----------------------------------------------------------------------------
// wxReTextCtrl creation
// ----------------------------------------------------------------------------
IMPLEMENT_DYNAMIC_CLASS(wxReTextCtrl, wxTextCtrl)
WX_IMPLEMENT_SCROLL_REDIRECT(wxReTextCtrl, wxTextCtrl)

wxReTextCtrl::wxReTextCtrl()
{
    m_ScrollController = (wxWindow*) NULL;
}

bool wxReTextCtrl::Create(wxWindow *parent, wxWindowID id,
                const wxString& value,
                const wxPoint& pos,
                const wxSize& size,
                long style,
                const wxValidator& validator,
                const wxString& name)
{
    if (parent && KnownController(parent))
		m_ScrollController = parent;
	else
		m_ScrollController = this;


    return wxTextCtrl::Create(parent, id, value, pos, size, style, validator, name);
}
// ----------------------------------------------------------------------------
wxReTextCtrl::~wxReTextCtrl()
{
	// All done by base!
}
// ----------------------------------------------------------------------------

// ****************************************************************************
// ** Test Redirection Macro. Create wxMyReScrolledWindow
// ****************************************************************************
class wxMyReScrolledWindow : public wxScrolledWindow
{
public:
	wxMyReScrolledWindow();
	wxMyReScrolledWindow( wxWindow *parent,
						wxWindowID id = -1,
						const wxPoint& pos = wxDefaultPosition,
						const wxSize& size = wxDefaultSize,
						long style = wxScrolledWindowStyle,
						const wxString& name = wxPanelNameStr)
	{
        Create(parent, id, pos, size, style, name);
    }

	~wxMyReScrolledWindow();

    bool Create(wxWindow *parent,
                wxWindowID id,
                const wxPoint& pos = wxDefaultPosition,
                const wxSize& size = wxDefaultSize,
                long style = wxScrolledWindowStyle,
                const wxString& name = wxPanelNameStr);
	
	WX_DEFINE_SCROLL_REDIRECT()
	DECLARE_ABSTRACT_CLASS(wxMyReScrolledWindow)	
};

// ----------------------------------------------------------------------------
// wxMyReScrolledWindow creation
// ----------------------------------------------------------------------------
IMPLEMENT_DYNAMIC_CLASS(wxMyReScrolledWindow, wxScrolledWindow)
WX_IMPLEMENT_SCROLL_REDIRECT(wxMyReScrolledWindow, wxScrolledWindow)

wxMyReScrolledWindow::wxMyReScrolledWindow()
{
    m_ScrollController = (wxWindow*) NULL;
}

bool wxMyReScrolledWindow::Create(wxWindow *parent,
                              wxWindowID id,
                              const wxPoint& pos,
                              const wxSize& size,
                              long style,
                              const wxString& name)
{
    if (parent && KnownController(parent))
		m_ScrollController = parent;
	else
		m_ScrollController = this;


    return wxScrolledWindow::Create(parent, id, pos, size, style, name);
}
// ----------------------------------------------------------------------------
wxMyReScrolledWindow::~wxMyReScrolledWindow()
{
	// All done by base!
}
// ----------------------------------------------------------------------------


// ****************************************************************************
// ** Creat my Canvas.
// ****************************************************************************
class MyCanvas: public wxScrolledWindow
{
public:
    MyCanvas() {};
    MyCanvas( wxWindow *parent, wxWindowID, const wxPoint &pos, const wxSize &size );
    ~MyCanvas();
    void OnPaint( wxPaintEvent &event );
    void OnQueryPosition( wxCommandEvent &event );
    void OnAddButton( wxCommandEvent &event );
    void OnDeleteButton( wxCommandEvent &event );
    void OnMoveButton( wxCommandEvent &event );
    void OnScrollWin( wxCommandEvent &event );
    void OnMouseDown( wxMouseEvent &event );
    void OnScroll( wxScrollWinEvent &event );

    wxButton *m_button;

    DECLARE_DYNAMIC_CLASS(MyCanvas)
    DECLARE_EVENT_TABLE()
};

// ****************************************************************************
// ** Creat My Frame
// ****************************************************************************
class MyFrame: public wxFrame
{
public:
    MyFrame();

    void OnAbout( wxCommandEvent &event );
    void OnQuit( wxCommandEvent &event );
    void OnDeleteAll( wxCommandEvent &event );
    void OnInsertNew( wxCommandEvent &event );

    MyCanvas         *m_canvas;
    wxTextCtrl       *m_log;

    DECLARE_DYNAMIC_CLASS(MyFrame)
    DECLARE_EVENT_TABLE()
};

// ****************************************************************************
// ** Creat My App
// ****************************************************************************
class MyApp: public wxApp
{
public:
    virtual bool OnInit();
};

// main program

IMPLEMENT_APP(MyApp)

// ids

#define   ID_ADDBUTTON    1
#define   ID_DELBUTTON    2
#define   ID_MOVEBUTTON   3
#define   ID_SCROLLWIN    4
#define   ID_QUERYPOS     5

#define   ID_NEWBUTTON    10

// MyCanvas

IMPLEMENT_DYNAMIC_CLASS(MyCanvas, wxScrolledWindow)

BEGIN_EVENT_TABLE(MyCanvas, wxScrolledWindow)
  EVT_PAINT(                  MyCanvas::OnPaint)
  EVT_MOUSE_EVENTS(           MyCanvas::OnMouseDown)
  EVT_BUTTON( ID_QUERYPOS,    MyCanvas::OnQueryPosition)
  EVT_BUTTON( ID_ADDBUTTON,   MyCanvas::OnAddButton)
  EVT_BUTTON( ID_DELBUTTON,   MyCanvas::OnDeleteButton)
  EVT_BUTTON( ID_MOVEBUTTON,  MyCanvas::OnMoveButton)
  EVT_BUTTON( ID_SCROLLWIN,   MyCanvas::OnScrollWin)
  EVT_SCROLLWIN(              MyCanvas::OnScroll)
END_EVENT_TABLE()

MyCanvas::MyCanvas( wxWindow *parent, wxWindowID id,
                    const wxPoint &pos, const wxSize &size )
        : wxScrolledWindow( parent, id, pos, size, wxSUNKEN_BORDER | wxTAB_TRAVERSAL, "test canvas" )
{
	// Load a bitmap.
	wxBitmap	smile_test(smile_xpm);

	// Set background.
	SetBackgroundColour( "WHEAT" );
	
	// Create test wxTabbedCtrl.
	wxTabbedCtrl *pTabCtrl = new wxTabbedCtrl( this, -1, wxPoint(10,10), wxSize(400,200));
	pTabCtrl->SetGenerateMasks(TRUE);

	// Create some new tab views.
	wxReDirectScrolledWindow *pTabView1 = new wxReDirectScrolledWindow(pTabCtrl, -1 );
	wxMyReScrolledWindow     *pTabView2 = new wxMyReScrolledWindow(pTabCtrl, -1 );
	wxListCtrl				 *pTabView3 = new wxListCtrl(pTabCtrl, -1);

	// Insert some new tabs.
	pTabCtrl->AddTab(pTabView1, "View 1", -1, &smile_test);
	pTabCtrl->AddTab(pTabView2, "View 2", -1, (wxBitmap*)NULL);
	pTabCtrl->AddTab(pTabView3, "View 3", -1, (wxBitmap*)NULL);

	// Create test wxPagingCtrl.
	wxPagingCtrl *pPagingCtrl = new wxPagingCtrl( this, -1, wxPoint(10,220), wxSize(300,100));

	// Create some new tab views.
	wxReDirectScrolledWindow *pPageView1 = new wxReDirectScrolledWindow(pPagingCtrl, -1 );
	wxMyReScrolledWindow     *pPageView2 = new wxMyReScrolledWindow(pPagingCtrl, -1 );
	wxListCtrl				 *pPageView3 = new wxListCtrl(pPagingCtrl, -1);
	wxTextCtrl				 *pPageView4 = new wxTextCtrl(pPagingCtrl, -1);

	// Insert some new tabs.
	pPagingCtrl->AddTab(pPageView1, "View 1", -1, &smile_test);
	pPagingCtrl->AddTab(pPageView2, "View 2", -1, (wxBitmap*)NULL);
	pPagingCtrl->AddTab(pPageView3, "View 3", -1, (wxBitmap*)NULL);
	pPagingCtrl->AddTab(pPageView4, "View 4", -1, (wxBitmap*)NULL);

	// *** Create other test wxPagingCtrl.
	wxPagingCtrl *pPagingCtrlIn1 = new wxPagingCtrl( pPagingCtrl, -1, wxPoint(0, 0), wxSize(300,100));
	pPagingCtrl->AddTab(pPagingCtrlIn1, "new Page", -1, (wxBitmap*)NULL);

	wxReTextCtrl			 *pPageViewIn11 = new wxReTextCtrl(pPagingCtrlIn1, -1);
	wxTextCtrl				 *pPageViewIn12 = new wxTextCtrl(pPagingCtrlIn1, -1, "scroll text", wxPoint(0, 0), wxDefaultSize, wxHSCROLL | wxTE_MULTILINE );

	pPagingCtrlIn1->AddTab(pPageViewIn11, "View 1", -1, (wxBitmap*)NULL);
	pPagingCtrlIn1->AddTab(pPageViewIn12, "View 2", -1, (wxBitmap*)NULL);

	// NOTE: wxReTextCtrl DOES NOT work... And the reason is !!! wxTextCtrl !!! 
	// Does not use the SetScrollbar and other functions... so srolling whon't 
	// work for this control. ***It's a native Ctrl***
	//pPageViewIn11->SetScrollbar(wxHORIZONTAL, 0, 16, 50, TRUE);
	//pPageViewIn12->SetScrollbar(wxHORIZONTAL, 0, 16, 50, TRUE);
	
	// *** Create other test wxPagingCtrl.
	wxPagingCtrl *pPagingCtrlIn2 = new wxPagingCtrl( pPagingCtrl, -1, wxPoint(0, 0), wxSize(300,100));
	pPagingCtrl->AddTab(pPagingCtrlIn2, "new Page", -1, (wxBitmap*)NULL);

	wxReTextCtrl			 *pPageViewIn21 = new wxReTextCtrl(pPagingCtrlIn2, -1);
	wxTextCtrl				 *pPageViewIn22 = new wxTextCtrl(pPagingCtrlIn2, -1);

	pPagingCtrlIn2->AddTab(pPageViewIn21, "View 1", -1, (wxBitmap*)NULL);
	pPagingCtrlIn2->AddTab(pPageViewIn22, "View 2", -1, (wxBitmap*)NULL);
}

MyCanvas::~MyCanvas()
{
}

void MyCanvas::OnMouseDown( wxMouseEvent &event )
{
    if (event.LeftDown())
    {
        wxPoint pt( event.GetPosition() );
        int x,y;
        CalcUnscrolledPosition( pt.x, pt.y, &x, &y );
        wxLogMessage( "Mouse down event at: %d %d, scrolled: %d %d", pt.x, pt.y, x, y );

        if ( !event.LeftIsDown() )
            wxLogMessage( "Error: LeftIsDown() should be TRUE if for LeftDown()" );
    }
}

void MyCanvas::OnPaint( wxPaintEvent &WXUNUSED(event) )
{
    wxPaintDC dc( this );
    PrepareDC( dc );
}

void MyCanvas::OnQueryPosition( wxCommandEvent &WXUNUSED(event) )
{
    wxPoint pt( m_button->GetPosition() );
    wxLogMessage( "Position of ""Query position"" is %d %d", pt.x, pt.y );
    pt = ClientToScreen( pt );
    wxLogMessage( "Position of ""Query position"" on screen is %d %d", pt.x, pt.y );
}

void MyCanvas::OnAddButton( wxCommandEvent &WXUNUSED(event) )
{
    wxLogMessage( "Inserting button at position 10,70..." );
    wxButton *button = new wxButton( this, ID_NEWBUTTON, "new button", wxPoint(10,70), wxSize(80,25) );
    wxPoint pt( button->GetPosition() );
    wxLogMessage( "-> Position after inserting %d %d", pt.x, pt.y );
}

void MyCanvas::OnDeleteButton( wxCommandEvent &event )
{
    wxLogMessage( "Deleting button inserted with ""Add button""..." );
    wxWindow *win = FindWindow( ID_NEWBUTTON );
    if (win)
       win->Destroy();
    else
       wxLogMessage( "-> No window with id = ID_NEWBUTTON found." );
}

void MyCanvas::OnMoveButton( wxCommandEvent &event )
{
    wxLogMessage( "Moving button 10 pixels downward.." );
    wxWindow *win = FindWindow( event.GetId() );
    wxPoint pt( win->GetPosition() );
    wxLogMessage( "-> Position before move is %d %d", pt.x, pt.y );
    win->Move( -1, pt.y + 10 );
    pt = win->GetPosition();
    wxLogMessage( "-> Position after move is %d %d", pt.x, pt.y );
}

void MyCanvas::OnScrollWin( wxCommandEvent &WXUNUSED(event) )
{
    wxLogMessage( "Scrolling 2 units up.\nThe white square and the controls should move equally!" );
    int x,y;
    GetViewStart( &x, &y );
    Scroll( -1, y+2 );
}

void MyCanvas::OnScroll( wxScrollWinEvent &event )
{
    if (( event.GetEventType() == wxEVT_SCROLLWIN_THUMBRELEASE ))
    {
        wxLogMessage( "Thumb released; position: %u", event.GetPosition() );
    }
    event.Skip();
}

// MyFrame

const int ID_QUIT  = 108;
const int ID_ABOUT = 109;
const int ID_DELETE_ALL = 110;
const int ID_INSERT_NEW = 111;

IMPLEMENT_DYNAMIC_CLASS( MyFrame, wxFrame )

BEGIN_EVENT_TABLE(MyFrame,wxFrame)
  EVT_MENU    (ID_DELETE_ALL, MyFrame::OnDeleteAll)
  EVT_MENU    (ID_INSERT_NEW,  MyFrame::OnInsertNew)
  EVT_MENU    (ID_ABOUT, MyFrame::OnAbout)
  EVT_MENU    (ID_QUIT,  MyFrame::OnQuit)
END_EVENT_TABLE()

MyFrame::MyFrame()
       : wxFrame( (wxFrame *)NULL, -1, "wxScrolledWindow sample",
                  wxPoint(20,20), wxSize(470,500) )
{
    wxMenu *file_menu = new wxMenu();
    file_menu->Append( ID_DELETE_ALL, "Delete all");
    file_menu->Append( ID_INSERT_NEW, "Insert new");
    file_menu->Append( ID_ABOUT, "&About..");
    file_menu->Append( ID_QUIT, "E&xit\tAlt-X");

    wxMenuBar *menu_bar = new wxMenuBar();
    menu_bar->Append(file_menu, "&File");

    SetMenuBar( menu_bar );

    CreateStatusBar(2);
    int widths[] = { -1, 100 };
    SetStatusWidths( 2, widths );

    m_canvas = new MyCanvas( this, -1, wxPoint(0,0), wxSize(100,100) );
    m_canvas->SetScrollbars( 10, 10, 50, 100 );

    m_log = new wxTextCtrl( this, -1, "This is the log window.\n", wxPoint(0,0), wxSize(100,100), wxTE_MULTILINE );
    wxLog *old_log = wxLog::SetActiveTarget( new wxLogTextCtrl( m_log ) );
    delete old_log;

    wxBoxSizer *topsizer = new wxBoxSizer( wxVERTICAL );

    topsizer->Add( m_canvas, 1, wxEXPAND );
    topsizer->Add( m_log, 0, wxEXPAND );

    SetAutoLayout( TRUE );
    SetSizer( topsizer );
}

void MyFrame::OnDeleteAll( wxCommandEvent &WXUNUSED(event) )
{
    m_canvas->DestroyChildren();
}

void MyFrame::OnInsertNew( wxCommandEvent &WXUNUSED(event) )
{
    (void)new wxButton( m_canvas, -1, "Hello", wxPoint(100,100) );
}

void MyFrame::OnQuit( wxCommandEvent &WXUNUSED(event) )
{
  Close( TRUE );
}

void MyFrame::OnAbout( wxCommandEvent &WXUNUSED(event) )
{
  (void)wxMessageBox( "wxPagingCtrl and wxTabbedCtrl demo\n"
                      "Hans Van Leemputten (c) 2001",
                      "About wxScroll Demo", wxICON_INFORMATION | wxOK );
}

//-----------------------------------------------------------------------------
// MyApp
//-----------------------------------------------------------------------------

bool MyApp::OnInit()
{
  wxFrame *frame = new MyFrame();
  frame->Show( TRUE );

  return TRUE;
}

