/////////////////////////////////////////////////////////////////////////////
// Name:        flipperb.cpp
// Purpose:     wxFlipperCtrlBase implementation
// Author:      Aleksandras Gluchovas & Hans Van Leemputten
// Modified by:
// Created:     21/May/2001
// RCS-ID:      $Id: flipperb.cpp,v 1.8 2001/08/20 10:06:23 hansvl Exp $
// Copyright:   (c) Hans Van Leemputten & Aleksandras Gluchovas
// Licence:     wxWindows license
/////////////////////////////////////////////////////////////////////////////

// ============================================================================
// declarations
// ============================================================================

// ----------------------------------------------------------------------------
// headers
// ----------------------------------------------------------------------------

#ifdef __GNUG__
    #pragma implementation "flipperb.h"
#endif

// For compilers that support precompilation, includes "wx.h".
#include "wx/wxprec.h"

#ifdef __BORLANDC__
    #pragma hdrstop
#endif

#ifndef WX_PRECOMP
    #include "wx/wx.h"
#endif

#include "wx/list.h"

#include "wx/wspp/flipperb.h"

// ============================================================================
// Implementation for the base class wxFlipperCtrlBase
// ============================================================================
IMPLEMENT_ABSTRACT_CLASS(wxFlipperCtrlBase, wxPanel)

// ----------------------------------------------------------------------------
// event tables
// ----------------------------------------------------------------------------
//BEGIN_EVENT_TABLE(wxFlipperCtrlBase, wxPanel)
//  EVT_PAINT(wxFlipperCtrlBase::OnPaint) => Declare these in the derived class!
//  EVT_SIZE (wxFlipperCtrlBase::OnSize)  => Declare these in the derived class!
//END_EVENT_TABLE()

// ----------------------------------------------------------------------------
// creation
// ----------------------------------------------------------------------------
wxFlipperCtrlBase::wxFlipperCtrlBase()
    :wxPanel()
{
    DoInitFlipperCtrl();
}
// ----------------------------------------------------------------------------
wxFlipperCtrlBase::wxFlipperCtrlBase( wxWindow *parent, wxWindowID id,
                                    const wxPoint& pos, const wxSize& size,
                                    long  style, const wxString& name)
{
    Create(parent, id, pos, size, style, name);
}
// ----------------------------------------------------------------------------
bool wxFlipperCtrlBase::Create( wxWindow *parent, wxWindowID id,
                                const wxPoint& pos, const wxSize& size,
                                long  style, const wxString& name)
{
    bool bReturn = wxPanel::Create(parent, id, pos, size, style, name);

    DoInitFlipperCtrl(bReturn);

    return bReturn;
}
// ---------------------------------------------------------------------------
wxFlipperCtrlBase::~wxFlipperCtrlBase()
{
/*
    wxNode* pTab = m_Tabs.First();
    
    while(pTab)
    {
        delete ((twTabInfo*)pTab->Data());
        pTab = pTab->Next();
    } 
*/
    m_Tabs.DeleteContents(TRUE);
    m_Tabs.Clear();
}
// ----------------------------------------------------------------------------
void wxFlipperCtrlBase::DoInitFlipperCtrl(bool bCreated)
{
    /* Do the default init in this function. */

    if(bCreated)
    {
        // Change background to that of the parent
        if(GetParent() == NULL)
            SetBackgroundColour(wxSystemSettings::GetSystemColour(wxSYS_COLOUR_BTNFACE));
        else
            SetBackgroundColour(GetParent()->GetBackgroundColour());    
    }

    // Init state variables.
    m_ActiveTab   = -1;

    // Init state layout variables
    m_LayoutType  = wxTITLE_IMG_AND_TEXT;

    // Init Size/layout related variables
    m_TitleHeight = 0;  /* Init this to the lowest possible value. Should be */
                        /* changed during the execution to the highest size  */
                        /* needed. (So the max value needed to draw a tab.)  */
    m_VertGap     = 0;  /* Make sure to take this gap into acount.           */
    m_HorizGap    = 0;  /* Make sure to take this gap into acount.           */

    // Init bitmap mask variables
    m_GenerateMasks = FALSE;
    m_MaskColor     = *wxWHITE;
}

// ----------------------------------------------------------------------------
// Misc.
// ----------------------------------------------------------------------------
int wxFlipperCtrlBase::GetTabCount()
{
    return m_Tabs.GetCount();
}
// ----------------------------------------------------------------------------
wxWindow* wxFlipperCtrlBase::GetTab(int tabNo)
{
    wxASSERT(HasTab());             // DBG::Check if there is a tab?
    wxASSERT(tabNo >= 0 && tabNo <= GetTabCount()); 
                                    // DBG::Check range of tabNo.

    return GetTabInfo(tabNo)->m_pContent;
}
// ----------------------------------------------------------------------------
wxWindow* wxFlipperCtrlBase::GetActiveTab()
{
    if (HasTab())
    {
        return GetTab(m_ActiveTab);
    }
    else
    {
        wxLogWarning(wxT("GetActiveTab == NULL, No active tab!"));
        return NULL;
    }
}
// ----------------------------------------------------------------------------
bool wxFlipperCtrlBase::HasTab()
{
    return (m_Tabs.GetCount() > 0);
}
// ----------------------------------------------------------------------------
void wxFlipperCtrlBase::SetActiveTab(int tabNo)
{
    wxASSERT(tabNo >= 0 && tabNo < GetTabCount()); // DBG:: Validate tabNo range.

    // If none debug handle it this way.
    if(tabNo >= GetTabCount())
        tabNo = GetTabCount() - 1;
    if (tabNo < 0)
        tabNo = 0;

    // Now set the value.
    m_ActiveTab = tabNo;
    RecalcLayout(TRUE);
    Refresh();
}

// ----------------------------------------------------------------------------
void wxFlipperCtrlBase::AddTab(wxWindow* pContent,
                            wxString  tabText,
                            int       tabID,              // tab ID Number
                            wxString  imageFileName,
                            long      imageType)
{
    twTabInfo* pTab = new twTabInfo();
    
    pTab->m_pContent = pContent;
    pTab->m_Text     = tabText;
    pTab->m_TabID    = tabID;
    
    if (wxFileExists(imageFileName) &&
        pTab->m_BitMap.LoadFile(imageFileName, imageType))
    {
        pTab->m_ImageFile = imageFileName;
        pTab->m_ImageType = imageType;
    }
    
    AddTab(pTab);
}
// ----------------------------------------------------------------------------
void wxFlipperCtrlBase::AddTab(wxWindow* pContent,
                            wxString  tabText,
                            int       tabID,              // tab ID Number  
                            wxBitmap* pImage  )
{
    twTabInfo* pTab = new twTabInfo();
    
    pTab->m_pContent = pContent;
    pTab->m_Text     = tabText;
    pTab->m_TabID    = tabID;
    
    if (pImage)
        pTab->m_BitMap = *pImage;
    
    AddTab(pTab);
}
// ----------------------------------------------------------------------------
void wxFlipperCtrlBase::RemoveTab(int tabNo)
{
    twTabInfo* pTab = GetTabInfo(tabNo);

    // Delete the window content
    pTab->m_pContent->Destroy();
    // Delete tab info.
    wxDELETE(pTab);
    m_Tabs.DeleteNode(m_Tabs.Item(tabNo));
    
    // Set new Active tab.
    if (m_ActiveTab >= GetTabCount())
        m_ActiveTab = GetTabCount() - 1;

    // If active tab < 0 we do not need to set a new active tab, 
    // there is no tab left to make active!
    if (m_ActiveTab >= 0)   
        SetActiveTab(m_ActiveTab);
}
// ----------------------------------------------------------------------------
void wxFlipperCtrlBase::OnTabAdded(twTabInfo* pInfo)
{
    // Note: pInfo is not used in the base class...

    if (m_ActiveTab < 0)
        SetActiveTab(0);
    else
        SetActiveTab(m_ActiveTab);
}

// ----------------------------------------------------------------------------
int wxFlipperCtrlBase::GetLayoutStyle()
{
    return m_LayoutType;
}   
// ----------------------------------------------------------------------------
void wxFlipperCtrlBase::SetLayoutStyle(int NewStyle)
{
    wxASSERT((NewStyle & wxTITLE_IMG_AND_TEXT) || 
             (NewStyle & wxTITLE_IMG_ONLY) || 
             (NewStyle & wxTITLE_BORDER_ONLY));     // DBG::Make sure style is OK

    // Do we need to redraw our control?
    if (m_LayoutType != NewStyle)
        RecalcLayout(TRUE) ;

    m_LayoutType = NewStyle;
}

// ----------------------------------------------------------------------------
void wxFlipperCtrlBase::Refresh(bool eraseBackground, const wxRect* rect)
{
    // Refresh Active tab element.
    if (HasTab())
        GetActiveTab()->Refresh(eraseBackground, rect);

    // Refresh the base class.
    wxPanel::Refresh(eraseBackground, rect);
}

// ----------------------------------------------------------------------------
void wxFlipperCtrlBase::SetGenerateMasks(bool bOn)
{
    m_GenerateMasks = bOn;
}
// ----------------------------------------------------------------------------
bool wxFlipperCtrlBase::GetGenerateMasks()
{
    return m_GenerateMasks;
}
// ----------------------------------------------------------------------------
void wxFlipperCtrlBase::SetMaskColor(const wxColour& color)
{
    m_MaskColor = color;
}
// ----------------------------------------------------------------------------
wxColour& wxFlipperCtrlBase::GetMaskColor()
{
    return m_MaskColor;
}
// ----------------------------------------------------------------------------
wxFont wxFlipperCtrlBase::GetLabelingFont()
{
    return (*wxNORMAL_FONT);
}
// ----------------------------------------------------------------------------

twTabInfo* wxFlipperCtrlBase::GetTabInfo(int tabNo)
{
    wxASSERT(HasTab());             // DBG::Check if there is a tab?
    wxASSERT(tabNo >= 0 && tabNo <= GetTabCount()); 
                                    // DBG::Check range of tabNo.

    return ((twTabInfo*)(m_Tabs.Item(tabNo)->Data()));
}
// ----------------------------------------------------------------------------
twTabInfo* wxFlipperCtrlBase::GetActiveTabInfo()
{
    if (HasTab())
    {
        return GetTabInfo(m_ActiveTab);
    }
    else
    {
        wxLogWarning(wxT("GetActiveTabInfo == NULL, No active tab!"));
        return NULL;
    }
}
// ----------------------------------------------------------------------------
void wxFlipperCtrlBase::AddTab(twTabInfo *pTabInfo)
{
    wxASSERT(pTabInfo != NULL);     // DBG::Validate tab info.

    PrepareBitmap(pTabInfo->m_BitMap);

    // Create it if it is not yet created.
    if (pTabInfo->m_pContent->GetParent() == NULL)
        pTabInfo->m_pContent->Create(this, -1);
    
    m_Tabs.Append((wxObject*)pTabInfo);
    OnTabAdded(pTabInfo);
}
// ----------------------------------------------------------------------------
void wxFlipperCtrlBase::GetTitleCoord(int *x1, int *y1, int *x2, int *y2)
{
    int width, height;
    GetClientSize(&width, &height);
    
    *x1 = m_HorizGap;
    *y1 = height - m_VertGap - m_TitleHeight;
    *x2 = width - m_HorizGap*2;
    *y2 = height - m_VertGap*2;
}
// ----------------------------------------------------------------------------
void wxFlipperCtrlBase::GetBodyCoord(int *x1, int *y1, int *x2, int *y2)
{
    int width, height;
    GetClientSize(&width, &height);
    
    *x1 = m_HorizGap;
    *y1 = m_VertGap;
    *x2 = width - m_HorizGap*2;
    *y2 = height - m_VertGap*2 - m_TitleHeight;
}
// ----------------------------------------------------------------------------
void wxFlipperCtrlBase::SizeBodyContent(int x,int y, int width, int height, bool repant)
{
    wxNode* pTabNode = m_Tabs.First();
    int nTab = 0;

    while(pTabNode)
    {
        twTabInfo& info = *((twTabInfo*)pTabNode->Data());
        
        if (nTab == m_ActiveTab)
        {
#ifdef __WXDEBUG__ // Only in Debug
            int x1, y1, x2, y2;
            
            GetBodyCoord(&x1, &y1, &x2, &y2);
            wxASSERT(x1 <= x && y1 <= y && (x2-x1) >= width && (y2-y1) >= height);
                // DBG:: Check that the size to be set fits in the body.
#endif
            info.m_pContent->SetSize(x, y, width, height, 0);
            info.m_pContent->Show(TRUE);
            info.m_pContent->Refresh();         
        }
        else
        {
            info.m_pContent->Show(FALSE);
        }
        
        pTabNode = pTabNode->Next();
        ++nTab;
    }
}
// ----------------------------------------------------------------------------
void wxFlipperCtrlBase::EasySizeBodyContent(bool repaint)
{
    // Get all coordinates ready to use.
    int curX, curY, xSize, ySize;
    GetBodyCoord(&curX, &curY, &xSize, &ySize); 

    xSize = xSize - curX;
    ySize = ySize - curY;
    
    // Make sure the body is displayed OK.
    SizeBodyContent(curX, curY, xSize, ySize, repaint);
}
// ----------------------------------------------------------------------------
int wxFlipperCtrlBase::HitTest(const wxPoint& pos)
{
    int curX, curY, endX, endY;
    GetTitleCoord(&curX, &curY, &endX, &endY);
    
    // First check if pos is in the Title area...
    if (!(pos.x >= curX && pos.y >= curY))
         // Can't do the following... it conflicts if we use a virtual hit point
         // pos.x <= endX && pos.y <= endY
        return -1;

    // Ok lets check if we hit a tab...
    int     tabNo = 0;
    wxNode* pNode = m_Tabs.First();
    
    while(pNode)
    {
        twTabInfo& tab = *((twTabInfo*)(pNode->Data()));
        
        // hit test rectangle of the currnet tab title bar
        if (pos.x >= curX && pos.x < curX + tab.m_Dims.x  &&
            pos.y >= curY && pos.y < curY + tab.m_Dims.y
            )
        {
            return tabNo;
        }
        
        curX += tab.m_Dims.x;
        
        pNode = pNode->Next();
        ++tabNo;
    }
    
    return -1;
}
// ----------------------------------------------------------------------------
void wxFlipperCtrlBase::PrepareBitmap(wxBitmap& bmp)
{
    if (!m_GenerateMasks) return;
    
    if (!bmp.Ok()) return;

    if (bmp.GetMask() != NULL) return;
    
    bmp.SetMask(new wxMask(bmp, GetMaskColor()));
}

// ----------------------------------------------------------------------------
// event handlers
// ----------------------------------------------------------------------------
void wxFlipperCtrlBase::OnPaint(wxPaintEvent& event)
{
    wxPaintDC dc(this);
    DrawDecorations(dc);
}
// ----------------------------------------------------------------------------
void wxFlipperCtrlBase::OnSize (wxSizeEvent& event)
{
    RecalcLayout(TRUE);
}


// ============================================================================
// Implementation for class twTabInfo
// ============================================================================
IMPLEMENT_DYNAMIC_CLASS(twTabInfo, wxObject)

// ---------------------------------------------------------------------------
twTabInfo::twTabInfo()
{
    m_TabID     = -1;
    m_pContent  = NULL;
    m_BitMap    = wxNullBitmap;
    m_Text      = wxEmptyString;
    m_Dims      = wxDefaultSize;
    
    m_ImageFile = wxEmptyString;
    m_ImageType = -1;

    m_HorizScrollBar = FALSE;
    m_VertScrollBar  = FALSE;
}
// ---------------------------------------------------------------------------
twTabInfo::~twTabInfo()
{
    // Nothing
}
// ---------------------------------------------------------------------------
int twTabInfo::ImgWidth()
{
    if (m_BitMap.Ok()) 
        return m_BitMap.GetWidth();
    else 
        return 0;
}
// ---------------------------------------------------------------------------
int twTabInfo::ImgHeight()
{
    if (m_BitMap.Ok()) 
        return m_BitMap.GetHeight();
    else 
        return 0;
}
// ---------------------------------------------------------------------------
int twTabInfo::ImageToTxtGap(int prefGap)
{
    if (m_BitMap.Ok()) 
        return prefGap;
    else 
        return 0;
}
// ---------------------------------------------------------------------------
bool twTabInfo::HasImg()
{
    return m_BitMap.Ok();
}
// ---------------------------------------------------------------------------
bool twTabInfo::HasText()
{
    return !m_Text.IsEmpty();
}
// ---------------------------------------------------------------------------
bool twTabInfo::HandleHoriz()
{
    return m_HorizScrollBar;
}
// ---------------------------------------------------------------------------
bool twTabInfo::HandleVert()
{
    return m_VertScrollBar;
}
// ---------------------------------------------------------------------------
wxBitmap& twTabInfo::GetImg()
{
    return m_BitMap;
}
// ---------------------------------------------------------------------------
wxString& twTabInfo::GetText()
{
    return m_Text;
}
// ---------------------------------------------------------------------------
wxWindow& twTabInfo::GetContent()
{
    return *m_pContent;
}

