/////////////////////////////////////////////////////////////////////////////
// Name:        pagectrl.cpp
// Purpose:     wxPagingCtrl implementation
// Author:      Aleksandras Gluchovas & Hans Van Leemputten
// Modified by:
// Created:     21/May/2001
// RCS-ID:      $Id: pagectrl.cpp,v 1.15 2001/09/15 09:25:55 hansvl Exp $
// Copyright:   (c) Hans Van Leemputten & Aleksandras Gluchovas
// Licence:     wxWindows license
/////////////////////////////////////////////////////////////////////////////

// ============================================================================
// declarations
// ============================================================================

// ----------------------------------------------------------------------------
// headers
// ----------------------------------------------------------------------------

#ifdef __GNUG__
    #pragma implementation "pagectrl.h"
#endif

// For compilers that support precompilation, includes "wx.h".
#include "wx/wxprec.h"

#ifdef __BORLANDC__
    #pragma hdrstop
#endif

#ifndef WX_PRECOMP
    #include "wx/wx.h"
#endif

#include "wx/wspp/espinbut.h"
#include "wx/wspp/pagectrl.h"

#include <math.h>

// ============================================================================
// wxTabbedCtrl implementation.
// ============================================================================
IMPLEMENT_DYNAMIC_CLASS(wxPagingCtrl, wxFlipperCtrlBase)

// ----------------------------------------------------------------------------
// event tables
// ----------------------------------------------------------------------------
BEGIN_EVENT_TABLE(wxPagingCtrl, wxFlipperCtrlBase)
    EVT_PAINT     (wxFlipperCtrlBase::OnPaint )
    EVT_SIZE      (wxFlipperCtrlBase::OnSize  )
    EVT_LEFT_DOWN (wxPagingCtrl::OnLButtonDown)
    EVT_LEFT_UP   (wxPagingCtrl::OnLButtonUp  )
    EVT_RIGHT_DOWN(wxPagingCtrl::OnRButtonDown)
    EVT_MOTION    (wxPagingCtrl::OnMouseMove  )
    EVT_SPIN      (-1, wxPagingCtrl::OnTabScroll)
    EVT_SCROLL    (wxPagingCtrl::OnScroll     )    
END_EVENT_TABLE()

// ----------------------------------------------------------------------------
// creation
// ----------------------------------------------------------------------------
wxPagingCtrl::wxPagingCtrl()
    : wxFlipperCtrlBase(),
      m_bInitDone(FALSE)     // Not finished with Init.
{
    DoInitPagingCtrl();
}
// ----------------------------------------------------------------------------
wxPagingCtrl::wxPagingCtrl( wxWindow *parent,
                            wxWindowID id,
                            const wxPoint& pos,
                            const wxSize& size,
                            long  style,
                            const wxString& name)
    : wxFlipperCtrlBase(),   // Tmp creation of the base class.
      m_bInitDone(FALSE)     // Not finished with Init.
{
    Create(parent, id, pos, size, style, name);
}
// ----------------------------------------------------------------------------
wxPagingCtrl::~wxPagingCtrl()
{
    /* cleanup scrollbars and spin button. */
    wxDELETE(m_pTabScroll);
    wxDELETE(m_pHorizScroll);
    wxDELETE(m_pVertScroll);
}
// ----------------------------------------------------------------------------
bool wxPagingCtrl::Create( wxWindow *parent, wxWindowID id,
                        const wxPoint& pos,
                        const wxSize& size,
                        long  style,
                        const wxString& name)
{
    bool bResult = wxFlipperCtrlBase::Create(parent, id, pos, size, style, name);

    DoInitPagingCtrl();

    // Create the spin button and scroll bars for this control.
    m_pTabScroll   =
            new wxEnhSpinButton(this, -1);
    m_pHorizScroll =
            new wxScrollBar(this, -1, wxDefaultPosition, wxDefaultSize, wxSB_HORIZONTAL);
    m_pVertScroll = 
            new wxScrollBar(this, -1, wxDefaultPosition, wxDefaultSize, wxSB_VERTICAL);

    // Hide them until we know the can be displayed.
    m_pTabScroll->Show(FALSE);
    m_pHorizScroll->Show(FALSE);    // Note: changing these will not help
    m_pVertScroll->Show(FALSE);     //       You need to change these also in SetActiveTab...

    return bResult;
}
// ----------------------------------------------------------------------------
void wxPagingCtrl::DoInitPagingCtrl()
{
    // Init the pen's.
    m_WhitePen = wxPen(wxColour(255,255,255), 1, wxSOLID);
    m_GrayPen  = wxPen(wxColour(192,192,192), 1, wxSOLID);
    m_DarkPen  = wxPen(wxColour(128,128,128), 1, wxSOLID);
    m_BlackPen = wxPen(wxColour( 0,  0,  0), 1, wxSOLID);
    m_TabPen   = wxPen(wxSystemSettings::GetSystemColour(wxSYS_COLOUR_BTNFACE), 1, wxSOLID);
    m_ActiveTabPen = wxPen(wxSystemSettings::GetSystemColour( wxSYS_COLOUR_WINDOW ), 1, wxSOLID);

    // Init the tab brush.
    m_TabBrush = wxBrush(m_TabPen.GetColour(), wxSOLID);
    m_ActiveTabBrush = wxBrush(m_ActiveTabPen.GetColour(), wxSOLID);

    // Protected Init Size/layout related variables
    m_TabTrianGap  = 5; /* Gap that causes the paper tab to be a trapezium.  */
    m_ImageTextGap = 2; /* Make sure to take this gap into acount.           */
    m_TitleHeight  = 16;/* Make sure to set it to the max value of all tabs  */ //FORNOW!!! Hard coded...
    m_TitleHorizGap= 10;/* This is a global class var.                       */
    m_TitleVertGap = 2; /* This is a global class var.                       */
    m_CurentRowOfs = 0; /* Display offset of the visble title bar part.      */
                        /* Space available for the title bar                 */
    m_TitleSpace = GetSize().GetWidth(); // Old... m_TitleSpace   = 300;
    
    // Private Init Size/layout related variables
    m_TabScrollRate = 20;   /* Use the Setxxx and Getxxx and not his varible!*/
    m_bIsDragged    = FALSE;/* We are not dragging the tab resizer. so FALSE.*/ 
    m_bCursorChanged= FALSE;/* No we did not change the cursor, so FALSE.    */

    // Init Cursor's
    m_ResizeCursor = wxCURSOR_SIZEWE;
    m_NormalCursor = wxCURSOR_ARROW;

    // Init Scroll redirection stuff.
    m_bScrollResult = FALSE;/* Result of the last scroll redirection.        */
    m_pCaller       = NULL; /* Last scroll redirection caller.               */
    m_bScrollRefresh= FALSE;/* Are we in scrollbar refresh mode...           */ 

    // Init all spin and scroll bars.
    m_pTabScroll    = NULL;
    m_pHorizScroll  = NULL;
    m_pVertScroll   = NULL;

    // Init the background color in the same color as the Tab backgound.
    SetBackgroundColour(m_TabPen.GetColour());

    // Ok now we are done with Init.
    m_bInitDone = TRUE;
}

// ----------------------------------------------------------------------------
// Misc.
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// width of the decorations border (2 shade-lines), should not be changed
// ----------------------------------------------------------------------------
#define BORDER_SZ 2
// ----------------------------------------------------------------------------
void wxPagingCtrl::DrawShadedRect(int x, int y, int width, int height,
                                    wxPen& upperPen, wxPen& lowerPen, wxDC& dc
                                    )
{
    // darw the lightened upper-left sides of the rectangle
    dc.SetPen(upperPen);
    dc.DrawLine(x,y, x, y + height - 1); // vert
    dc.DrawLine(x,y, x + width - 1,  y); // horiz
    
    // draw the unenlightened lower-right sides of the rectangle    
    dc.SetPen(lowerPen);
    dc.DrawLine(x + width - 1,  y, x + width - 1, y + height - 1); // vert
    dc.DrawLine(x, y + height - 1, x + width, y + height - 1);     // horiz
}
// ----------------------------------------------------------------------------
void wxPagingCtrl::DrawPaperBar(twTabInfo& tab, int x, int y,
                                wxBrush& brush, wxPen& pen, wxDC& dc)
{
    wxPoint poly[4];
    
    /********************************************************/
    /* step #1 - draw organizer-style paper outlet          */
    /********************************************************/
    
    poly[0].x = x - m_TabTrianGap;
    poly[0].y = y;
    
    poly[1].x = x + m_TabTrianGap;
    poly[1].y = y + tab.m_Dims.y-1;
    
    poly[2].x = x + tab.m_Dims.x - m_TabTrianGap;
    poly[2].y = y + tab.m_Dims.y-1;
    
    poly[3].x = x + tab.m_Dims.x + m_TabTrianGap;
    poly[3].y = y;
    
    dc.SetPen(pen);
    dc.SetBrush(brush);
    
    dc.DrawPolygon(4, poly);

    /********************************************************/
    /* step #2 - Draw Image in tab if any.                  */
    /********************************************************/
    
    long w,h;
    
    // set select default font of the window into it's device context
    dc.SetFont(GetLabelingFont());
    
    dc.SetTextBackground(brush.GetColour());
    
    dc.GetTextExtent(tab.m_Text, &w, &h);
    
    if (tab.HasImg() && tab.GetImg().Ok())
    {
        wxBitmap& bmp = tab.GetImg();
        
        bool hasMask = bmp.GetMask() != NULL;
        
        dc.DrawBitmap(bmp, x + m_TitleHorizGap,
            y + (tab.m_Dims.y - tab.ImgHeight()) / 2,
            hasMask);
    }

    /********************************************************/
    /* step #3 - Draw Text in tab if any.                   */
    /********************************************************/
    
    if (tab.HasText())
    {
        int tx = x + m_TitleHorizGap +
            tab.ImgWidth() + tab.ImageToTxtGap(m_ImageTextGap);
        
        dc.DrawText(tab.GetText(), tx, y + (tab.m_Dims.y - h) / 2);
    }
}
// ----------------------------------------------------------------------------
void wxPagingCtrl::GetTitleCoord(int *x1, int *y1, int *x2, int *y2)
{
    wxFlipperCtrlBase::GetTitleCoord(x1, y1, x2, y2);

    // Remove the size from the border.
    *x1 += BORDER_SZ;
    *x2 = *x1 + m_TitleSpace;
    // Shift the title bar up, the size of the border.
    *y1 -= BORDER_SZ;   
    *y2 -= BORDER_SZ;   


    // Correct the already calculated values if m_pTabScroll is shown. (Value of x1)
    if (m_pTabScroll != NULL && m_pTabScroll->IsShown())
    {
        int sizeX, sizeY, tempY;

        // Get the displayed position in the control.
        // Use tempY instead of y1, this way the title 
        // height is not depending on the height of the scrollbar.
        m_pTabScroll->GetPosition(x1, &tempY);

        // Add the with of the scrollbar to x1
        m_pTabScroll->GetSize(&sizeX, &sizeY);
        *x1 += sizeX;
    }

    // This causes problems when the window is minimized.
    // wxASSERT(*x2 >= *x1 && *x1 >= 0); // DBG:: Some debug validation...
    // wxASSERT(*y2 >= *y1 && *y1 >= 0); // DBG:: Some debug validation...
    // So no more debug checking, change it by real checking
    if (*x1 < 0) *x1 = 0;
    if (*x2 < 0) *x2 = 0;
    if (*y1 < 0) *y1 = 0;
    if (*y2 < 0) *y2 = 0;
    if (*y2 < *y1) *y2 = *y1;
    if (*x2 < *x1) *x2 = *x1;
}
// ----------------------------------------------------------------------------
void wxPagingCtrl::GetBodyCoord(int *x1, int *y1, int *x2, int *y2)
{
    // NOTE: This function can not be used to calc the psition of m_pVertScroll!!!
    //       This is because m_pVertScroll position is used to clac the return values.

    wxFlipperCtrlBase::GetBodyCoord(x1, y1, x2, y2);

    // Remove the size from the border.
    *x1 += BORDER_SZ;
    *x2 -= BORDER_SZ;
    *y1 += BORDER_SZ;
    *y2 -= BORDER_SZ;

    // And subtract the width of the Vert. Scrollbar if this one is displayed.
    if (m_pVertScroll != NULL && m_pVertScroll->IsShown())
    {
        int tempY;

        // Get the displayed position in the control.
        m_pVertScroll->GetPosition(x2, &tempY);
    }
}
// ----------------------------------------------------------------------------
void wxPagingCtrl::GetCtrlCoord(int *x1, int *y1, int *x2, int *y2)
{
    int width, height;
    GetClientSize(&width, &height);
    
    *x1 = m_HorizGap;
    *y1 = m_VertGap;
    *x2 = width - m_HorizGap*2;
    *y2 = height - m_VertGap*2;
}
// ----------------------------------------------------------------------------
int wxPagingCtrl::GetWholeTabRowLen(bool addTabSizer)
{
    wxNode* pNode = m_Tabs.First();
    
    int nLen = 0;

    if (pNode != NULL)
        nLen += m_TabTrianGap*2;
    
    while(pNode)
    {
        twTabInfo& tab = *((twTabInfo*)(pNode->Data()));
        
        nLen += tab.m_Dims.x;
        pNode = pNode->Next();
    }
    
    // If the the resize stick is shown adjust the len, so it adds the RESIZE_STICK_SIZE.
    if (addTabSizer && m_pHorizScroll != NULL && m_pHorizScroll->IsShown())        
        nLen += RESIZE_STICK_SIZE;

    return nLen;
}

// ----------------------------------------------------------------------------
void wxPagingCtrl::DrawDecorations(wxDC& dc)
{
    if (!m_bInitDone)
        return;     // Hell we are not ready to handle this!

    int width, height;
    GetClientSize(&width, &height);
    
    // check if there's at least a bit of space to draw things
    if (width  < m_HorizGap*2 + BORDER_SZ*2+1 ||
        height < m_VertGap*2  + BORDER_SZ*2+1 + m_TitleHeight)
        return;

    /********************************************************/
    /* step #1 - draw border around the tab content area    */
    /********************************************************/

    // setup position for kind of "pencil"
    int curX, curY, xSize, ySize;
    GetCtrlCoord(&curX, &curY, &xSize, &ySize);

    // Draw border around the total control...  
    DrawShadedRect(curX, curY, xSize, ySize,
        m_DarkPen, m_WhitePen, dc);
    
    DrawShadedRect(curX+1, curY+1, xSize-2, ySize-2,
        m_BlackPen, m_GrayPen, dc);
    
    /****************************************************************/
    /* step #2 - draw inactive tab title bars frist (left-to-right) */
    /****************************************************************/
    
    int xMaxSize, yMaxSize;
    GetTitleCoord(&curX, &curY, &xMaxSize, &yMaxSize);

    int nTitleHeight = yMaxSize - curY; // Same as m_TitleHeight
    int nTitleRowLen = xMaxSize - curX; 

    wxNode* pNode = m_Tabs.First();
    int     tabNo = 0;

    curX = m_TabTrianGap;
    curY = 0;
    
    // FOR NOW:: avoid creating bitmap with invalid dimensions
    if (nTitleRowLen < 1 || nTitleHeight < 1) return;
    
    wxMemoryDC tmpDc;
    wxBitmap   tmpBmp(nTitleRowLen, nTitleHeight);
    
    // Clean the background of the memory bitmap.
    tmpDc.SelectObject(tmpBmp);
    tmpDc.SetPen(m_GrayPen);
    tmpDc.SetBrush(m_TabBrush);
    tmpDc.DrawRectangle(0,0, nTitleRowLen, nTitleHeight);
    
    tmpDc.SetDeviceOrigin(m_CurentRowOfs, 0);
    
    while(pNode)
    {
        twTabInfo& tab = *((twTabInfo*)(pNode->Data()));
        
        if (tabNo != m_ActiveTab)
            DrawPaperBar(tab, curX, curY, m_TabBrush, m_BlackPen, tmpDc);
        
        curX += tab.m_Dims.x;
        
        pNode = pNode->Next();
        ++tabNo;
    }
    
    /********************************************************/
    /* step #2 - draw the active tab (white-filled)         */
    /********************************************************/
    
    pNode = m_Tabs.First();
    tabNo = 0;
    
    curX = m_TabTrianGap;
    
    while(pNode)
    {
        twTabInfo& tab = *((twTabInfo*)(pNode->Data()));
        
        if (tabNo == m_ActiveTab)
        {
            tmpDc.SetBrush(m_ActiveTabBrush);
            
            DrawPaperBar(tab, curX, curY, m_ActiveTabBrush, m_BlackPen, tmpDc);
            
            // Because DrawPaperBar resets the pen, we need to set the pen here.
            tmpDc.SetPen(m_ActiveTabPen);
            
            tmpDc.DrawLine(curX - m_TabTrianGap+1, curY,
                curX + tab.m_Dims.x + m_TabTrianGap, curY);
            break;
        }
        curX += tab.m_Dims.x;
        
        pNode = pNode->Next();
        ++tabNo;
    }
    
    // back to initial device origin
    tmpDc.SetDeviceOrigin(0, 0);
    
    // Get ready to do the real drawing stuff.
    GetTitleCoord(&curX, &curY, &xMaxSize, &yMaxSize);
    
    // Calc it to what we will be needing.
    nTitleHeight = yMaxSize - curY;
    nTitleRowLen = xMaxSize - curX; 

    // Draw a resize stick if the scrollbar is shown.
    if (m_pHorizScroll != NULL && m_pHorizScroll->IsShown())
    {
        wxASSERT(xMaxSize > RESIZE_STICK_SIZE); // DBG:: Make sure there is enough space to do this!
        
        // Continue with Calc.
        xMaxSize = nTitleRowLen - RESIZE_STICK_SIZE;
        
        // draw resize-hint-stick
        DrawShadedRect(xMaxSize+0, 0+0, RESIZE_STICK_SIZE,   nTitleHeight,   m_GrayPen,  m_BlackPen, tmpDc);
        DrawShadedRect(xMaxSize+1, 0+1, RESIZE_STICK_SIZE-2, nTitleHeight-2, m_WhitePen, m_DarkPen,  tmpDc);
        DrawShadedRect(xMaxSize+2, 0+2, RESIZE_STICK_SIZE-4, nTitleHeight-4, m_GrayPen,  m_GrayPen,  tmpDc);
    }
    
    // Draw the memory tab buffer to the screen.
    dc.Blit(curX, curY,
            nTitleRowLen, nTitleHeight,
            &tmpDc, 0,0, wxCOPY);   
}
// ----------------------------------------------------------------------------
void wxPagingCtrl::RecalcLayout(bool andRepaint)
{
    if (!m_bInitDone)
        return;     // Hell we are not ready to handle this!
        
    // layout tab title bars
    /****************************************************************/
    /* pass #1 - try to layout assuming it's wxTITLE_IMG_AND_TEXT   */
    /****************************************************************/

    if (m_LayoutType == wxTITLE_IMG_AND_TEXT)
    {   
        wxNode* pNode = m_Tabs.First();
        
        while(pNode)
        {
            twTabInfo& tab = *((twTabInfo*)(pNode->Data()));
            
            wxWindowDC dc(this);
            
            long w,h;
            
            // set select default font of the window into it's device context
            dc.SetFont(GetLabelingFont());
            dc.GetTextExtent(tab.m_Text, &w, &h);
            
            tab.m_Dims.x = w + tab.ImageToTxtGap(m_ImageTextGap) +
                tab.ImgWidth() + m_TitleHorizGap*2;
            
            tab.m_Dims.y = m_TitleHeight;
            
            pNode = pNode->Next();
        }
    }

    /****************************************************************/  
    /* pass #2 - try to layout assuming wxTITLE_IMG_ONLY            */
    /****************************************************************/

    if (m_LayoutType == wxTITLE_IMG_ONLY)
    {
        wxNode* pNode = m_Tabs.First();
        
        while(pNode)
        {
            twTabInfo& tab = *((twTabInfo*)(pNode->Data()));
            
            if (tab.HasImg())
                tab.m_Dims.x = tab.ImgWidth() + m_TitleHorizGap*2;
            else
                tab.m_Dims.x = 30; // Hardcoded!!!          
            tab.m_Dims.y = m_TitleHeight;
            
            pNode = pNode->Next();
        }
    }

    /****************************************************************/
    /* pass #3 - set the narrowest layout wxTITLE_BORDER_ONLY       */
    /****************************************************************/

    if (m_LayoutType == wxTITLE_BORDER_ONLY)
    {
        wxNode* pNode = m_Tabs.First();
        
        while(pNode)
        {
            twTabInfo& tab = *((twTabInfo*)(pNode->Data()));
            
            tab.m_Dims.x = 30; // Hardcoded!!!          
            tab.m_Dims.y = m_TitleHeight;
            
            pNode = pNode->Next();
        }
    }
    
    // Adjust Horiz and Vert settings if needed.
    RecalcHorizVertScroll();

    // Adjust TabScroll settings.
    RecalcTabScroll();
    
    // Make sure this gets done!
    // Defiantly needed when recalcing layout on tab added!
    // re-size the body content to the new Body size, if any...
    EasySizeBodyContent(andRepaint);

    // Move Focus to the active tab contol.
    if (HasTab())
        GetActiveTab()->SetFocus();
}
// ----------------------------------------------------------------------------
void wxPagingCtrl::RecalcTabScroll()
{
    // Adjust m_pTabScroll settings if already created.
    if(m_pTabScroll != NULL)
    {   
        int curX, curY, xMaxSize, yMaxSize;

        // Use the Title space that was set in wxFlipperCtrlBase.
        wxFlipperCtrlBase::GetTitleCoord(&curX, &curY, &xMaxSize, &yMaxSize);

        // TODO:: m_TitleHeight * 4 => the hard coded 4 should change if a 
        //                             other style is used for the spinbutton

        // Make sure there is enough space in the control title to set the scrollbar...
        if((curY - BORDER_SZ) + m_TitleHeight > yMaxSize || 
            curX + BORDER_SZ + m_TitleHeight * 4 > xMaxSize)
        {
            wxLogError(wxT("Tab Scrollbar to big for the current size of wxPagingCtrl!"));
            m_pTabScroll->Show(FALSE);
        }
        else
        {
            if(!m_pTabScroll->IsShown())
                m_pTabScroll->Show(TRUE);
        }

        // Calc the new size, location of m_pTabScroll.
        m_pTabScroll->SetSize(curX + BORDER_SZ, curY - BORDER_SZ, m_TitleHeight * 4, m_TitleHeight);
        
        // disable title-bar scroller if there's nowhere to scroll to
        // Get the available draw size. = (xMaxSize - curX)
        GetTitleCoord(&curX, &curY, &xMaxSize, &yMaxSize);  
        m_pTabScroll->Enable((xMaxSize - curX) < GetWholeTabRowLen(TRUE) || m_CurentRowOfs < 0 );           
    }
}
// ----------------------------------------------------------------------------
void wxPagingCtrl::RecalcHorizVertScroll()
{
    int curX, curY, xMaxSize, yMaxSize, tempX;

    // Use the Title space that was set in wxFlipperCtrlBase.
    wxFlipperCtrlBase::GetTitleCoord(&curX, &curY, &xMaxSize, &yMaxSize);

    // Adjust m_pTabScroll settings if already created and one of them is shown.
    if (m_pHorizScroll != NULL && m_pVertScroll != NULL &&
        (m_pHorizScroll->IsShown() || m_pVertScroll->IsShown())) 
    {
        // Make sure there is enough space in the control title to set the scrollbar...
        if((curY - BORDER_SZ) + m_TitleHeight > yMaxSize || 
            curX + BORDER_SZ + m_TitleHeight * 3 > xMaxSize)
        {
            wxLogError(wxT("Scrollbars too big for the current size of wxPagingCtrl!"));
            m_pHorizScroll->Show(FALSE);
            m_pVertScroll->Show(FALSE);
        }   
        
        // Set size of Horiz scrollbar
        int nHorizScrollSize = wxMax(m_TitleHeight * 4, xMaxSize - curX - BORDER_SZ*2 - m_TitleSpace);

        if (m_pHorizScroll->IsShown() && m_pVertScroll->IsShown())
        {
            // We have to leave gap behinde the scrollbar...
            tempX = xMaxSize - BORDER_SZ - nHorizScrollSize;
            m_pHorizScroll->SetSize(tempX, curY - BORDER_SZ, xMaxSize - tempX - BORDER_SZ - m_TitleHeight, m_TitleHeight);
            
            GetCtrlCoord(&curX, &curY, &xMaxSize, &yMaxSize);
            m_pVertScroll->SetSize(xMaxSize - BORDER_SZ - m_TitleHeight, curY + BORDER_SZ, m_TitleHeight, yMaxSize - curY - BORDER_SZ*2 - m_TitleHeight);
        }
        else
        {
            if (m_pHorizScroll->IsShown()) //  && !m_pVertScroll->IsShown()
            {
                // We do not have to leave gap behinde the scrollbar...
                tempX = xMaxSize - BORDER_SZ - nHorizScrollSize;
                m_pHorizScroll->SetSize(tempX, curY - BORDER_SZ, xMaxSize - tempX - BORDER_SZ, m_TitleHeight);
            }
            else
            {
                // We do not have to leave gap behinde the scrollbar...
                GetCtrlCoord(&curX, &curY, &xMaxSize, &yMaxSize);

                tempX = xMaxSize - m_TitleHeight - BORDER_SZ;
                m_pVertScroll->SetSize(tempX, curY + BORDER_SZ, m_TitleHeight, yMaxSize - curY - BORDER_SZ*2);
            }
        }

        // Update the Title space.
        wxFlipperCtrlBase::GetTitleCoord(&curX, &curY, &xMaxSize, &yMaxSize);
        m_TitleSpace = tempX - BORDER_SZ - curX;
    }
    else
    {
        // Update the Title space.
        m_TitleSpace = xMaxSize - BORDER_SZ*2 - curX;
    }
}
// ----------------------------------------------------------------------------
void wxPagingCtrl::OnTabAdded(twTabInfo* pInfo)
{
    // Set as default to handle scrollbars (only if possible).
    pInfo->m_VertScrollBar = TRUE;
    pInfo->m_HorizScrollBar= TRUE;

    // Call base OnTabAdded
    wxFlipperCtrlBase::OnTabAdded(pInfo);

    // Calc the settings of the m_pTabScroll spin button.
    SetTabScrollRate();
}
// ----------------------------------------------------------------------------
void wxPagingCtrl::SetActiveTab(int tabNo)
{
    // Only do this if we have a new active tab
    if (tabNo != m_ActiveTab)
    {
        // Re int these if a new tab is active.
        m_pHorizScroll->Show(FALSE);
        m_pVertScroll->Show(FALSE);

        wxASSERT(tabNo >= 0 && tabNo <= GetTabCount()); // DBG:: Validate tabNo range.

        // Set new Active tab.
        m_ActiveTab = tabNo;

        // Let the control refresh his scroll bars.
        ScrollRefresh();
    }

    // Call base SetActiveTab
    wxFlipperCtrlBase::SetActiveTab(tabNo);     
}

// ----------------------------------------------------------------------------
void wxPagingCtrl::SetTabScrollRate(int nRate)
{
    wxASSERT(m_pTabScroll != NULL); // DBG:: Make sure this is OK.

    if (nRate != 0)
        m_TabScrollRate = nRate;

    // (Re)Calc the settings of the m_pTabScroll spin button.
    // * 1.0 to make sure it divids in floats.
    int units = (int)ceil(GetWholeTabRowLen() * 1.0 / m_TabScrollRate);  
    m_pTabScroll->SetRange(0, units);
}
// ----------------------------------------------------------------------------
int wxPagingCtrl::GetTabScrollRate()
{
    // Nothing more nothing less...
    return m_TabScrollRate;
}

// ----------------------------------------------------------------------------
void wxPagingCtrl::SetScrollBars(int tabID, bool bHorizScroll, bool bVertScroll)
{
    wxASSERT_MSG(tabID <= GetTabCount(), wxT("SetScrollBars Tab ID invalid!")); // DBG::

    if (!HasTab())
    {
        wxLogWarning(wxT("Action is ignored, no Tab in wxTabbedWindow control!"));
        return;
    }

    if (tabID == -1)
        tabID = GetTabCount() - 1;

    // Update Tab info.
    GetTabInfo(tabID)->m_HorizScrollBar = bHorizScroll;
    GetTabInfo(tabID)->m_VertScrollBar  = bVertScroll;

    // Refresh srollbars
    if (tabID == m_ActiveTab)
        ScrollRefresh();

    // RecalcLayout...
    RecalcLayout(TRUE);
}
// ----------------------------------------------------------------------------
void wxPagingCtrl::ScrollRefresh()
{
    // Get the Scrollbars update... with a trick
    // I hope this works for all controls with aware scrolling...
    if (HasTab())
    {
        wxWindow *pActiveTab = GetActiveTab();

        wxSizeEvent event; // Fake event. It is just to get OnSize to be called.
        pActiveTab->GetEventHandler()->ProcessEvent(event);

        // Remove scrollbars from control if needed
        if (GetActiveTabInfo()->HandleHoriz())      
        {
            // Don't let the following scroll actions be handled by this class
            m_bScrollRefresh = TRUE;
            pActiveTab->SetScrollbar(wxHORIZONTAL, 0, 0, 0, TRUE);
        }
        if (GetActiveTabInfo()->HandleVert())
        {
            // Don't let the following scroll actions be handled by this class
            m_bScrollRefresh = TRUE;
            pActiveTab->SetScrollbar(wxVERTICAL, 0, 0, 0, TRUE);
        }

        // Set state variable back, now we are no longer in client scrollbar refresh mode.
        m_bScrollRefresh = FALSE;   // Normally it already has this value back... 
                                    // but we need to be sure it is really the case.

        // Make sure the current active tab gets active.
        pActiveTab->SetFocus();
    }
}

// ----------------------------------------------------------------------------
// Extra functions for scroll communication.
// ----------------------------------------------------------------------------
bool wxPagingCtrl::ResultLastScrollAction()
{
    bool bScrollResult = m_bScrollResult;
    
    // This function can only be called once for every scroll 
    // function redirection, so clean it...
    m_bScrollResult = FALSE; 

    // Clean this too
    m_pCaller = NULL;
    
    return bScrollResult;
}
// ----------------------------------------------------------------------------
void wxPagingCtrl::SetNextCaller(wxWindow *pCaller)
{
    wxASSERT(pCaller != NULL); // DBG:: Validate input.

    m_pCaller = pCaller;
}

// ----------------------------------------------------------------------------
// New scrolling implementation.
// ----------------------------------------------------------------------------
void wxPagingCtrl::SetScrollbar(int orientation, int position, int thumbSize, int range, bool refresh)
{
    // Note::pageSize is set to thumbSize, because its not known...
        
    m_bScrollResult = FALSE;

    // Make sure we have a tab and the active tab is the next caller.
    if (!HasTab())
        return;
    if (m_pCaller != GetActiveTab())
        return;

    // Check if we may continue handling this scroll action.
    if (m_bScrollRefresh)
    {
        // Put m_bScrollRefresh to false to avoid recursive issues!
        m_bScrollRefresh = FALSE;

        return;
    }

    // make some state variables
    bool bNewIsShown = (thumbSize < range);
    bool bOldIsShown = FALSE;
    
    // Test and set new values for the scrollbars.
    if (orientation == wxVERTICAL && m_pVertScroll && GetActiveTabInfo()->HandleVert())
    {
        // Set our scrollbar with the redirected values
        m_pVertScroll->SetScrollbar(position, thumbSize, range, thumbSize, refresh);

        // Now handle the showing and recalc of the layout.
        bOldIsShown = m_pVertScroll->IsShown();

        // Do we need to update any thing?
        if (bNewIsShown != bOldIsShown)
        {
            m_pVertScroll->Show(bNewIsShown);
            RecalcLayout(TRUE);
        }       

        // Set m_bScrollResult here, because RecalcLayout(TRUE) can cause it to 
        // be set back to FALSE, a recursive thing :-)
        m_bScrollResult = TRUE;
    }
    
    if (orientation == wxHORIZONTAL && m_pHorizScroll && GetActiveTabInfo()->HandleHoriz())
    {
        // Set our scrollbar with the redirected values
        m_pHorizScroll->SetScrollbar(position, thumbSize, range, thumbSize, refresh);
                
        // Now handle the showing and recalc of the layout.
        bOldIsShown = m_pHorizScroll->IsShown();

        // Do we need to update any thing?
        if (bNewIsShown != bOldIsShown)
        {
            m_pHorizScroll->Show(bNewIsShown);
            RecalcLayout(TRUE);
        }  
        
        // Set m_bScrollResult here, because RecalcLayout(TRUE) can cause it to 
        // be set back to FALSE, a recursive thing :-)
        m_bScrollResult = TRUE;
    }
}
// ----------------------------------------------------------------------------
void wxPagingCtrl::SetScrollPos(int orientation, int pos, bool refresh)
{
    m_bScrollResult = FALSE;

    // Make sure we have a tab and the active tab is the next caller.
    if (!HasTab())
        return;
    if (m_pCaller != GetActiveTab())
        return;
    
    // Test and set new values for the scrollbars.
    if (orientation == wxVERTICAL && m_pVertScroll && GetActiveTabInfo()->HandleVert())
    {
        m_pVertScroll->SetThumbPosition(pos);
        if (refresh)
            m_pVertScroll->Refresh();
        m_bScrollResult = TRUE;
    }
    
    if (orientation == wxHORIZONTAL && m_pHorizScroll && GetActiveTabInfo()->HandleHoriz())
    {
        m_pHorizScroll->SetThumbPosition(pos);
        if (refresh)
            m_pHorizScroll->Refresh();
        m_bScrollResult = TRUE;
    }
}
// ----------------------------------------------------------------------------
int wxPagingCtrl::GetScrollThumb(int orientation)
{
    int result = 0;
    
    m_bScrollResult = FALSE;

    // Make sure we have a tab and the active tab is the next caller.
    if (!HasTab())
        return result;
    if (m_pCaller != GetActiveTab())
        return result;
    
    // Test and get value form the scrollbar.       
    if (orientation == wxVERTICAL && m_pVertScroll && GetActiveTabInfo()->HandleVert())
    {
        result = m_pVertScroll->GetThumbSize();
        m_bScrollResult = TRUE;
    }
    
    if (orientation == wxHORIZONTAL && m_pHorizScroll && GetActiveTabInfo()->HandleHoriz())
    {
        result = m_pHorizScroll->GetThumbSize();
        m_bScrollResult = TRUE;
    }
    
    return result;
}
// ----------------------------------------------------------------------------
int wxPagingCtrl::GetScrollPos(int orientation)
{
    int result = 0;
    
    m_bScrollResult = FALSE;

    // Make sure we have a tab and the active tab is the next caller.
    if (!HasTab())
        return result;
    if (m_pCaller != GetActiveTab())
        return result;
    
    // Test and get value form the scrollbar.       
    if (orientation == wxVERTICAL && m_pVertScroll && GetActiveTabInfo()->HandleVert())
    {
        result = m_pVertScroll->GetThumbPosition();
        m_bScrollResult = TRUE;
    }
    
    if (orientation == wxHORIZONTAL && m_pHorizScroll && GetActiveTabInfo()->HandleHoriz())
    {
        result = m_pHorizScroll->GetThumbPosition();
        m_bScrollResult = TRUE;
    }
    
    return result;
}
// ----------------------------------------------------------------------------
int wxPagingCtrl::GetScrollRange(int orientation)
{
    int result = 0;
    
    m_bScrollResult = FALSE;

    // Make sure we have a tab and the active tab is the next caller.
    if (!HasTab())
        return result;
    if (m_pCaller != GetActiveTab())
        return result;
    
    // Test and get value form the scrollbar.       
    if (orientation == wxVERTICAL && m_pVertScroll && GetActiveTabInfo()->HandleVert())
    {
        result = m_pVertScroll->GetRange();
        m_bScrollResult = TRUE;
    }
    
    if (orientation == wxHORIZONTAL && m_pHorizScroll && GetActiveTabInfo()->HandleHoriz())
    {
        result = m_pHorizScroll->GetRange();
        m_bScrollResult = TRUE;
    }
    
    return result;
}

// ----------------------------------------------------------------------------
// event handlers
// ----------------------------------------------------------------------------
void wxPagingCtrl::OnLButtonDown(wxMouseEvent& event)
{
    if (!m_bInitDone)
        return;     // Hell we are not ready to handle this!

    if (m_bCursorChanged)
    {
        int x1, y1, x2, y2;

        // Calc Resize stick location.
        GetTitleCoord(&x1, &y1, &x2, &y2);
        x1 = x2 - RESIZE_STICK_SIZE;

        // Save all info needed and go into drag mode.
        m_bIsDragged     = TRUE;
        m_DagOrigin      = x1; //event.m_x;
        m_DragTitleSpace = m_TitleSpace;
        
        if (m_pTabScroll != NULL && m_pTabScroll->IsShown())
            m_DragMinTitleSpace = m_pTabScroll->GetSize().GetWidth() + RESIZE_STICK_SIZE;
        else
            m_DragMinTitleSpace = RESIZE_STICK_SIZE;
        
        CaptureMouse();
    }
    else
    {
        int x = event.m_x - m_CurentRowOfs; // Add the offset to the x... make it a virtual x
        int y = event.m_y;
        
        int tabNo = HitTest(wxPoint(x,y));
        
        if (tabNo != -1)
        {
            SetActiveTab(tabNo);
            twTabInfo* pInfo = GetTabInfo(tabNo);
            
            wxCommandEvent menuEvent(wxEVT_TAB_LEFT_CLICKED, pInfo->m_TabID); 
            menuEvent.SetEventObject(this);
            
            this->ProcessEvent(menuEvent);
        }
    }
}
// ----------------------------------------------------------------------------
void wxPagingCtrl::OnLButtonUp(wxMouseEvent& event)
{
    if (!m_bInitDone)
        return;     // Hell we are not ready to handle this!

    if (m_bIsDragged)
    {
        // Stop with dragging.
        m_bIsDragged     = FALSE;
        m_bCursorChanged = FALSE;
        SetCursor(m_NormalCursor);
        
        ReleaseMouse();
    }
}
// ----------------------------------------------------------------------------
void wxPagingCtrl::OnRButtonDown(wxMouseEvent& event)
{
    if (!m_bInitDone)
        return;     // Hell we are not ready to handle this!

    int x = event.m_x - m_CurentRowOfs; // Add the offset to the x... make it a virtual x
    int y = event.m_y;
    
    int tabNo = HitTest(wxPoint(x,y));
    
    if (tabNo != -1)
    {
        twTabInfo* pInfo = GetTabInfo(tabNo);
        
        wxCommandEvent menuEvent(wxEVT_TAB_RIGHT_CLICKED, pInfo->m_TabID); 
        menuEvent.SetEventObject(this);
        
        this->ProcessEvent(menuEvent);
    }
}
// ----------------------------------------------------------------------------
void wxPagingCtrl::OnMouseMove(wxMouseEvent& event)
{
    if (!m_bInitDone)
        return;     // Hell we are not ready to handle this!

    // Check if there is a scrollbar reziser.
    if (m_pHorizScroll != NULL && m_pHorizScroll->IsShown())
    {
        int x1, y1, x2, y2;

        // Calc Resize stick location.
        GetTitleCoord(&x1, &y1, &x2, &y2);
        x1 = x2 - RESIZE_STICK_SIZE;

        if (!m_bIsDragged)
        {
            // Check if we are on the resize stick.
            if (event.m_x >= x1 && event.m_y >= y1 &&
                event.m_x <= x2 && event.m_y <= y2)
            {
                if (!m_bCursorChanged)
                {
                    SetCursor(m_ResizeCursor);
                    m_bCursorChanged = TRUE;
                }
            }
            else
            {
                if (m_bCursorChanged)
                {
                    SetCursor(m_NormalCursor);      
                    m_bCursorChanged = FALSE;
                }
            }
        }
        else
        {
            // Calc the new m_TitleSpace... the rest will follow
            m_TitleSpace = wxMax(m_DragTitleSpace + (event.m_x - m_DagOrigin), m_DragMinTitleSpace);

            // Re size Code.
            RecalcLayout(FALSE);

            wxWindowDC dc(this);
            DrawDecorations(dc);
        }
    }
}
// ----------------------------------------------------------------------------
void wxPagingCtrl::OnScroll(wxScrollEvent& event)
{
    if (!m_bInitDone)
        return;     // Hell we are not ready to handle this!

    wxScrollBar* pSender = (wxScrollBar*)event.GetEventObject();
    
    if (pSender == m_pHorizScroll || pSender == m_pVertScroll)
    {
        // Check if we can handle it.
        if (pSender != m_pHorizScroll && pSender != m_pVertScroll && !HasTab())
        {
            event.Skip();
            return;
        }

        // Make new Event... a scroll Event that can be handled by 
        // a wxWindow object and send it to the active tab to be handled.
        wxScrollWinEvent newEvent;
                
        if (pSender == m_pHorizScroll)
            newEvent.SetOrientation(wxHORIZONTAL);
        if (pSender == m_pVertScroll)
            newEvent.SetOrientation(wxVERTICAL);
        
        newEvent.SetPosition(event.GetPosition());
        newEvent.m_eventObject = event.m_eventObject;
        
        // This is !!!not!!! a very clean way to convert a scrollbar event 
        // type too a scrolled window event. A switch does not work 
        // because wxEVT_XXXX_XXXX are not constant, 
        // and thus not allowed in a switch.
        newEvent.m_eventType = (event.m_eventType - wxEVT_SCROLL_TOP) + wxEVT_SCROLLWIN_TOP;
        
        // Let it be processed by the active tab
        GetActiveTab()->GetEventHandler()->ProcessEvent(newEvent);
        // GetActiveTab()->GetEventHandler()->ProcessEvent(event); // Old way.
        
        // Move Focus to the active tab contol.
        GetActiveTab()->SetFocus();
    }
    else
    {
        // Let a other event handler handle this case...
        event.Skip();
    }
}
// ----------------------------------------------------------------------------
void wxPagingCtrl::OnTabScroll(wxSpinEvent& event)
{
    if (!m_bInitDone)
        return;     // Hell we are not ready to handle this!

    if (!m_pTabScroll)
    {
        // There is no way we can handle this...
        event.Skip();
        return;     
    }

    int curX, curY, xMaxSize, yMaxSize;
    int nTitleRowLen, nWholeTabRowLen;
    int nTitleBeginOfs, nTitleEndOfs;
    
    // Get the available draw size. = (xMaxSize - curX)
    GetTitleCoord(&curX, &curY, &xMaxSize, &yMaxSize);
    nTitleRowLen = xMaxSize - curX;

    // Subtract the Resize stick if it is displayed!
    if (m_pTabScroll != NULL && m_pTabScroll->IsShown())
        nTitleRowLen -= RESIZE_STICK_SIZE;

    // Get total the length of the title bar.
    nWholeTabRowLen = GetWholeTabRowLen();

    // Calc the begin and end offset.
    nTitleBeginOfs  = m_pTabScroll->GetValue() * GetTabScrollRate();
    nTitleEndOfs    = nTitleBeginOfs + nTitleRowLen;
        
    // Correct the offset to match boarders of scrolling
    if (nTitleBeginOfs < 0)
    {
        nTitleBeginOfs = 0;
    }
    else
    {
        if (nTitleEndOfs > nWholeTabRowLen)
        {
            nTitleBeginOfs = nWholeTabRowLen - nTitleRowLen;
            
            if (nTitleBeginOfs < 0)
            {
                nTitleBeginOfs = 0;
            }

            // Adjust spin button so the max View value is the max scroll value 
            m_pTabScroll->SetValue((int)ceil(nTitleBeginOfs * 1.0 / GetTabScrollRate()));
        }
    }

    // Finally set the offset. (convert from pos to negative)
    m_CurentRowOfs = (- nTitleBeginOfs);

    // let' it automatically disable itself if it's time
    m_pTabScroll->Enable(nTitleRowLen < GetWholeTabRowLen() || m_CurentRowOfs < 0);        
    
    // repaint title bars
    wxWindowDC dc(this);
    DrawDecorations(dc);

    // Move Focus to the active tab contol.
    GetActiveTab()->SetFocus();
}
