/////////////////////////////////////////////////////////////////////////////
// Name:        tabctrl.cpp
// Purpose:     wxTabbedCtrl implementation
// Author:      Aleksandras Gluchovas & Hans Van Leemputten
// Modified by:
// Created:     21/May/2001
// RCS-ID:      $Id: tabbctrl.cpp,v 1.3 2001/09/15 09:25:55 hansvl Exp $
// Copyright:   (c) Hans Van Leemputten & Aleksandras Gluchovas
// Licence:     wxWindows license
/////////////////////////////////////////////////////////////////////////////

// ============================================================================
// declarations
// ============================================================================

// ----------------------------------------------------------------------------
// headers
// ----------------------------------------------------------------------------

#ifdef __GNUG__
    #pragma implementation "tabbctrl.h"
#endif

// For compilers that support precompilation, includes "wx.h".
#include "wx/wxprec.h"

#ifdef __BORLANDC__
    #pragma hdrstop
#endif

#ifndef WX_PRECOMP
    #include "wx/wx.h"
#endif

#include "wx/wspp/tabbctrl.h"

// ============================================================================
// wxTabbedCtrl implementation.
// ============================================================================
IMPLEMENT_DYNAMIC_CLASS(wxTabbedCtrl, wxFlipperCtrlBase)

// ----------------------------------------------------------------------------
// event tables
// ----------------------------------------------------------------------------
BEGIN_EVENT_TABLE(wxTabbedCtrl, wxFlipperCtrlBase)
    EVT_PAINT     (wxFlipperCtrlBase::OnPaint )
    EVT_SIZE      (wxFlipperCtrlBase::OnSize  )
    EVT_LEFT_DOWN (wxTabbedCtrl::OnLButtonDown)
    EVT_RIGHT_DOWN(wxTabbedCtrl::OnRButtonDown)
END_EVENT_TABLE()

// ----------------------------------------------------------------------------
// creation
// ----------------------------------------------------------------------------
wxTabbedCtrl::wxTabbedCtrl()
    : wxFlipperCtrlBase(),
      m_bInitDone(FALSE)     // Not finished with Init.
{
    DoInitTabbedCtrl();
}
// ----------------------------------------------------------------------------
wxTabbedCtrl::wxTabbedCtrl( wxWindow *parent,
                            wxWindowID id,
                            const wxPoint& pos,
                            const wxSize& size,
                            long  style,
                            const wxString& name)
    : wxFlipperCtrlBase(),   // Temp 
      m_bInitDone(FALSE)     // Not finished with Init.
{
    Create(parent, id, pos, size, style, name);
}
// ----------------------------------------------------------------------------
wxTabbedCtrl::~wxTabbedCtrl()
{
    // Done by base class.
}
// ----------------------------------------------------------------------------
bool wxTabbedCtrl::Create( wxWindow *parent, wxWindowID id,
                        const wxPoint& pos,
                        const wxSize& size,
                        long  style,
                        const wxString& name)
{
    bool bResult = wxFlipperCtrlBase::Create(parent, id, pos, size, style, name);

    DoInitTabbedCtrl();

    return bResult;
}
// ----------------------------------------------------------------------------
void wxTabbedCtrl::DoInitTabbedCtrl()
{
    // Init the pen's.
    m_WhitePen = wxPen(wxColour(255,255,255), 1, wxSOLID);
    m_GrayPen  = wxPen(wxColour(192,192,192), 1, wxSOLID);
    m_DarkPen  = wxPen(wxColour(128,128,128), 1, wxSOLID);
    m_BlackPen = wxPen(wxColour( 0,  0,  0), 1, wxSOLID);
    m_TabPen   = wxPen(wxSystemSettings::GetSystemColour(wxSYS_COLOUR_BTNFACE), 1, wxSOLID);

    // Init the tab brush.
    m_TabBrush = wxBrush(m_TabPen.GetColour(), wxSOLID);

    // Init Size/layout related variables
    m_ImageTextGap = 2; /* Make sure to take this gap into acount.           */
    m_TitleHeight  = 0; /* Make sure to set it to the max value of all tabs  */
    m_FirstTitleGap= 11;/* This is the gap before the first tab.             */
    m_TitleHorizGap= 4; /* This is a global class var.                       */
    m_TitleVertGap = 3; /* This is a global class var.                       */

    // Ok now we are done with Init.
    m_bInitDone = TRUE;
}

// ----------------------------------------------------------------------------
// Misc.
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// width of the decorations border (4 shade-lines), should not be changed
// ----------------------------------------------------------------------------
#define BORDER_SZ 4
// ----------------------------------------------------------------------------
void wxTabbedCtrl::DrawShadedRect(int x, int y, int width, int height,
                                    wxPen& upperPen, wxPen& lowerPen, wxDC& dc
                                    )
{
    // darw the lightened upper-left sides of the rectangle
    dc.SetPen(upperPen);
    dc.DrawLine(x,y, x, y + height - 1); // vert
    dc.DrawLine(x,y, x + width - 1,  y); // horiz
    
    // draw the unenlightened lower-right sides of the rectangle    
    dc.SetPen(lowerPen);
    dc.DrawLine(x + width - 1,  y, x + width - 1, y + height - 1); // vert
    dc.DrawLine(x, y + height - 1, x + width, y + height - 1);     // horiz
}
// ----------------------------------------------------------------------------
void wxTabbedCtrl::EasySizeBodyContent(bool repaint)
{
    // Get all coordinates ready to use.
    int curX, curY, xSize, ySize;
    GetBodyCoord(&curX, &curY, &xSize, &ySize); 

    xSize = xSize - curX - BORDER_SZ*2-1;
    ySize = ySize - curY - BORDER_SZ*2-1;
    curX = curX + BORDER_SZ;
    curY = curY  + BORDER_SZ;       
    
    // Make sure the body is displayed OK.
    SizeBodyContent(curX, curY, xSize, ySize, repaint);
}
// ----------------------------------------------------------------------------
void wxTabbedCtrl::GetTitleCoord(int *x1, int *y1, int *x2, int *y2)
{
    wxFlipperCtrlBase::GetTitleCoord(x1, y1, x2, y2);

    // Add the extra gap beffore the first tab.
    *x1 = *x1 + m_FirstTitleGap;
}

// ----------------------------------------------------------------------------
void wxTabbedCtrl::DrawDecorations(wxDC& dc)
{
    if (!m_bInitDone)
        return;     // Hell we are not ready to handle this!

    int width, height;
    GetClientSize(&width, &height);
    
    // check if there's at least a bit of space to draw things
    if (width  < m_HorizGap*2 + BORDER_SZ*2+1 ||
        height < m_VertGap*2  + BORDER_SZ*2+1 + m_TitleHeight)
        return;
    
    /********************************************************/
    /* step #1 - draw border around the tab content area    */
    /********************************************************/

    // setup position for kind of "pencil"
    int curX, curY, xSize, ySize;
    GetBodyCoord(&curX, &curY, &xSize, &ySize); 
    
    // layer 1 (upper white)
    DrawShadedRect(curX+0, curY+0, xSize-0, ySize-0,
        m_WhitePen, m_BlackPen, dc );
    
    // layer 2 (upper gray)
    DrawShadedRect(curX+1, curY+1, xSize-2-1, ySize-2-1,
        m_GrayPen, m_GrayPen, dc );
    
    // layer 3 (upper darkGray)
    DrawShadedRect(curX+2, curY+2, xSize-3-2, ySize-3-2,
        m_DarkPen, m_WhitePen, dc );
    
    // layer 4 (upper black)
    DrawShadedRect(curX+3, curY+3, xSize-4-3, ySize-4-3,
        m_BlackPen, m_GrayPen, dc );
    
    // add non-siemtric layer from the lower-right side (confroming to MFC-look)
    dc.SetPen(m_DarkPen);
    dc.DrawLine(curX+1, curY + ySize - 2, curX + xSize - 1, curY + ySize - 2);   // horiz
    dc.DrawLine(curX + xSize - 2, curY + 1, curX + xSize - 2, curY + ySize - 2); // vert
    
    /********************************************************/
    /* step #2 - draw tab title bars                        */
    /********************************************************/
    
    int xMaxSize, yMaxSize;
    GetTitleCoord(&curX, &curY, &xMaxSize, &yMaxSize);
    xSize = 0;
    ySize = 0;
    
    int tabNo = 0;
    wxNode* pNode = m_Tabs.First();
    
    while(pNode)
    {
        // "hard-coded metafile" for decorations
        twTabInfo& tab = *((twTabInfo*)(pNode->Data()));

        // We need to make sure we do not draw more tabs than there is space for.
        if ((curX + tab.m_Dims.x) > xMaxSize)
        {
            // TBD:: Handel this on the fly...
            wxLogWarning(wxT("Unable to draw tab, no more space left to draw!"));
            break;
        }
        
        // Set the size of the tab to draw.
        xSize = tab.m_Dims.x;
        ySize = m_TitleHeight;
        
        // Erase tab surface
        dc.SetPen(m_TabPen);
        dc.SetBrush(m_TabBrush);
        dc.DrawRectangle(curX, curY, xSize, ySize);

        // Remove some lines from the border, to show the active one in front.
        if (tabNo == m_ActiveTab)
        {
            dc.DrawLine(curX+1, curY-2, curX+xSize-2, curY-2);
            dc.DrawLine(curX+1, curY-1, curX+xSize-2, curY-1);
        }
        
        // Draw the tab borders.
        dc.SetPen(m_WhitePen);
        if (tabNo == m_ActiveTab)
            dc.DrawLine(curX, curY-2, curX, curY+ySize-2);
        else
            dc.DrawLine(curX, curY, curX, curY+ySize-2);
        
        dc.SetPen(m_DarkPen);
        dc.DrawLine(curX+1, curY+ySize-3, curX+1, curY+ySize-1); // to pix down
        dc.DrawLine(curX+2, curY+ySize-2, curX+xSize-2, curY+ySize-2);
        dc.DrawLine(curX+xSize-3, curY+ySize-3, curX+xSize-2, curY+ySize-3);
        if (tabNo == m_ActiveTab)
            dc.DrawLine(curX+xSize-2, curY+ySize-3, curX+xSize-2, curY-3);
        else
            dc.DrawLine(curX+xSize-2, curY+ySize-3, curX+xSize-2, curY-1);
        
        dc.SetPen(m_BlackPen);
        dc.DrawLine(curX+xSize-1, curY, curX+xSize-1, curY+ySize-2);
        dc.DrawLine(curX+xSize-2, curY+ySize-2, curX+xSize-3, curY+ySize-2);
        dc.DrawLine(curX+xSize-3, curY+ySize-1, curX+1, curY+ySize-1);
        
        /********************************************************/
        /* draw image and (or without) text centered within the */
        /* title bar rectangle                                  */
        /********************************************************/

        // Check if we may display a image and also check 
        // if there is a valid image, if so display it.
        if (m_LayoutType != wxTITLE_BORDER_ONLY && tab.HasImg())
        { 
            wxBitmap& bmp = tab.GetImg();
            
            if (bmp.Ok())
            {
                bool hasMask = bmp.GetMask() != NULL;
                
                dc.DrawBitmap(bmp, curX + m_TitleHorizGap, curY + (ySize - tab.ImgHeight()) / 2,
                    hasMask);
            }           
        }
        
        // Check if we may display some text and also check 
        // if there is some text avalible to display, if so display it.
        if (m_LayoutType == wxTITLE_IMG_AND_TEXT && tab.HasText())
        {
            long x,w,h;
            
            // set select default font of the window into it's device context
            dc.SetFont(GetLabelingFont());
            
            dc.SetTextBackground(GetBackgroundColour());
            
            dc.GetTextExtent(tab.m_Text, &w, &h);
            
            x = curX + m_TitleHorizGap +
                tab.ImgWidth() + tab.ImageToTxtGap(m_ImageTextGap);
            
            dc.DrawText(tab.GetText(), x, curY + (ySize - h) / 2);
        }

        // Get ready for the next loop.
        pNode = pNode->Next();
        ++tabNo;
        curX += xSize;      
    } // end of `while (pNode)'
}
// ----------------------------------------------------------------------------
void wxTabbedCtrl::RecalcLayout(bool andRepaint)
{   
    if (!m_bInitDone)
        return;     // Hell we are not ready to handle this!
    
    int width, height;
    GetClientSize(&width, &height);
    
    // check if there's at least a bit of space to draw things
    if (width  < m_HorizGap*2 + BORDER_SZ*2+1 ||
        height < m_VertGap*2  + BORDER_SZ*2+1 + m_TitleHeight)
        return;

    int curX, curY, xSize, ySize;
    GetTitleCoord(&curX, &curY, &xSize, &ySize);

    int org_curX = curX; // To keep track of the original value.

    int denomiator = GetTabCount();
    if (denomiator <= 0) denomiator = 1;
    int BorderOnlyWidth = (xSize - m_FirstTitleGap - curX) / denomiator;

    int OldTitleHeight = m_TitleHeight;

    wxNode* pNode = NULL;
    
    /****************************************************************/
    /* pass #1 - try to layout assuming it's wxTITLE_IMG_AND_TEXT   */
    /****************************************************************/
    
    if (m_LayoutType == wxTITLE_IMG_AND_TEXT)
    {
        pNode = m_Tabs.First();
        
        curX = org_curX; // ReInit.
        m_TitleHeight = 0;
        
        while(pNode)
        {
            twTabInfo& tab = *((twTabInfo*)(pNode->Data()));
            
            wxWindowDC dc(this);
            long w,h;
            
            // set select default font of the window into it's device context
            dc.SetFont(GetLabelingFont());
            
            dc.GetTextExtent(tab.m_Text, &w, &h);
            
            tab.m_Dims.x = w + tab.ImageToTxtGap(m_ImageTextGap) +
                tab.ImgWidth() + m_TitleHorizGap*2;
            
            tab.m_Dims.y  = wxMax(h, tab.ImgHeight()) + m_TitleVertGap*2;
            m_TitleHeight = wxMax(m_TitleHeight, tab.m_Dims.y);
            
            curX += tab.m_Dims.x;
            
            pNode = pNode->Next();
        }
        
        // make all title bars of equel height  
        pNode = m_Tabs.First();
        
        while(pNode)
        {
            ((twTabInfo*)(pNode->Data()))->m_Dims.y = m_TitleHeight;
            pNode = pNode->Next();
        }
        
        // if curX hasn't ran out of bounds, leave TITLE_IMG layout and return
        if (curX >= xSize)
        {
            if (OldTitleHeight != m_TitleHeight)
                // re-size the body content to the new m_TitleHeight.
                EasySizeBodyContent(andRepaint);
            
            return;
        }
    }
    
    /****************************************************************/  
    /* pass #2 - try to layout assuming wxTITLE_IMG_ONLY            */
    /****************************************************************/

    if (m_LayoutType == wxTITLE_IMG_ONLY)       
    {
        pNode = m_Tabs.First();
        
        curX = org_curX; // ReInit
        
        while(pNode)
        {
            twTabInfo& tab = *((twTabInfo*)(pNode->Data()));
            
            if (tab.HasImg())
            {
                tab.m_Dims.x = tab.ImgWidth()  + m_TitleHorizGap*2;
                tab.m_Dims.y = tab.ImgHeight() + m_TitleVertGap*2;
            }
            else
            {
                tab.m_Dims.x = BorderOnlyWidth;
                tab.m_Dims.y = m_TitleHeight;
            }

            m_TitleHeight = wxMax(m_TitleHeight, tab.m_Dims.y);
            
            curX += tab.m_Dims.x;
            
            pNode = pNode->Next();
        }

        // make all title bars of equel height  
        pNode = m_Tabs.First();
        
        while(pNode)
        {
            ((twTabInfo*)(pNode->Data()))->m_Dims.y = m_TitleHeight;
            pNode = pNode->Next();
        }
        
        // if curX has'nt ran out of bounds, leave IMG_ONLY layout and return
        if (curX >= xSize)
        {
            if (OldTitleHeight != m_TitleHeight)
                // re-size the body content to the new m_TitleHeight.
                EasySizeBodyContent(andRepaint);
            
            return;
        }
    }

    /****************************************************************/
    /* pass #3 - set the narrowest layout wxTITLE_BORDER_ONLY       */
    /****************************************************************/

    if (m_LayoutType == wxTITLE_BORDER_ONLY)
    {   
        pNode = m_Tabs.First();
        
        while(pNode)
        {
            twTabInfo& tab = *((twTabInfo*)(pNode->Data()));
            
            if (m_TitleHeight <= 0)
                m_TitleHeight = BORDER_ONLY_HEIGHT; // Set a default height...
            
            tab.m_Dims.x = BorderOnlyWidth;
            tab.m_Dims.y = m_TitleHeight;
            
            pNode = pNode->Next();
        }
    }

    // Make sure this gets done any way. 
    // Defiantly needed when recalcing layout on tab added!
    // re-size the body content to the new m_TitleHeight.
    EasySizeBodyContent(andRepaint);
}

// ----------------------------------------------------------------------------
// event handlers
// ----------------------------------------------------------------------------
void wxTabbedCtrl::OnLButtonDown(wxMouseEvent& event)
{
    if (!m_bInitDone)
        return;     // Hell we are not ready to handle this!

    int x = event.m_x;
    int y = event.m_y;
    
    int tabNo = HitTest(wxPoint(x,y));
    
    if (tabNo != -1)
    {
        SetActiveTab(tabNo);
        twTabInfo* pInfo = GetTabInfo(tabNo);
        
        wxCommandEvent menuEvent(wxEVT_TAB_LEFT_CLICKED, pInfo->m_TabID); 
        menuEvent.SetEventObject(this);
        
        this->ProcessEvent(menuEvent);
    }
}
// ----------------------------------------------------------------------------
void wxTabbedCtrl::OnRButtonDown(wxMouseEvent& event)
{
    if (!m_bInitDone)
        return;     // Hell we are not ready to handle this!

    int x = event.m_x;
    int y = event.m_y;
    
    int tabNo = HitTest(wxPoint(x,y));
    
    if (tabNo != -1)
    {
        twTabInfo* pInfo = GetTabInfo(tabNo);
        
        wxCommandEvent menuEvent(wxEVT_TAB_RIGHT_CLICKED, pInfo->m_TabID); 
        menuEvent.SetEventObject(this);
        
        this->ProcessEvent(menuEvent);
    }
}
